package algorithms;

import java.util.Random;

import Jama.EigenvalueDecomposition;
import Jama.Matrix;

import static java.lang.Math.*;

public class ICA {

	public int ID;

	public double[][] ev; // eigenvalues
	public double wscCost; // Cost of White Space Compensation in Bits
	boolean verbose = false;

	public double[] shift;
	public double transform[][];
	public Matrix transformMatrix;
	public double[][] whitened;
	public Matrix whithenedMatrix;

	double[][] centered;
	public Matrix mw;
	public Matrix transposedMW;

	public Matrix transformTIMESmw;
	

	public ICA(int dim) {
		this.dim = dim;
		transform = new double[dim][dim];
		transformMatrix = new Matrix(dim,dim);
		shift = new double[dim];
		mw = new Matrix(dim,dim);
		w = new double[dim][dim];
	}

	double log2 = log(2d);
	
	int dim, i, j, k, m, numObj, alliter, iter;
	double sum, eRight, scalar, g;
	double[] diag, old_w, wUnorth, eLeft;
	double[][] result, w;
	Matrix centeredMatrix, c, evm, ewm, res;
	EigenvalueDecomposition e;
	Random r = new Random();;

	boolean firstGuess = true;
	public void reset(){
		
//		w = new double[dim][dim] ;
		for (k=0 ; k<dim ; k++) {
	        for (i=0; i<dim ; i++)
	            w[k][i] = r.nextDouble();
	        sum = 0.0 ;
	        for (i=0; i<dim ; i++)
	            sum += w[k][i]*w[k][i] ;
	        for (i=0; i<dim ; i++)
	            w[k][i] /= sqrt(sum) ;
		}
		firstGuess = true;
		maxIterations = 1000;
	}
	
	int maxIterations = 100;
	double scalarTanH;
//	public double [][] optimizeEntropy() {
//        // fastICA algorithm acc. to p. 194 with deflationary orthogonalization
////        int numObj= whitened.length ;
////        dim = whitened[0].length ;
//        alliter = 0 ;
//        for (k=0 ; k<dim ; k++) {
//            iter = 0 ;
//            do {
//                old_w = w[k].clone() ;
//                eLeft = new double[dim] ;
//                eRight = 0.0 ;
//                for (int i=0 ; i<numObj ; i++) {
//                    scalar = 0.0 ;
//                    for (j=0 ; j<dim ; j++)
//                        scalar += w[k][j]*whitened[i][j] ;
//                    g = (scalarTanH = tanh(scalar)) / numObj ;
//                    for (j=0 ; j<dim ; j++)
//                        eLeft[j] += g * whitened[i][j] ;
//                    eRight +=  ( 1d - scalarTanH*scalarTanH) / numObj ;
//                }
//                for (j=0 ; j<dim ; j++)
//                    w[k][j] = eLeft[j] - eRight * w[k][j] ;
//                // deflationary orthogonalize w[n] w.r.t. vectors w[0]..w[n-1]
//                wUnorth = w[k].clone();
//                for (m=0 ; m<k ; m++) {
//                    scalar = 0.0 ;
//                    for (j=0 ; j<dim ; j++)
//                        scalar += wUnorth[j] * w[m][j] ;
//                    for (j=0 ; j<dim ; j++)
//                        w[k][j] -= scalar * w[m][j] ;
//                }
//                sum = 0.0 ;
//                for (i=0; i<dim ; i++)
//                    sum += w[k][i]*w[k][i] ;
//                for (i=0; i<dim ; i++)
//                    w[k][i] /= sqrt(sum) ;
//                sum = 0.0 ;
//                for (i=0; i<dim ; i++)
//                    sum += w[k][i] * old_w[i] ;
//                iter ++ ;
//            } while (abs(sum)<0.99999999 && iter<maxIterations); // <0.9999 && iter<100
//        
//            for (i=0; i<dim ; i++)
//            	mw.set(k,i, w[k][i]);
//            
//            
//            alliter += iter ;
//        }
//        
//        
//        
//        // To transform data such that independent components are located at
//        // the coordinate axes, we have to multiply the data with w^{-1} which
//        // does not coincide with wn' (unlike the Eigenvector matrix of PCA)
//        
//        //(new Matrix (data1)).times((new Matrix (w)).transpose()).print (6,4) ;
////        mw = new Matrix(w);
//        transposedMW = mw.transpose();
//        maxIterations = 100;
//        firstGuess = false;
//        
//        return w ;
//    }

	double res2, xsq, x;

	public void doICA(double[][] data, int coordinatesUsedUptoIndex) {
		// INIT
//		this.data = data;
		numObj = coordinatesUsedUptoIndex;
		
		centered = new double[numObj][dim];
		centeredMatrix = new Matrix(numObj, dim);
		ev = null;
		wscCost = 0d;
//		whitened = new double[numObj][dim];
		
		// CENTER DATA
		for (i = 0; i < dim; i++) {
			shift[i] = 0.0;
			for (j = 0; j < numObj; j++)
				shift[i] += data[j][i];
			shift[i] /= numObj;
			for (j = 0; j < numObj; j++)
				centeredMatrix.set(j, i, centered[j][i] = data[j][i] - shift[i]);
		}
		
		// WHITEN
//		whitened = new double[numObj][dim];
//		whithenedMatrix = new Matrix(numObj, dim);
//		centeredMatrix = new Matrix(centered);
		c = centeredMatrix.transpose().times(centeredMatrix).times(1.0 / numObj);
		e = new EigenvalueDecomposition(c);
		evm = e.getV();
		ewm = e.getD();
		ev = ewm.getArrayCopy();
		diag = new double[dim];
		for (i = 0; i < dim; i++) {
			diag[i] = sqrt(min(max(ewm.get(i, i), 1.0e-6), 1.0e+6));
			wscCost += log(diag[i]) / log2;
		}

		for (i = 0; i < dim; i++) {
			for (j = 0; j < dim; j++) {
				transformMatrix.set(i,j,transform[i][j] = evm.get(i, j) / diag[j]);
			}
		}

		whitened = (whithenedMatrix = centeredMatrix.times(transformMatrix)).getArrayCopy();
		
		// OPTIMIZE ENTROPY
	    for (k=0 ; k<dim ; k++) {
            iter = 0 ;
            do {
                old_w = w[k].clone() ;
                eLeft = new double[dim] ;
                eRight = 0.0 ;
                for (int i=0 ; i<numObj ; i++) {
                    scalar = 0.0 ;
                    for (j=0 ; j<dim ; j++)
                        scalar += w[k][j]*whitened[i][j] ;
                    g = (scalarTanH = tanh(scalar)) / numObj ;
                    for (j=0 ; j<dim ; j++)
                        eLeft[j] += g * whitened[i][j] ;
                    eRight +=  ( 1d - scalarTanH*scalarTanH) / numObj ;
                }
                for (j=0 ; j<dim ; j++)
                    w[k][j] = eLeft[j] - eRight * w[k][j] ;
                // deflationary orthogonalize w[n] w.r.t. vectors w[0]..w[n-1]
                wUnorth = w[k].clone();
                for (m=0 ; m<k ; m++) {
                    scalar = 0.0 ;
                    for (j=0 ; j<dim ; j++)
                        scalar += wUnorth[j] * w[m][j] ;
                    for (j=0 ; j<dim ; j++)
                        w[k][j] -= scalar * w[m][j] ;
                }
                sum = 0.0 ;
                for (i=0; i<dim ; i++)
                    sum += w[k][i]*w[k][i] ;
                for (i=0; i<dim ; i++)
                    w[k][i] /= sqrt(sum) ;
                sum = 0.0 ;
                for (i=0; i<dim ; i++)
                    sum += w[k][i] * old_w[i] ;
                iter ++ ;
            } while (abs(sum)<0.99999999 && iter<maxIterations); // <0.9999 && iter<100
        
            for (i=0; i<dim ; i++)
            	mw.set(k,i, w[k][i]);
            
        }
        
        
        
        // To transform data such that independent components are located at
        // the coordinate axes, we have to multiply the data with w^{-1} which
        // does not coincide with wn' (unlike the Eigenvector matrix of PCA)
        
        //(new Matrix (data1)).times((new Matrix (w)).transpose()).print (6,4) ;
//	        mw = new Matrix(w);
        transposedMW = mw.transpose();
        
        transformTIMESmw = transformMatrix.times(transposedMW);
        
        maxIterations = 100;
        firstGuess = false;
	        		
	}

//	public double g_slope(double x) {
//		return 1.0 - Math.pow(Math.tanh(x), 2);
//	}

//	public Matrix covar(Matrix data) {
//		return data.transpose().times(data).times(1.0 / data.getRowDimension());
//	}

}
