/*
 * Kmeans.java
 *
 * Created on February 17, 2005, 3:47 PM
 */

package clusterer;

/**
 *
 * @author  Administrator
 */
import java.util.*;
import compr.*;
import gengauss.*;
import visu.VisuKMeans;
import visu.VisuInfo;
import eval.SortPack;
import ica.Ica;
import ica.IcaCost;
import Jama.*;

public class IcaSplit {
    int k;
    DataObject[] data;
    DataObject[] cluster1;  //objects assigned to cluster 1
    DataObject[] cluster2;  //objects assigned to cluster 2
    int dim;
    int numObj;
    double[][] means;
    int iteration;
    boolean verbose = false;
    
    
    
    
    /** Creates a new instance of Kmeans */
    public IcaSplit(int k, DataObject[] data){
        this.k = k;
        this.data = data;
        numObj = data.length;
        dim = data[0].coord.length;
        means = new double[k][dim];
        iteration = 0;
    }
    
    
    
    public DataObject[] getObjects(int id){
        if(id == 1)
            return cluster1;
        if(id == 2)
            return cluster2;
        else
            return null;
    }
    
    
    
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
    
    
    public DataObject[] run(){
        IcaCluster[] res = new IcaCluster[k];
        Ica ica = new Ica();
        GenGauss gg = new GenGauss() ;
        double[][] d = new double[data.length][] ;
        for (int j=0 ; j< data.length ; j++)
            d[j] = data[j].coord.clone();
        double[][] transform = new double[dim][dim];
        double[][] retransform = new double[dim][dim];
        double[] shift = new double[dim];
        double centered[][] = ica.centerData(d, shift);
        double whitened[][] = ica.whitenData(centered, transform, retransform);
        //displayData(whitened, "white space");
        double bestcost = Double.MAX_VALUE;
        double[][] w = ica.optimizeEntropy(whitened);
        Matrix mwhitened = new Matrix(whitened);
        Matrix mw = new Matrix(w);
        double[][] projectedWhitened =( mwhitened.times(mw.transpose())).transpose().getArrayCopy();
        double[][] genGaussParams = new double[dim][] ;
        // now determine the shape parameter for every dimension
        //START: SELECT THE MOST SUBGAUSSIAN DIRECTION
        double betaMin = Double.MAX_VALUE;
        int minIndex = -1;
        for (int j=0 ; j<dim ; j++) {
            genGaussParams [j] = EpdMineo.optimizeMSP(projectedWhitened [j]) ;  //beta: genGaussParams[2]
            //System.out.println("beta: " + genGaussParams[j][2]);
            if(genGaussParams[j][2] < betaMin){
                betaMin = genGaussParams[j][2];
                minIndex = j;
            }
        }
        //END: SELECT THE MOST SUBGAUSSIAN DIRECTION
        //START: SELECT THE MOST NON-GAUSSIAN DIRECTION
//         double betaMin = 0.0;
//        int minIndex = -1;
//        for (int j=0 ; j<dim ; j++) {
//            genGaussParams [j] = EpdMineo.optimizeMSP(projectedWhitened [j]) ;  //beta: genGaussParams[2]
//            //System.out.println("beta: " + genGaussParams[j][2]);
//            if(Math.abs(genGaussParams[j][2]) > betaMin){
//                betaMin = genGaussParams[j][2];
//                minIndex = j;
//            }
//        }
        //END: SELECT THE MOST NON-GAUSSIAN DIRECTION
        double[] projCoord = new double[numObj];
        double[] wRow = new double[dim];
        //select row number minIndex of w
        for(int i = 0; i < dim; i++)
            wRow[i] = w[minIndex][i];
        //transpose wRow
        Matrix wt = rowVector(wRow).transpose();
        //compute projCoord = data * transform * wt
        for(int i = 0;  i < numObj; i++){
            projCoord[i] = rowVector(data[i].coord).times(new Matrix(transform)).times(wt).get(0,0);
            //System.out.println(projCoord[i]);
        }
        //50% split: sort projCoord
        SortPack[] s = SortPack.array(projCoord);
        int m = s.length/2;
        //set clusterIDs: index < m: ID 0, index >=m: ID1
        for(int i = 0; i < data.length; i++)
            if(i < m)
                data[s[i].index].clusterID = 1;
            else
                data[s[i].index].clusterID = 2; //clusterIDs 1 and 2 as in KMeans
        //split into 2 clusters
        res[0] = new IcaCluster();
        res[0].ID = 0;
        res[0].numObj = s.length/2;
        res[1] = new IcaCluster();
        res[1].ID = 1;
        res[1].numObj = s.length - s.length/2;
        //------------------START: assign points with epd-bisecting k-means, Problem: Kosten Double.MAX_VALUE, cluster 0 frisst im ersten Schritt alle Punkte weg---------------------------------------------------------
//        boolean terminated = false;
//        int iter = 0;
//        while(!terminated && iter < 100){
//            //determine genGaussParams for each cluster
//            double[][] genGaussParams1D = new double[k][];
//            for(int i = 0; i < k; i++){
//                double[] coordK = new double[res[i].numObj];
//                int counter = 0;
//                for(int j = 0; j < numObj; j++){
//                    if (data[j].clusterID == res[i].ID){
//                        coordK[counter] = projCoord[j];
//                        counter++;
//                    }
//                }
//                genGaussParams1D[i] = EpdMineo.optimizeMSP(coordK);
//            }
//            //print parameters
//            for(int i = 0; i < res.length; i++)
//                System.out.println("cluster: " + i + " mean: " + genGaussParams1D[i][0] + " sigma: " + genGaussParams[i][1] + " beta: " + genGaussParams[i][2]);
//            GenGauss g = new GenGauss();
//            boolean clusterChanged = false;
//            //reset number of objects
//            for(int i = 0; i < k; i++)
//                res[i].numObj = 0;
//            for(int i = 0; i < numObj; i++){
//                double minCost = Double.MAX_VALUE;
//                int min= -1;
//                //assign points
//                for(int j = 0; j < k; j++){
//                    double costDist = g.costV(projCoord[i], genGaussParams[j]);
//                    //System.out.println("costDist: "  + costDist);
//                    if(Double.isInfinite(costDist))
//                        System.err.println("g.costV("+ projCoord[i] + ",  genGaussParams[j]: " + genGaussParams[j][0] + " " + genGaussParams[j][1] + " " + genGaussParams[j][2] + " Double.MAX_VALUE");
//                    if(costDist < minCost){
//                        minCost = costDist;
//                        min = j;
//                    }
//                }
//                res[min].numObj++;
//                if(data[i].clusterID != min){
//                    data[i].clusterID = min;
//                    clusterChanged = true;
//                }
//            }
//            //TEST:  check clusterID
//            int count0 = 0;
//            int count1 = 0;
//            for(int i = 0; i < numObj; i++){
//                if(data[i].clusterID == 0)
//                    count0++;
//                if(data[i].clusterID == 1)
//                    count1++;
//            }
//            System.out.println("Cluster 0: " + res[0].numObj + " " + count0);
//            System.out.println("Cluster 1: " + res[1].numObj + " " + count1);
//            //TEST
//            if(! clusterChanged)
//                terminated = true;
//            System.out.println("iter: " + iter);
//            iter++;
//        }
        ///------------------END: epd bisecting k-means-------------------------------------------
        //-------------------START: assign points with bisecting k-means -------------------------
        boolean terminated = false;
        int iter = 0;
        double[] means = new double[2];
        //init with the farthest objects
        means[0] = s[0].value;
        means[1] = s[s.length-1].value;
        while(!terminated){
            boolean clusterChanged = false;
            //reset number of objects
            for(int i = 0; i < k; i++)
                res[i].numObj = 0;
            //assign points
            for(int i = 0; i < numObj; i++){
                double minDist = Double.MAX_VALUE;
                int min= -1;
                for(int j = 0; j < k; j++){
                    double dist = Math.sqrt((projCoord[i] - means[j]) * (projCoord[i] - means[j]));
                    if(dist < minDist){
                        minDist = dist;
                        min = j;
                    }
                }
                res[min].numObj++;
                if(data[i].clusterID != (min+1)){
                    data[i].clusterID = min+1;
                    clusterChanged = true;
                }
            }
            //update means
            means[0] = 0.0;
            means[1] = 0.0;
            for(int i = 0; i < numObj; i++){
                if(data[i].clusterID == 1)
                    means[0] += projCoord[i];
                if(data[i].clusterID == 2)
                    means[1] += projCoord[i];
            }
            means[0]/=(double)res[0].numObj;
            means[1]/=(double)res[1].numObj;
//            System.out.println("Cluster0: numObj: " + res[0].numObj + " mean: " + means[0]);
//            System.out.println("Cluster1: numObj: " + res[1].numObj + " mean: " + means[1]);
            if(! clusterChanged)
                terminated = true;
            //System.out.println("iter: " + iter);
            iter++;
        }
        //---END: bisecting k-means---------------------------------------------------------------------
        //set cluster1 and cluster2
        int count1 = 0;
        int count2 = 0;
        for(int i = 0; i < numObj; i++){
            if(data[i].clusterID == 1)
                count1++;
            if(data[i].clusterID == 2)
                count2++;
        }
        cluster1 = new DataObject[count1];
        count1 = 0;
        for(int i = 0; i < numObj; i++){
            if(data[i].clusterID == 1){
                cluster1[count1] = data[i];
                count1++;
            }
        }
        cluster2 = new DataObject[count2];
        count2 = 0;
        for(int i = 0; i < numObj; i++){
            if(data[i].clusterID == 2){
                cluster2[count2] = data[i];
                count2++;
            }
        }
        
        return data;
    }
    
    
    public IcaCluster[] split(){
        IcaCluster[] res = new IcaCluster[k];
        Ica ica = new Ica();
        GenGauss gg = new GenGauss() ;
        double[][] d = new double[data.length][] ;
        for (int j=0 ; j< data.length ; j++)
            d[j] = data[j].coord.clone();
        double[][] transform = new double[dim][dim];
        double[][] retransform = new double[dim][dim];
        double[] shift = new double[dim];
        double centered[][] = ica.centerData(d, shift);
        double whitened[][] = ica.whitenData(centered, transform, retransform);
        displayData(whitened, "white space");
        double[][] w = ica.optimizeEntropy(whitened);
        Matrix mwhitened = new Matrix(whitened);
        Matrix mw = new Matrix(w);
        System.out.println("w: ");
        mw.print(3,3);
        double[][] projectedWhitened =( mwhitened.times(mw.transpose())).transpose().getArrayCopy();
        double[][] genGaussParams = new double[dim][] ;
        // now determine the shape parameter for every dimension
        double betaMin = Double.MAX_VALUE; //determine the most subgaussian dimension
        int minIndex = -1;
        for (int j=0 ; j<dim ; j++) {
            genGaussParams [j] = EpdMineo.optimizeMSP(projectedWhitened [j]) ;  //beta: genGaussParams[2]
            System.out.println("beta: " + genGaussParams[j][2]);
            if(genGaussParams[j][2] < betaMin){
                betaMin = genGaussParams[j][2];
                minIndex = j;
            }
        }
        double[] projCoord = new double[numObj];
        double[] wRow = new double[dim];
        //select row number minIndex of w
        for(int i = 0; i < dim; i++)
            wRow[i] = w[minIndex][i];
        //transpose wRow
        Matrix wt = rowVector(wRow).transpose();
        //compute projCoord = data * transform * wt
        for(int i = 0;  i < numObj; i++){
            projCoord[i] = rowVector(data[i].coord).times(new Matrix(transform)).times(wt).get(0,0);
            //System.out.println(projCoord[i]);
        }
        //50% split: sort projCoord
        SortPack[] s = SortPack.array(projCoord);
        int m = s.length/2;
        //set clusterIDs: index < m: ID 0, index >=m: ID1
        for(int i = 0; i < data.length; i++)
            if(i < m)
                data[s[i].index].clusterID = 1;
            else
                data[s[i].index].clusterID = 2; //clusterIDs 1 and 2 as in KMeans
        //split into 2 clusters
        res[0] = new IcaCluster();
        res[0].ID = 0;
        res[0].numObj = s.length/2;
        res[1] = new IcaCluster();
        res[1].ID = 1;
        res[1].numObj = s.length - s.length/2;
        //------------------START: assign points with epd-bisecting k-means, Problem: Kosten Double.MAX_VALUE, cluster 0 frisst im ersten Schritt alle Punkte weg---------------------------------------------------------
//        boolean terminated = false;
//        int iter = 0;
//        while(!terminated && iter < 100){
//            //determine genGaussParams for each cluster
//            double[][] genGaussParams1D = new double[k][];
//            for(int i = 0; i < k; i++){
//                double[] coordK = new double[res[i].numObj];
//                int counter = 0;
//                for(int j = 0; j < numObj; j++){
//                    if (data[j].clusterID == res[i].ID){
//                        coordK[counter] = projCoord[j];
//                        counter++;
//                    }
//                }
//                genGaussParams1D[i] = EpdMineo.optimizeMSP(coordK);
//            }
//            //print parameters
//            for(int i = 0; i < res.length; i++)
//                System.out.println("cluster: " + i + " mean: " + genGaussParams1D[i][0] + " sigma: " + genGaussParams[i][1] + " beta: " + genGaussParams[i][2]);
//            GenGauss g = new GenGauss();
//            boolean clusterChanged = false;
//            //reset number of objects
//            for(int i = 0; i < k; i++)
//                res[i].numObj = 0;
//            for(int i = 0; i < numObj; i++){
//                double minCost = Double.MAX_VALUE;
//                int min= -1;
//                //assign points
//                for(int j = 0; j < k; j++){
//                    double costDist = g.costV(projCoord[i], genGaussParams[j]);
//                    //System.out.println("costDist: "  + costDist);
//                    if(Double.isInfinite(costDist))
//                        System.err.println("g.costV("+ projCoord[i] + ",  genGaussParams[j]: " + genGaussParams[j][0] + " " + genGaussParams[j][1] + " " + genGaussParams[j][2] + " Double.MAX_VALUE");
//                    if(costDist < minCost){
//                        minCost = costDist;
//                        min = j;
//                    }
//                }
//                res[min].numObj++;
//                if(data[i].clusterID != min){
//                    data[i].clusterID = min;
//                    clusterChanged = true;
//                }
//            }
//            //TEST:  check clusterID
//            int count0 = 0;
//            int count1 = 0;
//            for(int i = 0; i < numObj; i++){
//                if(data[i].clusterID == 0)
//                    count0++;
//                if(data[i].clusterID == 1)
//                    count1++;
//            }
//            System.out.println("Cluster 0: " + res[0].numObj + " " + count0);
//            System.out.println("Cluster 1: " + res[1].numObj + " " + count1);
//            //TEST
//            if(! clusterChanged)
//                terminated = true;
//            System.out.println("iter: " + iter);
//            iter++;
//        }
        ///------------------END: epd bisecting k-means-------------------------------------------
        //-------------------START: assign points with bisecting k-means -------------------------
        boolean terminated = false;
        int iter = 0;
        double[] means = new double[dim];
        //init with the farthest objects
        means[0] = s[0].value;
        means[1] = s[s.length-1].value;
        while(!terminated){
            boolean clusterChanged = false;
            //reset number of objects
            for(int i = 0; i < k; i++)
                res[i].numObj = 0;
            //assign points
            for(int i = 0; i < numObj; i++){
                double minDist = Double.MAX_VALUE;
                int min= -1;
                for(int j = 0; j < k; j++){
                    double dist = Math.sqrt((projCoord[i] - means[j]) * (projCoord[i] - means[j]));
                    if(dist < minDist){
                        minDist = dist;
                        min = j;
                    }
                }
                res[min].numObj++;
                if(data[i].clusterID != (min+1)){
                    data[i].clusterID = min+1;
                    clusterChanged = true;
                }
            }
            //update means
            for(int i = 0; i < numObj; i++){
                if(data[i].clusterID == 1)
                    means[0] += projCoord[i];
                if(data[i].clusterID == 2)
                    means[1] += projCoord[i];
            }
            means[0]/=(double)res[0].numObj;
            means[1]/=(double)res[1].numObj;
            if(! clusterChanged)
                terminated = true;
            System.out.println("iter: " + iter);
            iter++;
        }
        //---END: bisecting k-means---------------------------------------------------------------------
        //set cluster1 and cluster2
        int count1 = 0;
        int count2 = 0;
        for(int i = 0; i < numObj; i++){
            if(data[i].clusterID == 1)
                count1++;
            if(data[i].clusterID == 2)
                count2++;
        }
        cluster1 = new DataObject[count1];
        count1 = 0;
        for(int i = 0; i < numObj; i++){
            if(data[i].clusterID == 1){
                cluster1[count1] = data[i];
                count1++;
            }
        }
        cluster2 = new DataObject[count2];
        count2 = 0;
        for(int i = 0; i < numObj; i++){
            if(data[i].clusterID == 2){
                cluster2[count2] = data[i];
                count2++;
            }
        }
        return res;
        
    }
    
    
    private void displayData(double[][] d, String title){
        DataObject[] data = new DataObject[d.length];
        for(int i = 0; i < d.length; i++){
            double[] coord = new double[dim];
            for(int j = 0; j < dim; j++)
                coord[j] = d[i][j];
            data[i] = new DataObject(coord, i);
        }
        VisuInfo visu = new VisuInfo(data, title);
        visu.setSize(600,600);
        visu.setLocation(100,0);
        visu.setVisible(true);
    }
    
    
    
    
    
    
}
