

/*
 * TopDownKM.java
 *
 * Created on 18. April 2006, 15:19
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package clusterer;
import clusterer.*;
import java.util.*;
import Jama.*;
import eval.Clustering;
import eval.SortPack;
//import eval.Cluster;
import eval.ClusteringEvaluation;
import compr.DataObject;
import compr.DB;
import filterCluster.FilterKM;
import ica.IcaCost;
import visu.VisuSimple;
import visu.VisuInfo;
import visu.VisuKMeans;
import visu.Visu2D;
import io.IO;
import gengauss.GenGauss;

/**
 * Class to support both IcaSplit and bisecting K-Means (PCASplit). Modified methods (in comparison to class FastTopDownIcaSplitKMeans: symmetricSplit, splitCluster, splitClusterPCA, splitClusterICA. New global variables to save clusterIDs and clusters of both variants during splitting: bestClustersICA, bestClustersPCA, idICA, idPCA.
 * @author claudia
 */
public class TopDownICAPCASplit{
    DataObject[] data;
    Vector<IcaCluster> clusters; //current clusters
    Vector<IcaCluster> noiseClusters;
    IcaCluster[] bestClusters; //cheapest clusters
    IcaCluster[] bestNoiseClusters;
    int[] aktClusterID; //current cluster IDs;
    IcaCluster[] bestClustersICA; //current best clusters obtained with Ica
    IcaCluster[] bestNoiseClustersICA;
    IcaCluster[] bestClustersPCA; //current best clusters obtained with KM
    IcaCluster[] bestNoiseClustersPCA;
    int[] idsICA; //current best clusterIDs bisecting K-Means
    int[] idsPCA; //current best clusterIDs IcaSplit
    DB db;
    double bestCost;
    int iter;
    int numObj;
    int dim;
    int maxClusterID;      //Max ID of the clusters (not noise!)
    int numTry = 0;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    static int minNumObj = 10; //minimum number of objects in a cluster for ica
    //static double paramCost =  2000;
    //static double paramCost =  3168;
    
    //static double paramCost =  4 * 32 * 2 + 4 * 4 * 32;
    
    
    /** Creates a new instance of TopDownKM */
    public TopDownICAPCASplit(DataObject[] data) {
        this.data = data;
        db = new DB(data);
        numObj = data.length;
        iter = 0;
        dim = data[0].coord.length;
        maxClusterID = 0;
        bestCost = 0.0;
        clusters = new Vector<IcaCluster>();
        noiseClusters = new Vector<IcaCluster>();
        idsPCA = new int[numObj];
        idsICA = new int[numObj];
    }
    
    /**
     * Filter clusters after splitting with bisecting K-Means.
     * COSTS: no parameter costs added here, this is done for the noise objects in the merge method.
     */
    private void finalFiltering(){
        System.out.println("--- finalFiltering ---");
        Vector<IcaCluster> filteredClusters = new Vector<IcaCluster>();
        Vector<IcaCluster> filteredNoise = new Vector<IcaCluster>();
        for(int i = 0; i < bestClusters.length; i++){
            System.out.println("filtering cluster " + bestClusters[i].ID + " numObj " + bestClusters[i].numObj);
            DataObject[] o = getObjects(bestClusters[i]);
            System.out.println("o.length " + o.length);
            IcaCluster[] result = getBestFilteredClusters(o, true, true);
            result[0].ID = bestClusters[i].ID;
            result[1].ID = bestClusters[i].ID +1;
            //adjust clusterIDs
            for(int j = 0; j < aktClusterID.length; j++){
                if(aktClusterID[j] == 1)
                    data[o[j].number].clusterID = result[1].ID;
                if(aktClusterID[j] == 0)
                    data[o[j].number].clusterID = result[0].ID;
            }
            filteredClusters.add(result[0]);
            filteredNoise.add(result[1]);
        }
        IcaCluster[] filteredC = new IcaCluster[filteredClusters.size()];
        for(int i = 0; i < filteredC.length; i++)
            filteredC[i] = (IcaCluster)filteredClusters.elementAt(i);
        bestClusters = filteredC;
        IcaCluster[] filteredN = new IcaCluster[filteredNoise.size()];
        for(int i = 0; i < filteredN.length; i++)
            filteredN[i] = (IcaCluster)filteredNoise.elementAt(i);
        bestNoiseClusters = filteredN;
    }
    
    /**
     * MODIFIED: returns filtered or unfiltered clusters. To enable no filtering during splitting.
     * COSTS: Parameter costs for the cluster are added during splitting. For final filtering parameter costs for the cluster are added. Parameter costs for noise clusters are added in merge method.
     * @param o
     * @param filter if noise should be filtered
     * @param verbose
     * @return Cluster[2]: result[0]: filtered cluster, result[1]: filtered noise, can be empty (numObj = 0).
     */
    private IcaCluster[] getBestFilteredClusters(DataObject[] o, boolean filter, boolean verbose){
        if(verbose)
            System.out.println("---getBestFilteredClusters, filter: " + filter);
        if (filter){
            if(o.length < minNumObj){
                if(verbose)
                    System.out.println("Cluster too small for filtering");
                IcaCluster[] result = new IcaCluster[2];
                IcaCost ic = new IcaCost(dim);
                result[0] = ic.mergedClusterNoRot(o);
                result[1] = new IcaCluster(1);
                //clusterID
                int[] ids = new int[o.length];
                for(int i = 0; i < ids.length; i++)
                    ids[i] = 0;
                aktClusterID = ids;
                return result;
            }
            IcaCost ic = new IcaCost(o);
            IcaCluster[] result = ic.selectBestFilteredCluster();
//            //add parameter costs
//            double paramCostCluster = paramCost; //add parameter costs for cluster
//            double paramCostNoise = 0.0;
//            result[0].cost += paramCostCluster;
//            result[1].cost += paramCostNoise;
//            if(verbose){
//                System.out.println("coding cost cluster: " + result[0].cost + " including " + paramCostCluster + " parameter costs.");
//                System.out.println("coding cost noise: " + result[1].cost + " including " + paramCostNoise + " parameter costs.");
//            }
            int[] ids = ic.getClusterID();
            aktClusterID = ids;
            return result;
        }else{ //all objects in cluster
            if(o.length < minNumObj){
                if(verbose)
                    System.out.println("Cluster too small for ica");
                IcaCluster[] result = new IcaCluster[2];
                IcaCost ic = new IcaCost(dim);
                result[0] = ic.mergedClusterNoRot(o);
                result[1] = new IcaCluster(1);
                //clusterID
                int[] ids = new int[o.length];
                for(int i = 0; i < ids.length; i++)
                    ids[i] = 0;
                aktClusterID = ids;
                aktClusterID = ids;
                return result;
            }
            
            int[] ids = new int[o.length];
            for(int i = 0; i < ids.length; i++)
                ids[i] = 0;
            aktClusterID = ids;
            IcaCost ic = new IcaCost(dim);
            IcaCluster[] result = new IcaCluster[2];
            result[0] = ic.selectBestCluster(o);
            result[0].ID = 0;
            //result[0].cost += paramCost;
            result[0].numObj = o.length;
            result[1] = new IcaCluster(1);
            result[1].numObj = 0;
            return result;
        }
    }
    
    public void run(){
        double splitMinCost = Double.MAX_VALUE; //global cost minimum during splitting
        IcaCluster[] splitMinClusters = new IcaCluster[1]; //clusters with minimum cost achieved only by splitting without filtering
        IcaCluster[] result = getBestFilteredClusters(data, false, false);
        int[] splitIds = new int[data.length];  //save clusterIDs after split
        for(int i = 0; i < data.length; i++)
            data[i].clusterID = aktClusterID[i];
        bestClusters = new IcaCluster[1];
        bestClusters[0] = result[0];
        bestNoiseClusters = new IcaCluster[1];
        bestNoiseClusters[0] = result[1];
        double costFiltered = currentCost();
        splitMinCost = costFiltered;
        double minCost = costFiltered;
        System.out.println("filtered data set " + costFiltered);
        boolean improvement = true;
        int aktTry = 0;
        IcaCluster[] lastLocalMinClusters = new IcaCluster[1];        //save Clusters and ClusterIDs of the last local minimum
        IcaCluster[] lastLocalMinNoise = new IcaCluster[1];
        int[] lastLocalMinIDs = new int[numObj];
        IcaCluster[] minClusters = new IcaCluster[1];        //save Clusters and ClusterIDs of the global cost minimum
        IcaCluster[] minNoise = new IcaCluster[1];
        int[] minIDs = new int[numObj];
        //first global minimum: filtered data set
        minClusters = bestClusters;
        minNoise = bestNoiseClusters;
        for(int i = 0; i < data.length; i++)
            minIDs[i] = data[i].clusterID;
        while(improvement || aktTry < numTry){
            //while(improvement && aktTry < numTry){
            System.out.println("-------------------------------------");
            double costUnsplit = currentCost();
            int numClUnsplit = numClustersNumNoiseClusters()[0];
            int numNUnsplit = numClustersNumNoiseClusters()[1];
            System.out.println("SYMMETRIC SPLIT NUMBER: " + iter + " CURRENT COST " + costUnsplit);
            improvement = symmetricSplit(aktTry);
            if(improvement){
                minCost = currentCost();
                //new global cost minimum found
                if(minCost < splitMinCost){
                    splitMinCost = minCost;
                    splitMinClusters = bestClusters;
                    for(int i = 0; i < splitIds.length; i++)
                        splitIds[i] = data[i].clusterID;
                    minClusters = bestClusters;
                    minNoise = bestNoiseClusters;
                    for(int i = 0; i < numObj; i++)
                        minIDs[i] = data[i].clusterID;
                    System.out.println("GLOBAL IMPROVEMENT: " + " CURRENT COST " + minCost + " COST OLD: " + costUnsplit);
                } else
                    System.out.println("IMPROVEMENT FOR AT LEAST ONE CLUSTER: " + minCost + " COST OLD: " + costUnsplit);
                lastLocalMinClusters = bestClusters;
                lastLocalMinNoise = bestNoiseClusters;
                for(int i = 0; i < numObj; i++)
                    lastLocalMinIDs[i] = data[i].clusterID;
                aktTry = 0;
            } else{
                aktTry++;
                System.out.println("NO GLOBAL IMPROVEMENT: CURRENT COST " + currentCost()/*currentCostWithoutIDCosts()*/ + " COST OLD: " + costUnsplit + " TRYING FOR ANOTHER " + (numTry - aktTry) + " TIMES");
                IO ea = new IO();
//                ea.printResult(data);
//                System.out.println("m");
            }
            System.out.println("OLD: " + numClUnsplit + " clusters, " + numNUnsplit + " noiseClusters.");
            System.out.println("NEW: " + numClustersNumNoiseClusters()[0] + " clusters " + numClustersNumNoiseClusters()[1] + " noiseClusters.");
            System.out.println("-------------------------------------");
            iter++;
        }
        System.out.println("done. " + iter + " iterations");
        
        //adjust clusters and IDs as from the last local minimum
        if(minClusters.length > 1 || minNoise.length > 1){
            bestClusters = minClusters;
            bestNoiseClusters = minNoise;
            for(int i = 0; i < data.length; i++)
                data[i].clusterID = minIDs[i];
            System.out.println("BEST: " + bestClusters.length + " clusters ");
        }
        //checkClusterIDs();
        finalFiltering();
        resetClusterIDs();
        //boolean impr = finalSplit();
        //mergeClusters();
        //simpleGreedyMerge(0);
        greedyMerge(0, true);
        minCost = currentCost();
        System.out.println("BEST COST: " + minCost);
        if(splitMinCost < minCost){
            bestClusters = splitMinClusters;
            for(int i = 0; i < data.length; i++)
                data[i].clusterID = splitIds[i];
            System.out.println("NO COST IMPROVEMENT DUE TO FILTERING AND MERGING.");
            //reset clusterIDs
        }
        
        //printInfo();
        GenGauss g = new GenGauss();
        //g.selectivityEstimation(data, bestClusters, 10);
        //printPercentage();
        //evaluatePDF(1110); //cluster 0
        //evaluatePDF(1003); //gut im Vergleich zu EM auf String dir = "3cluster2dnoiseErzeugt.txt";
        //print data of specified clusters
        IO ea = new IO();
        //ea.printClusters(data, 4);
        //ea.printClusters(data, 2);
        //ea.printClusters(data, 0);
        //ea.printClusters(data, 10);
        
        //print clustering result
        //ea.printResult(data);
        
        VisuInfo vi = new VisuInfo(data, "result");
        vi.setSize(600,600);
        vi.setLocation(200,0);
        vi.setVisible(true);
    }
    
    public void printInfo(){
        System.out.println("---------PRINT INFO-------");
        for (int i = 0; i < bestClusters.length; i++){
            System.out.println("----------Cluster " + bestClusters[i].ID + "----------");
            System.out.println("number of objects: " + bestClusters[i].numObj);
            if(bestClusters[i].numObj > minNumObj && bestClusters[i].realRotation != null){
                System.out.println("realRotation:");
                for(int j = 0; j < bestClusters[i].realRotation.length; j++)
                    for(int k = 0; k < bestClusters[i].realRotation[j].length; k++)
                        System.out.print(bestClusters[i].realRotation[j][k] + "\t");
                System.out.println();
                System.out.print("realCenter: ");
                if(bestClusters[i].realCenter != null){
                    for(int j = 0; j < bestClusters[i].realCenter.length; j++)
                        System.out.print(bestClusters[i].realCenter[j] + "\t");
                    System.out.println();
                }
                System.out.println("beta: ");
                if(bestClusters[i].realCenter != null){
                    for(int j = 0; j < bestClusters[i].beta.length; j++)
                        System.out.print(bestClusters[i].beta[j] + "\t");
                    System.out.println();
                }
            }
        }
        
    }
    
    private int[] numClustersNumNoiseClusters(){
        int numCl = 0;
        int numN = 0;
        for(int i = 0; i < bestClusters.length; i++)
            if(bestClusters[i].numObj > 0)
                numCl++;
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                numN++;
        int[] result = new int[2];
        result[0] = numCl;
        result[1] = numN;
        return result;
    }
    
    /**
     * Evaluate how well the clustering result fits to the discovered pdfs. Determine the likelihood of the cluster points and how many points actually are in this location.
     */
    public void evaluatePDF(int rp){
        double windowSize = 2.5;
        int steps = 1000000 ;
        double[] lb = new double[dim];
        double ub[] = new double[dim];
        double[] overallLb = new double[dim]; //for KL-divergence
        double[] overallUb = new double[dim];
        for(int i = 0; i < dim; i++){
            overallLb[i] = Double.MAX_VALUE;
            overallUb[i] = -Double.MAX_VALUE;
        }
        GenGauss gg = new GenGauss();
        for(int i = 0; i < dim; i++){
            lb[i] = data[rp].coord[i]-windowSize;
            ub[i] = data[rp].coord[i]+windowSize;
        }
        //count Objects in window
        int counter = 0;
        for(int i = 0; i < data.length; i++){
            boolean contained = true;
            for(int j = 0; j < dim; j++){
                if(data[i].coord[j] > ub[j] || data[i].coord[j] < lb[j])
                    contained = false;
                if(data[i].coord[j] < overallLb[j])
                    overallLb[j] = data[i].coord[j];
                if(data[i].coord[j] > overallUb[j])
                    overallUb[j] = data[i].coord[j];
            }
            if(contained){
                counter++;
                data[i].clusterID = 100;
            }
        }
        overallLb[0] = 16.0;
        overallLb[1] = 6.0;
        overallUb[0] = 66.0;
        overallUb[1] = 47.0;
        
        double prob = 0.0;
        for(int i = 0; i < bestClusters.length; i++){
            System.out.println(bestClusters[i].ID + " " + gg.integral(lb, ub, bestClusters[i].realCenter, new Matrix(bestClusters[i].realRotation).getArrayCopy(), bestClusters[i].beta, steps));
            prob += (double)bestClusters[i].numObj/(double)numObj * gg.integral(lb, ub, bestClusters[i].realCenter, new Matrix(bestClusters[i].realRotation).getArrayCopy(), bestClusters[i].beta, steps);
        }
        System.out.println("prob: " + prob + " " + counter + " objects found. Percentage:  " + (double)counter/(double)numObj);
        System.out.println("EM: "+gg.gaussEvaluation(lb,ub,steps)) ;
        System.out.println("KL-divergence: " + gg.klEvaluation(bestClusters, overallLb, overallUb, steps, numObj));
    }
    
    
    
    /**
     * MODIFIED. Returns true if either PCASplit or IcaSplit is successful. Sets bestClusters and clusterIDs to the current best solution.
     * @param aktTry number of tried splits.
     * @return boolean: improvement of cost.
     */
    private boolean symmetricSplit(int aktTry){
        //aktTry = numTry;
        IcaCluster[] currentClusters = bestClusters.clone();
        double[] improvement_pca = new double[currentClusters.length]; //save improvement of costs with bisecting K-Means
        double[] improvement_ica = new double[currentClusters.length];
        boolean[] improvement = new boolean[currentClusters.length];
        
        for(int i = 0; i < currentClusters.length; i++){
            int[] aktID = new int[data.length]; //save IDs if split is not successful
            for(int j = 0; j < data.length; j++)
                aktID[j] = data[j].clusterID;
            //reset idsPCA and idsICA
            for(int j = 0; j < data.length; j++){
                idsPCA[j] = data[j].clusterID;
                idsICA[j] = data[j].clusterID;
            }
            try{
                improvement_pca[i] = splitClusterPCA(i);
                maxClusterID = maxClusterID-2;
                improvement_ica[i] = splitClusterICA(i);
                maxClusterID = maxClusterID-2;
                System.out.println("maxClusterID after ICA split " + maxClusterID);
                if(improvement_pca[i] > 0.0  || improvement_ica[i] > 0.0){ //improvement due to PCA or ICASplit
                    //set clusterIDs and bestClusters to the best solution, i.e. to the maximum improvement.
                    improvement[i] = true;
                    maxClusterID = maxClusterID+2;
                    if(improvement_pca[i] > improvement_ica[i]){
                        System.out.println("KM SPLIT");
                        bestCost= bestCost-improvement_pca[i];
                        bestClusters = bestClustersPCA;
                        bestNoiseClusters = bestNoiseClustersPCA;
                        for(int j = 0; j < data.length; j++)
                            data[j].clusterID = idsPCA[j];
                        //TEST
                        checkClusterIDs();
                        //TEST
                    } else{
                        System.out.println("ICA SPLIT");
                        bestCost= bestCost-improvement_ica[i];
                        bestClusters = bestClustersICA;
                        bestNoiseClusters = bestNoiseClustersICA;
                        for(int j = 0; j < data.length; j++)
                            data[j].clusterID = idsICA[j];
                        //TEST
                        //checkClusterIDs();
                        //TEST
                    }
                } else{
                    System.out.println("SPLIT UNSUCCESSFUL.");
                    for(int j = 0; j < data.length; j++)
                        data[j].clusterID = aktID[j];
                    //TEST
                    checkClusterIDs();
                    //TEST
                }//else
                if(improvement_pca[i] == Double.NEGATIVE_INFINITY && improvement_ica[i] == Double.NEGATIVE_INFINITY){ //cluster too small for splitting or with both PcaSplit and IcaSplit one of the resulting cluster becomes empty.
                    improvement[i] = false;
                    for(int j = 0; j < data.length; j++)
                        data[j].clusterID = aktID[j];
                }
                
            } catch(java.lang.NullPointerException e){
                improvement[i] = false;
                for(int j = 0; j < data.length; j++)
                    data[j].clusterID = aktID[j];
            } catch(java.lang.ArrayIndexOutOfBoundsException e){
                System.err.println("mist");
                improvement[i] = false;
                for(int j = 0; j < data.length; j++)
                    data[j].clusterID = aktID[j];
            }
        }
        boolean impr = false;
        for(int i = 0; i < improvement.length; i++)
            if(improvement[i])
                impr = true;
        if(!impr){
            if(aktTry < numTry){
                System.out.println("FORCED SPLIT");
                for(int i = 0; i < currentClusters.length; i++){
                    int[] aktID = new int[data.length]; //save IDs if split is not successful
                    improvement[i] = true;
                    try{
                        improvement_pca[i] = splitClusterPCA(i);
                        maxClusterID = maxClusterID-2;
                        improvement_ica[i] = splitClusterICA(i);
                        maxClusterID = maxClusterID-2;
                        //set bestClusters and clusterIDs to the solution with minimum cost gain
                        maxClusterID = maxClusterID+2;
                        if(improvement_pca[i] > improvement_ica[i]){
                            System.out.println("KM SPLIT");
                            bestCost= bestCost-improvement_pca[i];
                            bestClusters = bestClustersPCA;
                            bestNoiseClusters = bestNoiseClustersPCA;
                            for(int j = 0; j < data.length; j++)
                                data[j].clusterID = idsPCA[j];
                        } else{
                            System.out.println("ICA SPLIT");
                            bestCost= bestCost-improvement_ica[i];
                            bestClusters = bestClustersICA;
                            bestNoiseClusters = bestNoiseClustersICA;
                            for(int j = 0; j < data.length; j++)
                                data[j].clusterID = idsICA[j];
                        }
                        if(improvement_pca[i] == Double.NEGATIVE_INFINITY && improvement_ica[i] == Double.NEGATIVE_INFINITY){ //cluster too small for splitting or with both PcaSplit and IcaSplit one of the resulting cluster becomes empty.
                            improvement[i] = false;
                            for(int j = 0; j < data.length; j++)
                                data[j].clusterID = aktID[j];
                        }
                    } catch(java.lang.NullPointerException e){
                        improvement[i] = false;
                        for(int j = 0; j < data.length; j++)
                            data[j].clusterID = aktID[j];
                    } catch(java.lang.ArrayIndexOutOfBoundsException e){
                        improvement[i] = false;
                        for(int j = 0; j < data.length; j++)
                            data[j].clusterID = aktID[j];
                    }
                }
            }
        }
        return impr;
    }
    
    
    /**
     * written because of bug: wrong number of objects after filtering in the case of dublicates.
     */
    private void resetClusterIDs(){
        System.out.println("---resetClusterIDs-----");
        for(int i = 0; i < bestClusters.length; i++){
            int count = 0;
            for(int j = 0; j < data.length; j++)
                if(data[j].clusterID == bestClusters[i].ID)
                    count++;
            bestClusters[i].numObj = count;
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            int count = 0;
            for(int j = 0; j < data.length; j++)
                if(data[j].clusterID == bestNoiseClusters[i].ID)
                    count++;
            bestNoiseClusters[i].numObj = count;
        }
        
    }
    
    
    
    /**
     * MODIFIED: method determines the cost of KMSplit. Writes global variables idsPCA and bestClustersPCA.
     * @return double: cost change: > 0 -> improvement of costs, < 0 -> cost decline. Double.NEGATIVEINFINITY if cluster is too small for splitting.
     * @param splitIndex index of cluster to be split
     */
    private double splitClusterPCA(int splitIndex){
        double impr = 0.0;
        //System.out.println("split cluster " + bestClusters[splitIndex].ID);
        if(bestClusters[splitIndex].numObj < minNumObj){ //cluster is too small for splitting
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
        int[] currentID1 = new int[numObj]; //clusterIDs for cluster and noise cluster if filtered
        int[] currentID2;
        boolean improvement = false;
        boolean filtered1 = false; //k-means cluster 1 filtered?
        boolean filtered2 = false;
        //int splitIndex = chooseSplitCluster();
        //get all objects of the cluster to split
        double costOld = bestClusters[splitIndex].cost + bestNoiseClusters[splitIndex].cost;
        DataObject[] cobj = db.getObjects(bestClusters[splitIndex].ID); // hier obj kopieren und ClusterID auf 0 setzen!
        DataObject[] nobj = db.getObjects(bestNoiseClusters[splitIndex].ID);
        int numAll = cobj.length+nobj.length;
        DataObject[] all = new DataObject[numAll];
        for(int i = 0; i < numAll; i++)
            if(i < cobj.length)
                all[i] = cobj[i];
            else
                all[i] = nobj[i-cobj.length];
        //        START: K-MEANS
        DB kmDb = new DB(all);
        KMeans k = new KMeans(2, kmDb);
        DataObject[] newIDs = k.run(); //ClusterIDs 1 and 2
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = kmDb.getObjects(1);
        if(id_1.length < minNumObj){ //MODIFIED: NO CLUSTERS < minNumObj
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
//         END: K-MEANS
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for(int i = 0; i < id_1.length; i++){
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap h_1 = new HashMap(id_1.length);
        for(int i = 0; i < id_1.length; i++)
            h_1.put(internalIndex_1[i], numbers_1[i]);
        for(int i = 0; i < id_1.length; i++)
            id_1[i].number = internalIndex_1[i];
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        IcaCluster[] clusters1 = getBestFilteredClusters(id_1, false, false);
        currentID1 = aktClusterID;
        clusters1[0].ID = bestClusters[splitIndex].ID;
        clusters1[1].ID = bestNoiseClusters[splitIndex].ID;
        
////objects assigned to clusterID 2 by kMeans
        //K-MEANS
        DataObject[] id_2 = kmDb.getObjects(2);
        if(id_2.length < minNumObj){
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
        //K-MEANS
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for(int i = 0; i < id_2.length; i++){
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap h_2 = new HashMap(id_2.length);
        for(int i = 0; i < id_2.length; i++)
            h_2.put(internalIndex_2[i], numbers_2[i]);
        for(int i = 0; i < id_2.length; i++)
            id_2[i].number = internalIndex_2[i];
        IcaCluster[] clusters2 = getBestFilteredClusters(id_2, false, false);
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;
        clusters2[1].ID = maxClusterID + 3;
        //update current clusters
        for(int i = 0; i < bestClusters.length; i++)
            if(i != splitIndex)
                clusters.add(bestClusters[i]);
            else
                clusters.add(clusters1[0]);
        clusters.add(clusters2[0]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(i != splitIndex)
                noiseClusters.add(bestNoiseClusters[i]);
            else
                noiseClusters.add(clusters1[1]);
        noiseClusters.add(clusters2[1]);
        //determine saved cost
        impr = savedCost(false);
        //write global variables bestClustersPCA and bestNoiseClustersPCA
        //bestClustersPCA, bestNoiseClustersPCA
        IcaCluster[] newClusters = new IcaCluster[clusters.size()];
        IcaCluster[] newNoiseClusters = new IcaCluster[clusters.size()];
        for(int i = 0; i < clusters.size(); i++){
            newClusters[i] = (IcaCluster)clusters.elementAt(i);
            int aktID = newClusters[i].ID;
            for(int j = 0; j < noiseClusters.size(); j++){
                if(((IcaCluster)noiseClusters.elementAt(j)).ID == aktID+1)
                    newNoiseClusters[i] = (IcaCluster)noiseClusters.elementAt(j);
            }
            if(newNoiseClusters[i] == null){
                newNoiseClusters[i] = new IcaCluster();
                newNoiseClusters[i].ID = aktID+1;
            }
        }
        for(int i = 0; i < noiseClusters.size(); i++)
            newNoiseClusters[i] = (IcaCluster)noiseClusters.elementAt(i);
        bestClustersPCA = newClusters;
        bestNoiseClustersPCA = newNoiseClusters;
        //clusterIDS
        maxClusterID = maxClusterID+2;
        int splittedID = bestClusters[splitIndex].ID;
        
        for(int i = 0; i < currentID1.length; i++){
            int index = ((Integer)h_1.get(i)).intValue();
            if(currentID1[i] == 0)
                idsPCA[index] = splittedID;
            else
                idsPCA[index] = splittedID+1;
        }
        for(int i = 0; i < currentID2.length; i++){
            int index = ((Integer)h_2.get(i)).intValue();
            if(currentID2[i] == 0)
                idsPCA[index] = maxClusterID;
            else
                idsPCA[index] = maxClusterID + 1;
        }
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        return impr;
    }
    
    
    /**
     * MODIFIED: method determines the cost of ICASplit. Writes global variables idsICA and bestClustersICA.
     * @return double: cost change: > 0 -> improvement of costs, < 0 -> cost decline. Double.NEGATIVEINFINITY if cluster is too small for splitting.
     * @param splitIndex index of cluster to be split
     */
    private double splitClusterICA(int splitIndex){
        double impr = 0.0;
        //System.out.println("split cluster " + bestClusters[splitIndex].ID);
        if(bestClusters[splitIndex].numObj < minNumObj){ //cluster is too small for splitting
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
        int[] currentID1 = new int[numObj]; //clusterIDs for cluster and noise cluster if filtered
        int[] currentID2;
        boolean improvement = false;
        boolean filtered1 = false; //k-means cluster 1 filtered?
        boolean filtered2 = false;
        //int splitIndex = chooseSplitCluster();
        //get all objects of the cluster to split
        double costOld = bestClusters[splitIndex].cost + bestNoiseClusters[splitIndex].cost;
        DataObject[] cobj = db.getObjects(bestClusters[splitIndex].ID); // hier obj kopieren und ClusterID auf 0 setzen!
        DataObject[] nobj = db.getObjects(bestNoiseClusters[splitIndex].ID);
        int numAll = cobj.length+nobj.length;
        DataObject[] all = new DataObject[numAll];
        for(int i = 0; i < numAll; i++)
            if(i < cobj.length)
                all[i] = cobj[i];
            else
                all[i] = nobj[i-cobj.length];
        
        //MODIFIED: USE ICASPLIT INSTEAD OF ORDINARY K-MEANS
        //START: ICASPLIT
        IcaSplit is = new IcaSplit(2, all);
        DataObject[] newIDs = is.run(); //ClusterIDs 1 and 2
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = is.getObjects(1);
        if(id_1.length < minNumObj) {//MODIFIED: NO CLUSTERS < minNumObj
            maxClusterID = maxClusterID+2;
            return Double.NEGATIVE_INFINITY;
        }
////        //END: ICASPLIT
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for(int i = 0; i < id_1.length; i++){
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap h_1 = new HashMap(id_1.length);
        for(int i = 0; i < id_1.length; i++)
            h_1.put(internalIndex_1[i], numbers_1[i]);
        for(int i = 0; i < id_1.length; i++)
            id_1[i].number = internalIndex_1[i];
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        IcaCluster[] clusters1 = getBestFilteredClusters(id_1, false, false);
        currentID1 = aktClusterID;
        clusters1[0].ID = bestClusters[splitIndex].ID;
        clusters1[1].ID = bestNoiseClusters[splitIndex].ID;
        
////objects assigned to clusterID 2 by kMeans
//        //ICASPLIT
        DataObject[] id_2 = is.getObjects(2);
        if(id_2.length < minNumObj){
            maxClusterID = maxClusterID + 2;
            return Double.NEGATIVE_INFINITY;
        }
////        //ICASPLIT
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for(int i = 0; i < id_2.length; i++){
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap h_2 = new HashMap(id_2.length);
        for(int i = 0; i < id_2.length; i++)
            h_2.put(internalIndex_2[i], numbers_2[i]);
        for(int i = 0; i < id_2.length; i++)
            id_2[i].number = internalIndex_2[i];
        IcaCluster[] clusters2 = getBestFilteredClusters(id_2, false, false);
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;
        clusters2[1].ID = maxClusterID + 3;
        //update current clusters
        for(int i = 0; i < bestClusters.length; i++)
            if(i != splitIndex)
                clusters.add(bestClusters[i]);
            else
                clusters.add(clusters1[0]);
        clusters.add(clusters2[0]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(i != splitIndex)
                noiseClusters.add(bestNoiseClusters[i]);
            else
                noiseClusters.add(clusters1[1]);
        noiseClusters.add(clusters2[1]);
        //determine saved cost
        impr = savedCost(false);
        //write global variables bestClustersICA and bestNoiseClustersICA
        //bestClustersICA, bestNoiseClustersICA
        IcaCluster[] newClusters = new IcaCluster[clusters.size()];
        IcaCluster[] newNoiseClusters = new IcaCluster[clusters.size()];
        for(int i = 0; i < clusters.size(); i++){
            newClusters[i] = (IcaCluster)clusters.elementAt(i);
            int aktID = newClusters[i].ID;
            for(int j = 0; j < noiseClusters.size(); j++){
                if(((IcaCluster)noiseClusters.elementAt(j)).ID == aktID+1)
                    newNoiseClusters[i] = (IcaCluster)noiseClusters.elementAt(j);
            }
            if(newNoiseClusters[i] == null){
                newNoiseClusters[i] = new IcaCluster();
                newNoiseClusters[i].ID = aktID+1;
            }
        }
        for(int i = 0; i < noiseClusters.size(); i++)
            newNoiseClusters[i] = (IcaCluster)noiseClusters.elementAt(i);
        bestClustersICA = newClusters;
        bestNoiseClustersICA = newNoiseClusters;
        //clusterIDS
        maxClusterID = maxClusterID+2;
        int splittedID = bestClusters[splitIndex].ID;
        
        for(int i = 0; i < currentID1.length; i++){
            int index = ((Integer)h_1.get(i)).intValue();
            if(currentID1[i] == 0)
                idsICA[index] = splittedID;
            else
                idsICA[index] = splittedID+1;
        }
        for(int i = 0; i < currentID2.length; i++){
            int index = ((Integer)h_2.get(i)).intValue();
            if(currentID2[i] == 0)
                idsICA[index] = maxClusterID;
            else
                idsICA[index] = maxClusterID + 1;
        }
        
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        
        return impr;
    }
    
    
    
    /**
     * Determines the next cluster to split, splits the cluster and returns if improvment of cost results from this split.
     * @param aktTry aktTry: number of trys after minimum cost has been reached
     * @return boolean: improvement of cost
     */
    private boolean splitCluster(boolean splitAnyway, int splitIndex){
        //System.out.println("split cluster " + bestClusters[splitIndex].ID);
        if(bestClusters[splitIndex].numObj < minNumObj)
            return false;
        int[] currentID1 = new int[numObj]; //clusterIDs for cluster and noise cluster if filtered
        int[] currentID2;
        boolean improvement = false;
        boolean filtered1 = false; //k-means cluster 1 filtered?
        boolean filtered2 = false;
        //int splitIndex = chooseSplitCluster();
        //get all objects of the cluster to split
        double costOld = bestClusters[splitIndex].cost + bestNoiseClusters[splitIndex].cost;
        DataObject[] cobj = db.getObjects(bestClusters[splitIndex].ID); // hier obj kopieren und ClusterID auf 0 setzen!
        DataObject[] nobj = db.getObjects(bestNoiseClusters[splitIndex].ID);
        int numAll = cobj.length+nobj.length;
        DataObject[] all = new DataObject[numAll];
        for(int i = 0; i < numAll; i++)
            if(i < cobj.length)
                all[i] = cobj[i];
            else
                all[i] = nobj[i-cobj.length];
        
        //MODIFIED: USE ICASPLIT INSTEAD OF ORDINARY K-MEANS
        //START: ICASPLIT
        IcaSplit is = new IcaSplit(2, all);
        DataObject[] newIDs = is.run(); //ClusterIDs 1 and 2
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = is.getObjects(1);
        if(id_1.length < minNumObj) //MODIFIED: NO CLUSTERS < minNumObj
            return false;
////        //END: ICASPLIT
        
//        START: K-MEANS
//        DB kmDb = new DB(all);
//        KMeans k = new KMeans(2, kmDb);
//        DataObject[] newIDs = k.run(); //ClusterIDs 1 and 2
//        //objects assigned to cluster 1 by kMeans
//        DataObject[] id_1 = kmDb.getObjects(1);
//        if(id_1.length < minNumObj) //MODIFIED: NO CLUSTERS < minNumObj
        //           return false;
//         END: K-MEANS
        
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for(int i = 0; i < id_1.length; i++){
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap h_1 = new HashMap(id_1.length);
        for(int i = 0; i < id_1.length; i++)
            h_1.put(internalIndex_1[i], numbers_1[i]);
        for(int i = 0; i < id_1.length; i++)
            id_1[i].number = internalIndex_1[i];
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        IcaCluster[] clusters1 = getBestFilteredClusters(id_1, false, false);
        currentID1 = aktClusterID;
        clusters1[0].ID = bestClusters[splitIndex].ID;
        clusters1[1].ID = bestNoiseClusters[splitIndex].ID;
        
////objects assigned to clusterID 2 by kMeans
//        //ICASPLIT
        DataObject[] id_2 = is.getObjects(2);
        if(id_2.length < minNumObj)
            return false;
////        //ICASPLIT
        
        //K-MEANS
//        DataObject[] id_2 = kmDb.getObjects(2);
//        if(id_2.length < minNumObj)
//            return false;
        //K-MEANS
        
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for(int i = 0; i < id_2.length; i++){
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap h_2 = new HashMap(id_2.length);
        for(int i = 0; i < id_2.length; i++)
            h_2.put(internalIndex_2[i], numbers_2[i]);
        for(int i = 0; i < id_2.length; i++)
            id_2[i].number = internalIndex_2[i];
        IcaCluster[] clusters2 = getBestFilteredClusters(id_2, false, false);
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;
        clusters2[1].ID = maxClusterID + 3;
        //update current clusters
        for(int i = 0; i < bestClusters.length; i++)
            if(i != splitIndex)
                clusters.add(bestClusters[i]);
            else
                clusters.add(clusters1[0]);
        clusters.add(clusters2[0]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(i != splitIndex)
                noiseClusters.add(bestNoiseClusters[i]);
            else
                noiseClusters.add(clusters1[1]);
        noiseClusters.add(clusters2[1]);
        //boolean improvement = true;
        //boolean improvement = evaluateCost(splitIndex, cl, noise);
        if(splitAnyway)
            improvement = true;
        else
            improvement = improvement(true);
        if(improvement){
            //set new clusters and new noiseClusters as best clusters.
            IcaCluster[] newClusters = new IcaCluster[clusters.size()];
            IcaCluster[] newNoiseClusters = new IcaCluster[clusters.size()];
            for(int i = 0; i < clusters.size(); i++){
                newClusters[i] = (IcaCluster)clusters.elementAt(i);
                int aktID = newClusters[i].ID;
                for(int j = 0; j < noiseClusters.size(); j++){
                    if(((IcaCluster)noiseClusters.elementAt(j)).ID == aktID+1)
                        newNoiseClusters[i] = (IcaCluster)noiseClusters.elementAt(j);
                }
                if(newNoiseClusters[i] == null){
                    newNoiseClusters[i] = new IcaCluster();
                    newNoiseClusters[i].ID = aktID+1;
                }
            }
            for(int i = 0; i < noiseClusters.size(); i++)
                newNoiseClusters[i] = (IcaCluster)noiseClusters.elementAt(i);
            bestClusters = newClusters;
            bestNoiseClusters = newNoiseClusters;
            //reset bestCost
            bestCost = 0.0;
            for(int i = 0; i < bestClusters.length; i++)
                bestCost += bestClusters[i].cost;
            for(int i = 0; i < bestNoiseClusters.length; i++)
                bestCost += bestNoiseClusters[i].cost;
            
            maxClusterID = maxClusterID+2;
            int splittedID = bestClusters[splitIndex].ID;
            
            for(int i = 0; i < currentID1.length; i++){
                int index = ((Integer)h_1.get(i)).intValue();
                if(currentID1[i] == 0)
                    data[index].clusterID = splittedID;
                else
                    data[index].clusterID = splittedID+1;
            }
            for(int i = 0; i < currentID2.length; i++){
                int index = ((Integer)h_2.get(i)).intValue();
                if(currentID2[i] == 0)
                    data[index].clusterID = maxClusterID;
                else
                    data[index].clusterID = maxClusterID + 1;
            }
        }
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        if(improvement)
            return true;
        else{
            return false;
        }
    }
    
    /**
     * Try once to split all clusters and all noise clusters individually. Called after top-down has reached first cost minimum and before merging.
     * @return boolean: improvement for at least one cluster or noiseCluster
     */
    private boolean finalSplit(){
        boolean impr = false;
        Vector<IcaCluster> nonEmpty = new Vector<IcaCluster>();
        for(int i = 0; i < bestClusters.length; i++)
            if(bestClusters[i].numObj > 0)
                nonEmpty.add(bestClusters[i]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                nonEmpty.add(bestNoiseClusters[i]);
        boolean[] improvement = new boolean[nonEmpty.size()];
        IcaCluster[] actClusters = new IcaCluster[nonEmpty.size()];
        for(int i = 0; i < actClusters.length; i++)
            actClusters[i] = (IcaCluster)nonEmpty.elementAt(i);
        //relable noiseClusters and generate dummy noiseClusters to use the method splitCluster()
        int clusterCount = 0;
        for(int i = 0; i < actClusters.length; i++){
            int oldClusterID = actClusters[i].ID;
            if(oldClusterID != clusterCount){
                actClusters[i].ID = clusterCount;
                adjustClusterIDs(oldClusterID, clusterCount);
            }
            clusterCount = clusterCount+2;
        }
        maxClusterID = clusterCount;
        /*
        VisuSimple vk = new VisuSimple(data);
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
         */
        
        IcaCluster[] actNoise = new IcaCluster[actClusters.length];
        for(int i = 0; i < actNoise.length; i++)
            actNoise[i] = new IcaCluster(actClusters[i].ID + 1);
        bestClusters = actClusters;
        bestNoiseClusters = actNoise;
        for(int i = 0; i < improvement.length; i++){
            improvement[i] = splitCluster(true, i);
            if(improvement[i])
                impr = true;
        }
        if(impr){
            System.out.println("IMPROVEMENT DUE TO FINAL SPLIT. NEW: " + bestClusters.length + " clusters");
        }
        return impr;
    }
    
    private void setNewBestClusters(){
        IcaCluster[] newBest = new IcaCluster[clusters.size()];
        IcaCluster[] newBestNoise = new IcaCluster[clusters.size()];
        for(int i = 0; i < clusters.size(); i++){
            newBest[i] = (IcaCluster)clusters.elementAt(i);
            newBestNoise[i] = new IcaCluster(-1);    //dummy noise f�r merge
        }
        bestNoiseClusters = newBestNoise;
        bestClusters = newBest;
    }
    
    private void adjustClusterIDs(int oldID, int newID){
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == oldID)
                data[i].clusterID = newID;
        
    }
    
    
    public void greedyMerge(int numTry, boolean verbose){
        if(verbose)
            System.out.println("---GREEDY MERGE, numTry: " + numTry);
        int aktTry = 0;
        Vector<IcaCluster> aktClusters = new Vector<IcaCluster>(); //current clusters and non-empty noiseClusters
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        for(int i = 0; i < bestClusters.length; i++)
            aktClusters.add(bestClusters[i]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0){
            bestNoiseClusters[i].cost += dim*32*2;
            aktClusters.add(bestNoiseClusters[i]);
            }
        clusters = aktClusters;
        //save Clusters and IDs of the last cost minimum
        int[] lastLocalMinIDs = new int[numObj];
        for(int i = 0; i < numObj; i++)
            lastLocalMinIDs[i] = data[i].clusterID;
        IcaCluster[] lastLocalMinClusters = new IcaCluster[clusters.size()];
        for(int i = 0; i < lastLocalMinClusters.length; i++)
            lastLocalMinClusters[i] = (IcaCluster)clusters.elementAt(i);
        bestCost = currentCost();
        boolean improvement = true;
        //determine cost table
        //save the order of the clusters in the cost table by saving their ID
        int[] order = new int[aktClusters.size()];
        for(int i = 0; i < aktClusters.size(); i++)
            order[i] = aktClusters.elementAt(i).ID;
        if(verbose){
            System.out.println("order of clusters in cost table by ID: ");
            for(int i = 0; i < order.length; i++)
                System.out.print(order[i] + "\t");
            System.out.println();
        }
        IcaCluster[] intermediate = new IcaCluster[aktClusters.size()];
        double[][] costTable = new double[aktClusters.size()][aktClusters.size()];
        //init cost table with -1 to check if cost has already been computed
        for(int i = 0; i < costTable.length; i++)
            for(int j = 0; j < costTable.length; j++)
                costTable[i][j] = -1;
        for (int i = 0; i < aktClusters.size(); i++){
            for(int j = 0; j < aktClusters.size(); j ++){
                if(i != j && costTable[i][j] == -1){ //cost needs to be determined
                    for(int k = 0; k < intermediate.length; k++)
                        intermediate[k] = (IcaCluster)aktClusters.elementAt(k);
                    IcaCluster a = (IcaCluster)aktClusters.elementAt(i);
                    IcaCluster b = (IcaCluster)aktClusters.elementAt(j);
                    IcaCluster m = merge(a, b);
//                    if(verbose)
//                        System.out.println("merge: " + a.ID + " and " + b.ID);
                    aktClusters.remove(a);
                    aktClusters.remove(b);
                    aktClusters.add(m);
                    costTable[i][j] = savedCost(false);
                    aktClusters.removeAllElements();
                    for(int k = 0; k < intermediate.length; k++)
                        aktClusters.add(intermediate[k]);
                } else
                    costTable[i][j] = 0.0;
            }
        }
        //print cost table
        if(verbose){
            System.out.println("cost table");
            for(int i = 0; i < costTable.length; i++){
                for(int j = 0; j < costTable.length; j++)
                    System.out.print(costTable[i][j] + "\t");
                System.out.println();
            }
        }
        while(improvement && aktClusters.size() > 1){
            //determine the index of the clusters to be merged next, i.e. the index with maximum saved cost.
            double maxSaved = -Double.MAX_VALUE;
            int i_index = -1;
            int j_index = -1;
            for(int i = 0; i < costTable.length; i++){
                for(int j = 0; j < costTable.length; j++){
                    if(i != j && costTable[i][j] > maxSaved){
                        maxSaved = costTable[i][j];
                        i_index = i;
                        j_index = j;
                    }
                }
            }
            
            if(verbose)
                System.out.println("index of clusters to be merged in cost table: i_index: " + i_index + " j_index: " + j_index + " maxSaved: " + maxSaved);
            if(maxSaved < 0 && aktTry >= numTry)  //stop
                improvement = false;
            else{ //improvement due to merging
                aktTry++;
                IcaCluster a = (IcaCluster)(aktClusters.elementAt(i_index));
                IcaCluster b = (IcaCluster)(aktClusters.elementAt(j_index));
                IcaCluster m = merge(a, b);
                aktClusters.remove(a);
                aktClusters.remove(b);
                aktClusters.add(Math.min(i_index, j_index), m);
                int minID = Math.min(a.ID, b.ID);
                int maxID = Math.max(a.ID, b.ID);
                adjustClusterIDs(maxID, minID);
                setNewBestClusters();
                //determine cost of the new cluster to all other clusters
                double[] mCosts = new double[aktClusters.size()+1];
                //init with -1
                for(int i = 0; i < mCosts.length; i++)
                    mCosts[i] = -1;
                mCosts[Math.min(i_index, j_index)] = 0.0; //index of new cluster
                mCosts[Math.max(i_index, j_index)] = -Double.MAX_VALUE; //index of removed cluster
                IcaCluster[] temp = new IcaCluster[aktClusters.size()]; //save current Clusters
                for(int i = 0; i < temp.length; i++)
                    temp[i] = (IcaCluster)(aktClusters.elementAt(i));
                int checked = 0;
                for(int i = 0; i < mCosts.length; i++){
                    if(mCosts[i] == -1){ //cost needs to be determined
                        IcaCluster akt = (IcaCluster)aktClusters.elementAt(checked);
                        IcaCluster merged = merge(m, akt);
                        aktClusters.remove(m);
                        aktClusters.remove(akt);
                        aktClusters.add(merged);
                        mCosts[i] = savedCost(false);
                        aktClusters.removeAllElements();
                        for(int j = 0; j < temp.length; j++)
                            aktClusters.addElement(temp[j]);
                        checked++;
                    }
                    if(mCosts[i] == 0.0)
                        checked++;
                }
                //update cost table
                //copy old cost table into arry costsNew
                double[][] costsNew = new double[costTable.length][costTable.length];
                for(int i = 0; i < costsNew.length; i++)
                    for(int j = 0; j < costsNew.length; j++)
                        costsNew[i][j] = costTable[i][j];
                //set row and column of removed cluster to -Double.MaxValue
                int indexRemoved = Math.max(i_index, j_index);
                for(int i = 0; i < costsNew.length; i++)
                    costsNew[indexRemoved][i] = -Double.MAX_VALUE;
                for(int i = 0; i < costsNew.length; i++)
                    costsNew[i][indexRemoved] = -Double.MAX_VALUE;
                //set row and colum of new cluster to mCosts
                int indexNew = Math.min(i_index, j_index);
                for(int i = 0; i < costsNew.length; i++)
                    costsNew[indexNew][i] = mCosts[i];
                for(int i = 0; i < costsNew.length; i++)
                    costsNew[i][indexNew] = mCosts[i];
                //copy all values != -Double.MaxValue to array costTableNew
                double[][] costTableNew = new double[aktClusters.size()][aktClusters.size()];
                int written_i = 0;
                int written_j = 0;
                for(int i = 0; i < costsNew.length; i++){
                    for(int j = 0; j < costsNew.length; j++){
                        if(costsNew[i][j] != -Double.MAX_VALUE){
                            costTableNew[written_i][written_j] = costsNew[i][j];
                            written_j++;
                            if(written_j == costTableNew.length){ //row full
                                written_i++;
                                written_j = 0;
                            }
                        }
                    }
                }
                costTable = costTableNew;
                //print cost table
                order = new int[aktClusters.size()];
                for(int i = 0; i < aktClusters.size(); i++)
                    order[i] = aktClusters.elementAt(i).ID;
                if(verbose){
                    System.out.println("order of clusters in cost table by ID: ");
                    for(int i = 0; i < order.length; i++)
                        System.out.print(order[i] + "\t");
                    System.out.println();
                }
                if(verbose){
                    System.out.println("cost table");
                    for(int i = 0; i < costTable.length; i++){
                        for(int j = 0; j < costTable.length; j++)
                            System.out.print(costTable[i][j] + "\t");
                        System.out.println();
                    }
                }
                if(maxSaved > 0){ //local cost minimum: save clusters and IDs as lastlocalMinClusters and lastlocalMinIDs
                    for(int i = 0; i < numObj; i++)
                        lastLocalMinIDs[i] = data[i].clusterID;
                    lastLocalMinClusters = bestClusters;
                }
            } //else (improvement due to merging
        } //while loop
    }
    
    private void checkClusterIDs(){
//collect current clusterIDs
        System.out.println("---checkClusterIDs-----");
        Vector<Integer> id = new Vector<Integer>();
        for(int i = 0; i < bestClusters.length; i++)
            id.addElement(bestClusters[i].ID);
        System.out.println("CURRENT CLUSTER IDS:");
        for(int i = 0; i < id.size(); i++){
            System.out.print(id.elementAt((i)) + " ");
            int count = 0;
            for(int j = 0; j < data.length; j++)
                if(data[j].clusterID == id.elementAt(i))
                    count++;
            System.out.print(count);
            System.out.println();
        }
    }
    
    private IcaCluster merge(IcaCluster c, IcaCluster d){
        boolean verbose = false;
        if(verbose)
            System.out.println("merge clusters: " + c.ID + " " + d.ID);
        //determine best model: rotated or not and TODO: best distribution
        DataObject[] mObj = getMergedObjects(c, d);
        if(verbose)
            System.out.println("c.numObj: " + c.numObj + " d.numObj: " + d.numObj + " m.numObj: " + mObj.length);
        int minID = Math.min(c.ID, d.ID);
        int maxID = Math.max(c.ID, d.ID);
        if(minID == maxID)
            System.err.println("trying to merge clusters with the same id: " + minID);
        if(mObj.length < minNumObj){
            if(verbose)
                System.out.println("Cluster too small for ica");
            IcaCost ic = new IcaCost(dim);
            IcaCluster m = ic.mergedClusterNoRot(mObj);
            m.numObj = mObj.length;
            m.ID = minID;
            return m;
        }
        IcaCost icaC = new IcaCost(dim);
        IcaCluster m = icaC.selectBestCluster(mObj);
        //m.cost += paramCost;
        m.ID = minID;
        return m;
    }
    
    private DataObject[] getObjects(IcaCluster c){
        DataObject[] members = new DataObject[c.numObj];
        int counter = 0;
        for(int j = 0; j < data.length; j++)
            if(data[j].clusterID == c.ID){
            members[counter] = data[j];
            counter++;
            }
        return members;
    }
    
    private DataObject[] getMergedObjects(IcaCluster c, IcaCluster d){
        //DataObject[] members = new DataObject[c.numObj + d.numObj];
        DataObject[] members = new DataObject[countObj(c.ID, d.ID)];
        int counter = 0;
        for(int j = 0; j < data.length; j++)
            if(data[j].clusterID == c.ID || data[j].clusterID == d.ID){
            members[counter] = data[j];
            counter++;
            }
        return members;
    }
    
    public int countObj(int cId, int dId){
        int counter = 0;
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == cId || data[i].clusterID == dId)
                counter++;
        return counter;
    }
    
    /**
     * Returns the current cost of bestClusters and bestNoiseClusters including ID cost
     * @return double: cost
     */
    private double currentCost(){
        boolean verbose = true;
        if(verbose)
            System.out.println("currentCost");
        double cost = 0.0;
        for(int i = 0; i < bestClusters.length; i++){
            IcaCluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Cluster " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                cost += costID;
                cost += bestClusters[i].cost;
            }
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            IcaCluster akt = bestNoiseClusters[i];
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Noise " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                cost += costID;
                cost += bestNoiseClusters[i].cost;
            }
        }
        return cost;
    }
    
    /**
     * Checks if the costs are lower after one split. Cost for clusterID are counted here.
     * @return boolean: improvement of cost
     */
    private boolean improvement(boolean verbose){
        double costNew = 0.0;
        double costOld = 0.0;
        double costIDNew = 0.0;
        double costIDOld = 0.0;
        int clusterCountNew = 0;
        int clusterCountOld = 0;
        if (verbose)
            System.out.println("NEW:");
        for(int i = 0; i < clusters.size(); i++){
            IcaCluster akt = (IcaCluster)clusters.elementAt(i);
            if(clusters.elementAt(i).numObj > 0){
                clusterCountNew ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("cluster " + akt.ID +  " numObj " + akt.numObj + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                costNew += akt.cost;
            } else
                if (verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < noiseClusters.size(); i++){
            IcaCluster akt = (IcaCluster)noiseClusters.elementAt(i);
            if(noiseClusters.elementAt(i).numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                clusterCountNew ++;
                costNew += akt.cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        if(verbose)
            System.out.println("OLD:");
        for(int i = 0; i < bestClusters.length; i++){
            IcaCluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                clusterCountOld ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                if(verbose)
                    System.out.println("cluster " + akt.ID +  " numObj " + akt.numObj + " cost " + akt.cost + " costID: " + costID);
                costOld += bestClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            IcaCluster akt = bestNoiseClusters[i];
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                clusterCountOld ++;
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestNoiseClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        if(costNew < costOld){
            if(verbose)
                System.out.println("improvement : cost old: " + costOld + " " + clusterCountOld + " clusters, cost new: " + " " + costNew + " " + clusterCountNew + " clusters.");
            return true;
        }
        
        else{
            if(verbose)
                System.out.println("no improvement : cost old: " + costOld + " " + clusterCountOld + " clusters, cost new: " + " " + costNew + " " + clusterCountNew + " clusters.");
            return false;
        }
    }
    
    private double savedCost(boolean verbose){
        double costNew = 0.0;
        double costOld = 0.0;
        double costIDNew = 0.0;
        double costIDOld = 0.0;
        int clusterCountNew = 0;
        int clusterCountOld = 0;
        double savedCost;
        if (verbose)
            System.out.println("NEW:");
        for(int i = 0; i < clusters.size(); i++){
            IcaCluster akt = (IcaCluster)clusters.elementAt(i);
            if(clusters.elementAt(i).numObj > 0){
                clusterCountNew ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                costNew += akt.cost;
            } else
                if (verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < noiseClusters.size(); i++){
            IcaCluster akt = (IcaCluster)noiseClusters.elementAt(i);
            if(noiseClusters.elementAt(i).numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                clusterCountNew ++;
                costNew += akt.cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        if(verbose)
            System.out.println("OLD:");
        for(int i = 0; i < bestClusters.length; i++){
            IcaCluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                clusterCountOld ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            IcaCluster akt = bestNoiseClusters[i];
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                clusterCountOld ++;
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestNoiseClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        return costOld-costNew;
    }
    
    public int countClusterMembers(int clusterID){
        int counter = 0;
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == clusterID)
                counter++;
        return counter;
    }
    
    private int countClassMembers(int classID){
        int clM =0;
        for(int i = 0; i < data.length; i++)
            if(data[i].classID == classID)
                clM++;
        return clM;
    }
    
    private int countClassClusterMembers(int clusterID, int classID){
        int clM =0;
        for(int i = 0; i < data.length; i++)
            if(data[i].classID == classID && data[i].clusterID == clusterID)
                clM++;
        return clM;
    }
    
    private int countClasses(){
        Vector clCount = new Vector();
        for(int i = 0; i < data.length; i++){
            Integer clID = (Integer)(data[i].classID);
            if(!clCount.contains(clID))
                clCount.add(clID);
        }
        return clCount.size();
    }
    
    private void printPercentage(){
        int numClasses = countClasses();
        int[] classMembers = new int[numClasses];
        for(int i = 0; i < numClasses; i++)
            classMembers[i] = countClassMembers(i+1);
        for(int i = 0; i < bestClusters.length; i++){
            IcaCluster akt = bestClusters[i];
            System.out.println("cluster " + akt.ID);
            for(int j = 0; j < numClasses; j++){
                //System.out.println("class " + (j+1) + " " + ((double)countClassClusterMembers(akt.ID, j+1)/(double)classMembers[j]) + " percent ");
                System.out.println("class " + (j+1) + " " + countClassClusterMembers(akt.ID, j+1));
                //System.out.println();
            }
            System.out.println();
        }
    }
    
    
    
}




