/*
 * TopDownKM.java
 *
 * Created on 18. April 2006, 15:19
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package clusterer;
import java.util.*;
import Jama.*;
import eval.Clustering;
import eval.SortPack;
import eval.Cluster;
import compr.DataObject;
import compr.DB;
import visu.VisuSimple;
import visu.VisuKMeans;
import visu.Visu2D;

/**
 * Split all the objects belonging to clusters and noiseClusters separately.
 * @author claudia
 */
public class TopDownKMSplitAll {
    DataObject[] data;
    Vector<Cluster> clusters; //current clusters
    Cluster[] bestClusters; //cheapest clusters
    int[] aktClusterID; //current cluster IDs;
    DB db;
    double bestCost;
    int iter;
    int numObj;
    int dim;
    int precision;
    int maxClusterID;      //Max ID of the clusters (not noise!)
    int numTry = 0;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    static int gauss = 1;
    static int laplace = 2;
    
    /** Creates a new instance of TopDownKM */
    public TopDownKMSplitAll(DataObject[] data, int precision) {
        this.data = data;
        db = new DB(data);
        numObj = data.length;
        iter = 0;
        dim = data[0].coord.length;
        maxClusterID = -1;
        this.precision = precision;
        bestCost = 0.0;
        clusters = new Vector<Cluster>();
    }
    
    /**
     * Returns the cluster with the best model: Rotated or not, different distributions. Cost per Cluster
     * @param obj DataObject[]: objects belonging to the cluster
     * @return Cluster: if rotated with rotation matrix.
     */
    private Cluster bestModel(DataObject[] obj, boolean verbose){
        //verbose = true;
        double gaussCost = 0.0;
        double laplaceCost = 0.0;
        if(verbose)
            System.out.println("bestModel: " + obj.length + " objects.");
        Cluster best = new Cluster();
        double addVar = 0.0;
        Cluster clNotRot = new Cluster(dim, 0);
        clNotRot.mean = mean(obj);
        clNotRot.numObj = obj.length;
        clNotRot.noiseCluster = false;
        for(int i = 0; i < dim; i++){
            clNotRot.var[i] = var(obj, clNotRot.mean, i);
            addVar += clNotRot.var[i];
        }
        for(int i = 0; i < dim; i++){
            for(int j = 0; j < obj.length; j++){
                gaussCost += gaussCost(obj[j].coord[i], clNotRot.mean[i], clNotRot.var[i], precision);
                laplaceCost += laplaceCost(obj[j].coord[i], clNotRot.mean[i], clNotRot.var[i], precision);
            }
        }
        clNotRot.cost += Math.min(gaussCost, laplaceCost);
        if(gaussCost < laplaceCost){
            clNotRot.distr = gauss;
            if(verbose)
                System.out.println("gaussian");
        } else{
            clNotRot.distr = laplace;
            if(verbose)
                System.out.println("laplacian");
        }
        //Parameter ////////////////////////
        clNotRot.cost += dim * 32 * 2;
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("Cost not rotated " + clNotRot.cost);
        
        gaussCost = 0.0;
        laplaceCost = 0.0;
        Cluster clRot = new Cluster(dim, 0);
        clRot.mean = median(obj);
        clRot.numObj = obj.length;
        clRot.noiseCluster = false;
        robustEstimate(clRot, obj, false);
        //transform objects to rotated space
        DataObject[] red = new DataObject[obj.length];
        Matrix m = new Matrix(clRot.dir);
        for(int l = 0; l < red.length; l++){
            double[] redCoord = new double[dim];
            Matrix x = rowVector(obj[l].coord);
            Matrix akt = x.times(m);
            for(int j = 0; j < dim; j++){
                redCoord[j] = akt.get(0,j);
            }
            DataObject redObj = new DataObject(redCoord, obj[l].number);
            redObj.clusterID = obj[l].clusterID;
            red[l] = redObj;
        }
        double[] redMeanCoord = new double[dim];
        Matrix mx = rowVector(clRot.mean);
        Matrix akt = mx.times(m);
        for(int j = 0; j < dim; j++){
            redMeanCoord[j] = akt.get(0,j);
        }
        for(int i = 0; i < dim; i++)
            clRot.var[i] = var(red, redMeanCoord, i);
        for(int i = 0; i < dim; i++){
            for(int j = 0; j < obj.length; j++){
                gaussCost += gaussCost(red[j].coord[i], redMeanCoord[i], clRot.var[i], precision);
                laplaceCost += laplaceCost(red[j].coord[i], redMeanCoord[i], clRot.var[i], precision);
            }
        }
        clRot.cost += Math.min(gaussCost, laplaceCost);
        if(gaussCost < laplaceCost)
            clRot.distr = gauss;
        else
            clRot.distr = laplace;
        //Parameter ////////////////////////
        clRot.cost += dim*32*2 + dim*dim * 32;
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("Cost rotated " + clRot.cost);
        
        //all objects noise
        if(verbose)
            System.out.println("---all objects noise---");
        Cluster allNoise = new Cluster(dim, 0);
        allNoise.noiseCluster = false;
        allNoise.mean = clNotRot.mean;
        allNoise.numObj = obj.length;
        for(int i = 0; i < dim; i++)
            allNoise.var[i] = addVar;
        for(int i = 0; i < obj.length; i++)
            for(int j = 0; j < dim; j++)
                //allNoise.cost += gaussCost(obj[i].coord[j], allNoise.mean[j], allNoise.var[j], precision);
                //TEST
                allNoise.cost += gaussCost(obj[i].coord[j], allNoise.mean[j], clNotRot.var[j], precision);
                //TEST
        //////////////////////////
        allNoise.cost += dim * 32 * 2;
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("Cost noise " + allNoise.cost);
        if(clRot.cost < clNotRot.cost && clRot.cost < allNoise.cost)
            return clRot;
        if(clNotRot.cost < clRot.cost && clNotRot.cost < allNoise.cost || allNoise.cost == clNotRot.cost)
            return clNotRot;
        else
            return allNoise;
    }
    
    /**
     * Returns the filtered clusters with the best cost. Rotated or not, splitted in cluster and noise cluster or not. Cost of resulting clusters: Cost with parameters but without ID cost.
     * Cluster: clusterID 0, Noise: clusterID 1
     * @param data DataObject[]: objects to be filtered
     * @return Cluster[2]: result[0]: filtered cluster, result[1]: filtered noise, can be empty (numObj = 0).
     */
    private Cluster[] getBestFilteredClusters(DataObject[] data, boolean verbose){
        if(verbose)
            System.out.println("getBestFilteredClusters: numObj " + data.length);
        if(data.length == 1){
            if(verbose)
                System.out.println("Singleton cluster");
            Cluster[] result = new Cluster[2];
            result[0] = new Cluster(0);
            result[0].cost = dim * 32 * 2;
            result[0].numObj = 1;
            result[1] = new Cluster(1);
            //clusterID
            int[] ids = new int[1];
            ids[0] = 0;
            aktClusterID = ids;
            return result;
        }
        Cluster c = bestModel(data, false);
        if(verbose)
            System.out.println("--- All objects in cluster: " + c.cost);
        if(verbose)
            System.out.println("--filtered cluster and noise: gausssian---");
//get Filtered Clusters gauss and laplace, rotated and not rotated
        DataObject[] filteredRotatedData_g = new DataObject[data.length];
        for(int i = 0; i < data.length; i++)
            filteredRotatedData_g[i] = new DataObject(data[i]);
        KMCost kmkr_g = new KMCost(filteredRotatedData_g, 1, precision, true, 1);
        kmkr_g.run();
        Cluster filteredRotatedCluster_g = kmkr_g.getClusters()[0];
        filteredRotatedCluster_g.noiseCluster = false;
        Cluster filteredRotatedNoise_g = kmkr_g.getNoiseClusters()[0];
        filteredRotatedNoise_g.noiseCluster = true;
        double costFilteredRotated_g = filteredRotatedCluster_g.cost + filteredRotatedNoise_g.cost;
        double costIDClusterRot_g = 0.0;
        double costIDNoiseRot_g = 0.0;
        if(filteredRotatedCluster_g.numObj > 0 && filteredRotatedNoise_g.numObj > 0){
            costIDClusterRot_g = filteredRotatedCluster_g.numObj * Math.log((double)data.length/(double)filteredRotatedCluster_g.numObj)/Math.log(2);
            costIDNoiseRot_g = filteredRotatedNoise_g.numObj * Math.log((double)data.length/(double)filteredRotatedNoise_g.numObj)/Math.log(2);
            if(verbose)
                System.out.println("costIDClusterRot_g " + costIDClusterRot_g + " costIDNoiseRot_g " + costIDNoiseRot_g);
        }
        //Parameter ////////////////////////
        costFilteredRotated_g += dim * 32 * 2 + dim * dim * 32  + costIDClusterRot_g + costIDNoiseRot_g;
        filteredRotatedCluster_g.cost += dim * 32 * 2 + dim * dim * 32;
        //filteredRotatedNoise.cost += dim * 32 * 2;
        if(verbose)
            System.out.println("cost cluster_g: " + filteredRotatedCluster_g.cost);
        if(verbose)
            System.out.println("cost noise_g: " + filteredRotatedNoise_g.cost);
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("Cost filteredRotated_g " + costFilteredRotated_g);
        
        DataObject[] filteredData_g = new DataObject[data.length];
        for(int i = 0; i < filteredData_g.length; i++)
            filteredData_g[i] = new DataObject(data[i]);
        KMCost kmk_g= new KMCost(filteredData_g, 1, precision, false, 1);
        kmk_g.run();
        Cluster filteredCluster_g = kmk_g.getClusters()[0];
        filteredCluster_g.noiseCluster = false;
        Cluster filteredNoise_g = kmk_g.getNoiseClusters()[0];
        filteredNoise_g.noiseCluster = true;
        double costFilteredNotRotated_g = filteredCluster_g.cost + filteredNoise_g.cost;
        double costIDCluster = 0.0;
        double costIDNoise = 0.0;
        if(filteredCluster_g.numObj > 0 && filteredNoise_g.numObj > 0){
            costIDCluster = filteredCluster_g.numObj * Math.log((double)data.length/(double)filteredCluster_g.numObj)/Math.log(2);
            costIDNoise = filteredNoise_g.numObj * Math.log((double)data.length/(double)filteredNoise_g.numObj)/Math.log(2);
            if(verbose)
                System.out.println("CostIDCluster_g " + costIDCluster + " costIDNoise_g " + costIDNoise);
        }
        //Parameter////////////////////////
        costFilteredNotRotated_g += dim * 32 * 2 + costIDCluster + costIDNoise;
        filteredCluster_g.cost += dim * 32 * 2;
        //filteredNoise_g.cost += dim * 32 * 2;
        if(verbose)
            System.out.println("cost cluster_g: " + filteredCluster_g.cost);
        if(verbose)
            System.out.println("cost noise_g: " + filteredNoise_g.cost);
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("CostfilteredNotRotated_g " + costFilteredNotRotated_g);
        
        /////////////////////////////////////////////////////////////////////////////////////////////////////////////
        if(verbose)
            System.out.println("--filtered cluster and noise: laplacian---");
        DataObject[] filteredRotatedData_l = new DataObject[data.length];
        for(int i = 0; i < data.length; i++)
            filteredRotatedData_l[i] = new DataObject(data[i]);
        KMCost kmkr_l = new KMCost(filteredRotatedData_l, 1, precision, true, 2);
        kmkr_l.run();
        Cluster filteredRotatedCluster_l = kmkr_l.getClusters()[0];
        filteredRotatedCluster_l.noiseCluster = false;
        Cluster filteredRotatedNoise_l = kmkr_l.getNoiseClusters()[0];
        filteredRotatedNoise_l.noiseCluster = true;
        double costFilteredRotated_l = filteredRotatedCluster_l.cost + filteredRotatedNoise_l.cost;
        double costIDClusterRot_l = 0.0;
        double costIDNoiseRot_l = 0.0;
        if(filteredRotatedCluster_l.numObj > 0 && filteredRotatedNoise_l.numObj > 0){
            costIDClusterRot_l = filteredRotatedCluster_l.numObj * Math.log((double)data.length/(double)filteredRotatedCluster_l.numObj)/Math.log(2);
            costIDNoiseRot_l = filteredRotatedNoise_l.numObj * Math.log((double)data.length/(double)filteredRotatedNoise_l.numObj)/Math.log(2);
            if(verbose)
                System.out.println("costIDClusterRot_l " + costIDClusterRot_l + " costIDNoiseRot " + costIDNoiseRot_l);
        }
        //Parameter ////////////////////////
        costFilteredRotated_l += dim * 32 * 2 + dim * dim * 32  + costIDClusterRot_l + costIDNoiseRot_l;
        filteredRotatedCluster_l.cost += dim * 32 * 2 + dim * dim * 32;
        //filteredRotatedNoise.cost += dim * 32 * 2;
        if(verbose)
            System.out.println("cost cluster_l: " + filteredRotatedCluster_l.cost);
        if(verbose)
            System.out.println("cost noise_l: " + filteredRotatedNoise_l.cost);
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("Cost filteredRotated " + costFilteredRotated_l);
        DataObject[] filteredData_l = new DataObject[data.length];
        for(int i = 0; i < filteredData_l.length; i++)
            filteredData_l[i] = new DataObject(data[i]);
        KMCost kmk_l = new KMCost(filteredData_l, 1, precision, false, 2);
        kmk_l.run();
        Cluster filteredCluster_l = kmk_l.getClusters()[0];
        filteredCluster_l.noiseCluster = false;
        Cluster filteredNoise_l = kmk_l.getNoiseClusters()[0];
        filteredNoise_l.noiseCluster = true;
        double costFilteredNotRotated_l = filteredCluster_l.cost + filteredNoise_l.cost;
        double costIDCluster_l = 0.0;
        double costIDNoise_l = 0.0;
        if(filteredCluster_l.numObj > 0 && filteredNoise_l.numObj > 0){
            costIDCluster_l = filteredCluster_l.numObj * Math.log((double)data.length/(double)filteredCluster_l.numObj)/Math.log(2);
            costIDNoise_l = filteredNoise_l.numObj * Math.log((double)data.length/(double)filteredNoise_l.numObj)/Math.log(2);
            if(verbose)
                System.out.println("CostIDCluster " + costIDCluster_l + " costIDNoise " + costIDNoise_l);
        }
        //Parameter////////////////////////
        costFilteredNotRotated_l += dim * 32 * 2 + costIDCluster_l + costIDNoise_l;
        filteredCluster_l.cost += dim * 32 * 2;
        //filteredNoise_l.cost += dim * 32 * 2;
        if(verbose)
            System.out.println("cost cluster: " + filteredCluster_l.cost);
        if(verbose)
            System.out.println("cost noise: " + filteredNoise_l.cost);
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("CostfilteredNotRotated " + costFilteredNotRotated_l);
        double minFiltered_g = Math.min(costFilteredRotated_g, costFilteredNotRotated_g);
        double minFiltered_l = Math.min(costFilteredRotated_l, costFilteredNotRotated_l);
        double minFiltered = Math.min(minFiltered_g, minFiltered_l);
        Cluster[] result = new Cluster[2];
        int[] ids = new int[data.length];
        if(minFiltered < c.cost){
            //save current cluster IDs
            if(minFiltered_g < minFiltered_l){
                if(costFilteredRotated_g < costFilteredNotRotated_g){
                    //save current cluster IDS
                    for(int i = 0; i < ids.length; i++)
                        ids[i] = filteredRotatedData_g[i].clusterID;
                    aktClusterID = ids;
                    result[0] = filteredRotatedCluster_g;
                    result[0].distr = gauss;
                    result[1] = filteredRotatedNoise_g;
                } else{
                    for(int i = 0; i < ids.length; i++)
                        ids[i] = filteredData_g[i].clusterID;
                    aktClusterID = ids;
                    result[0] = filteredCluster_g;
                    result[0].distr = gauss;
                    result[1] = filteredNoise_g;
                }
                return result;
            } else{
                if(costFilteredRotated_l < costFilteredNotRotated_l){
                    //save current cluster IDS
                    ids = new int[filteredRotatedData_l.length];
                    for(int i = 0; i < ids.length; i++)
                        ids[i] = filteredRotatedData_l[i].clusterID;
                    aktClusterID = ids;
                    result[0] = filteredRotatedCluster_l;
                    result[0].distr = laplace;
                    result[1] = filteredRotatedNoise_l;
                } else{
                    ids = new int[filteredData_l.length];
                    for(int i = 0; i < ids.length; i++)
                        ids[i] = filteredData_l[i].clusterID;
                    aktClusterID = ids;
                    result[0] = filteredCluster_l;
                    result[0].distr = laplace;
                    result[1] = filteredNoise_l;
                }
                result[1].noiseCluster = true;
                return result;
            }
        } else{
            //all objects in cluster, save current cluster IDS
            for(int i = 0; i < ids.length; i++)
                ids[i] = 0;
            aktClusterID = ids;
            result[0] = c;
            result[1] = new Cluster(1);
            return result;
        }
    }
    
    public void testRun(){
        Cluster[] result = getBestFilteredClusters(data, true);
        VisuSimple vk = new VisuSimple(data);
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
    }
    
    public void run(){
        //save clusters and clusterIDs of the last local minimum
        Cluster[] lastLocalMinClusters = new Cluster[1];
        int[] lastLocalMinIDS = new int[numObj];
        Cluster[] result = getBestFilteredClusters(data, false);
        //adjust clusterIDs
        for(int i = 0; i < data.length; i++)
            data[i].clusterID = aktClusterID[i];
        //set best Clusters as clusters and the non-empty noiseClusters
        if(result[1].numObj > 0){
            bestClusters = new Cluster[2];
            bestClusters[0] = result[0];
            bestClusters[1] = result[1];
            maxClusterID = maxClusterID + 2;
        } else{
            bestClusters = new Cluster[1];
            bestClusters[0] = result[0];
            maxClusterID++;
        }
        double costFiltered = currentCost();
        double minCost = costFiltered;
        System.out.println("filtered data set " + costFiltered);
        int[] filteredID = new int[data.length];
        for(int i = 0; i < data.length; i++)
            filteredID[i] = data[i].clusterID;
        boolean improvement = true;
        int aktTry = 0;
        lastLocalMinClusters = bestClusters;        //save Clusters and ClusterIDs of the last local minimum
        int[] lastLocalMinIDs = new int[numObj];
        for(int i = 0; i < numObj; i++)
            lastLocalMinIDs[i] = data[i].clusterID;
        while(improvement || aktTry < numTry){
            System.out.println("-------------------------------------");
            double costUnsplit = currentCost();
            int numClUnsplit = bestClusters.length;
            System.out.println("SYMMETRIC SPLIT NUMBER: " + iter + " CURRENT COST " + costUnsplit);
            improvement = symmetricSplit(aktTry);
            if(improvement){
                minCost = currentCost();
                System.out.println("GLOBAL IMPROVEMENT: " + " CURRENT COST " + minCost + " COST OLD: " + costUnsplit);
                lastLocalMinClusters = bestClusters;
                for(int i = 0; i < numObj; i++)
                    lastLocalMinIDs[i] = data[i].clusterID;
                aktTry = 0;
            } else{
                aktTry++;
                System.out.println("NO GLOBAL IMPROVEMENT: CURRENT COST " + currentCost() + " COST OLD: " + costUnsplit + " TRYING FOR ANOTHER " + (numTry - aktTry) + " TIMES");
            }
            System.out.println("OLD: " + numClUnsplit + " clusters");
            System.out.println("NEW: " + bestClusters.length + " clusters. ");
            System.out.println("-------------------------------------");
            iter++;
        }
        System.out.println("done. " + iter + " iterations");
        
        //adjust clusters and IDs as from the last local minimum
        if(lastLocalMinClusters.length > 1){
            bestClusters = lastLocalMinClusters;
            for(int i = 0; i < data.length; i++)
                data[i].clusterID = lastLocalMinIDs[i];
            System.out.println("BEST: " + bestClusters.length + " clusters ");
        }
        
        //mergeClusters();
        minCost = currentCost();
        System.out.println("BEST COST: " + minCost);
        //testMerge();
        
        if(costFiltered < minCost)
            for(int i = 0; i < data.length; i++)
                data[i].clusterID = filteredID[i];
        
        VisuSimple vk = new VisuSimple(data);
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
    }
    
    /*
    private int[] numClustersNumNoiseClusters(){
        int numCl = 0;
        int numN = 0;
        for(int i = 0; i < bestClusters.length; i++)
            if(bestClusters[i].numObj > 0)
                numCl++;
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                numN++;
        int[] result = new int[2];
        result[0] = numCl;
        result[1] = numN;
        return result;
    }
     */
    
    private boolean symmetricSplit(int aktTry){
        Cluster[] currentClusters = bestClusters.clone();
        boolean[] improvement = new boolean[currentClusters.length];
        for(int i = 0; i < currentClusters.length; i++){
            improvement[i] = split(currentClusters[i], false);
        }
        boolean impr = false;
        for(int i = 0; i < improvement.length; i++)
            if(improvement[i])
                impr = true;
        if(!impr && aktTry < numTry){
            for(int i = 0; i < currentClusters.length; i++)
                improvement[i] = split(currentClusters[i], true);
        }
        return impr;
    }
    
    private boolean split(Cluster c, boolean splitAnyway){
        if(c.numObj < 2)
            return false;
        boolean improvement;
        int[] currentID1 = new int[numObj]; //temporary clusterIDs for clusters and noise clusters
        int[] currentID2;
        DataObject[] all = db.getObjects(c.ID);
        DB kmDb = new DB(all);
        KMeans k = new KMeans(2, kmDb);
        DataObject[] newIDs = k.run(); //ClusterIDs 1 and 2
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = kmDb.getObjects(1);
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for(int i = 0; i < id_1.length; i++){
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap h_1 = new HashMap(id_1.length);
        for(int i = 0; i < id_1.length; i++)
            h_1.put(internalIndex_1[i], numbers_1[i]);
        for(int i = 0; i < id_1.length; i++)
            id_1[i].number = internalIndex_1[i];
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        Cluster[] clusters1 = getBestFilteredClusters(id_1, false);
        currentID1 = aktClusterID;
        clusters1[0].ID = c.ID;                             //cluster1: c.ID
        clusters1[1].ID = maxClusterID+1;                   //noise1: maxClusterID+1;
        clusters1[1].noiseCluster = true;
        //objects assigned to clusterID 2 by kMeans
        DataObject[] id_2 = kmDb.getObjects(2);
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for(int i = 0; i < id_2.length; i++){
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap h_2 = new HashMap(id_2.length);
        for(int i = 0; i < id_2.length; i++)
            h_2.put(internalIndex_2[i], numbers_2[i]);
        for(int i = 0; i < id_2.length; i++)
            id_2[i].number = internalIndex_2[i];
        Cluster[] clusters2 = getBestFilteredClusters(id_2, false);
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;     //cluster2: maxID +2
        clusters2[1].ID = maxClusterID + 3;     //noise2: maxID + 3
        clusters2[1].noiseCluster = true;
        //update current clusters
        for(int i = 0; i < bestClusters.length; i++)
            if(bestClusters[i].ID != c.ID)
                clusters.add(bestClusters[i]);
        clusters.add(clusters1[0]);
        clusters.add(clusters1[1]);
        clusters.add(clusters2[0]);
        clusters.add(clusters2[1]);
        if(splitAnyway)
            improvement = true;
        else
            improvement = improvement(false);
        //set new clusters as bestClusters
        if(improvement){
            Cluster[] newClusters = new Cluster[clusters.size()];
            for(int i = 0; i < clusters.size(); i++)
                newClusters[i] = (Cluster)clusters.elementAt(i);
            bestClusters = newClusters;
            //reset bestCost
            bestCost = 0.0;
            for(int i = 0; i < bestClusters.length; i++)
                bestCost += bestClusters[i].cost;
            //update clusterIDs
            for(int i = 0; i < currentID1.length; i++){
                int index = ((Integer)h_1.get(i)).intValue();
                if(currentID1[i] == 0)
                    data[index].clusterID = c.ID;
                else
                    data[index].clusterID = maxClusterID + 1;
            }
            for(int i = 0; i < currentID2.length; i++){
                int index = ((Integer)h_2.get(i)).intValue();
                if(currentID2[i] == 0)
                    data[index].clusterID = maxClusterID + 2;
                else
                    data[index].clusterID = maxClusterID + 3;
            }
            maxClusterID = maxClusterID + 3;
        }
        clusters.removeAllElements();
        return improvement;
    }
    
    
    
    
    
    /**
     * Returns the clusters resulting from splitting a cluster.
     * @param c Cluster to split
     * @return Cluster[]: 0: cluster1, 1: noise1, 2: cluster2, 3: noise2. ClusterIDs: 0: c.ID, 1-3 maxClusterID++
     */
    private Cluster[] getSplitted(Cluster c){
        Cluster[] result = new Cluster[4];
        int[] currentID1 = new int[numObj]; //temporary clusterIDs for clusters and noise clusters
        int[] currentID2;
        DataObject[] all = db.getObjects(c.ID);
        DB kmDb = new DB(all);
        KMeans k = new KMeans(2, kmDb);
        DataObject[] newIDs = k.run(); //ClusterIDs 1 and 2
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = kmDb.getObjects(1);
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for(int i = 0; i < id_1.length; i++){
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap h_1 = new HashMap(id_1.length);
        for(int i = 0; i < id_1.length; i++)
            h_1.put(internalIndex_1[i], numbers_1[i]);
        for(int i = 0; i < id_1.length; i++)
            id_1[i].number = internalIndex_1[i];
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        Cluster[] clusters1 = getBestFilteredClusters(id_1, false);
        currentID1 = aktClusterID;
        clusters1[0].ID = c.ID;                             //cluster1: c.ID
        clusters1[1].ID = maxClusterID+1;                   //noise1: maxClusterID+1;
        clusters1[1].noiseCluster = true;
        //objects assigned to clusterID 2 by kMeans
        DataObject[] id_2 = kmDb.getObjects(2);
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for(int i = 0; i < id_2.length; i++){
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap h_2 = new HashMap(id_2.length);
        for(int i = 0; i < id_2.length; i++)
            h_2.put(internalIndex_2[i], numbers_2[i]);
        for(int i = 0; i < id_2.length; i++)
            id_2[i].number = internalIndex_2[i];
        Cluster[] clusters2 = getBestFilteredClusters(id_2, false);
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;         //cluster2
        clusters2[1].ID = maxClusterID + 3;         //noise2
        clusters2[1].noiseCluster = true;
        result[0] = clusters1[0];
        result[1] = clusters1[1];
        result[2] = clusters2[0];
        result[3] = clusters2[1];
        return result;
    }
    
    
    
    
    private void setNewBestClusters(){
        Cluster[] newBest = new Cluster[clusters.size()];
        for(int i = 0; i < clusters.size(); i++)
            newBest[i] = (Cluster)clusters.elementAt(i);
        bestClusters = newBest;
    }
    
    private void adjustClusterIDs(int oldID, int newID){
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == oldID)
                data[i].clusterID = newID;
        
    }
    /*
    private void mergeClusters(){
        int numRandom = 1000;
        Random r = new Random(2);
        Cluster c;
        Cluster d;
        Vector<Cluster> aktClusters = new Vector<Cluster>(); //current clusters and non-empty noiseClusters
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        for(int i = 0; i < bestClusters.length; i++)
            aktClusters.add(bestClusters[i]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                aktClusters.add(bestNoiseClusters[i]);
        clusters = aktClusters;
        //all clusters and noiseClusters to be merged in bestClusters, dummy noiseCluster.
        Cluster[] nonEmpty = new Cluster[aktClusters.size()];
        for(int i = 0; i < nonEmpty.length; i++)
            nonEmpty[i] = (Cluster)aktClusters.elementAt(i);
        bestClusters = nonEmpty;
        Cluster[] newNoise = new Cluster[1];
        newNoise[0] = new Cluster(1);
        newNoise[0].numObj = 0;
        bestNoiseClusters = newNoise; //dummy noisecluster
        for(int i = 0; i < numRandom; i++){
            if(bestClusters.length > 1){
                //merge random clusters and noiseClusters
                int nextMerge1 = r.nextInt(bestClusters.length);
                int nextMerge2 = r.nextInt(bestClusters.length);
                while(nextMerge2 == nextMerge1)
                    nextMerge2 = r.nextInt(bestClusters.length);
                //if(bestClusters[nextMerge2].ID == bestClusters[nextMerge1].ID+1)       //do not try to merge cluster with its own noise cluster.
                //  nextMerge2++;
                //objects with this clusterID need to be relabeled, since the merged cluster gets the minimumID of c and d.
                c = bestClusters[nextMerge1];
                d = bestClusters[nextMerge2];
                int maxID = Math.max(c.ID, d.ID);
                int minID = Math.min(c.ID, d.ID);
                Cluster m = merge(c, d);
                clusters.remove(c);
                clusters.remove(d);
                clusters.add(m);
                if(improvement(false)){
                    //adjust clusterIDs
                    adjustClusterIDs(maxID, minID);
                    //set new clusters as best clusters
                    setNewBestClusters();
                } else{
                    clusters.remove(m);
                    clusters.add(c);
                    clusters.add(d);
                }
            }
            //clusters.removeAllElements();
            //noiseClusters.removeAllElements();
        }
    }
     */
    
    private Cluster merge(Cluster c, Cluster d){
        System.out.println("merge clusters: " + c.ID + " " + d.ID);
        //determine best model: rotated or not and TODO: best distribution
        DataObject[] mObj = getMergedObjects(c, d);
        int minID = Math.min(c.ID, d.ID);
        int maxID = Math.max(c.ID, d.ID);
        if(minID == maxID)
            System.err.println("trying to merge clusters with the same id: " + minID);
        Cluster m = bestModel(mObj, true);
        m.ID = minID;
        return m;
    }
    
    private DataObject[] getMergedObjects(Cluster c, Cluster d){
        //DataObject[] members = new DataObject[c.numObj + d.numObj];
        DataObject[] members = new DataObject[countObj(c.ID, d.ID)];
        int counter = 0;
        //if(c.ID == 6 && d.ID == 5)
        System.out.println("obj count" + countObj(c.ID,  d.ID) + " numObj " + (c.numObj + d.numObj));
        for(int j = 0; j < data.length; j++)
            if(data[j].clusterID == c.ID || data[j].clusterID == d.ID){
            members[counter] = data[j];
            counter++;
            }
        return members;
    }
    
    public int countObj(int cId, int dId){
        int counter = 0;
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == cId || data[i].clusterID == dId)
                counter++;
        return counter;
    }
    
    /**
     * Returns the current cost of bestClusters and bestNoiseClusters including ID cost
     * @return double: cost
     */
    private double currentCost(){
        boolean verbose = false;
        if(verbose)
            System.out.println("currentCost");
        double cost = 0.0;
        for(int i = 0; i < bestClusters.length; i++){
            Cluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Cluster " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                cost += costID;
                cost += bestClusters[i].cost;
            }
        }
        return cost;
    }
    
    /**
     * Checks if the costs are lower after one split. Cost for clusterID are counted here.
     * @return boolean: improvement of cost
     */
    private boolean improvement(boolean verbose){
        double costNew = 0.0;
        double costOld = 0.0;
        double costIDNew = 0.0;
        double costIDOld = 0.0;
        int clusterCountNew = 0;
        int clusterCountOld = 0;
        if (verbose)
            System.out.println("NEW:");
        for(int i = 0; i < clusters.size(); i++){
            Cluster akt = (Cluster)clusters.elementAt(i);
            if(clusters.elementAt(i).numObj > 0){
                clusterCountNew ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                costNew += akt.cost;
            } else
                if (verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        if(verbose)
            System.out.println("OLD:");
        for(int i = 0; i < bestClusters.length; i++){
            Cluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                clusterCountOld ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        if(costNew < costOld){
            if(verbose)
                System.out.println("improvement : cost old: " + costOld + " " + clusterCountOld + " clusters, cost new: " + " " + costNew + " " + clusterCountNew + " clusters.");
            return true;
        }
        
        else{
            if(verbose)
                System.out.println("no improvement : cost old: " + costOld + " " + clusterCountOld + " clusters, cost new: " + " " + costNew + " " + clusterCountNew + " clusters.");
            return false;
        }
    }
    
    private void robustEstimate(Cluster c, DataObject[] data, boolean verbose){
        if(verbose)
            System.out.println("Cluster " + c.ID);
        DataObject[] obj = data;
        double[] median = median(obj);
        c.mean = median;
        double[][] covarm = covarByMedian(data, median);
        Matrix covar = new Matrix(covarm);
        EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
        Matrix vectors = eig.getV();
        Matrix values = eig.getD();
        if(verbose){
            System.out.println("EW");
            values.print(5, 3);
            System.out.println("EV");
            vectors.print(5, 3);
        }
        c.dir = vectors.getArrayCopy();
        double maxEigen = -Double.MAX_VALUE ;
        for (int i=0 ; i<dim ; i++)
            maxEigen = Math.max(maxEigen, values.get(i,i));
        for (int i=0 ; i<dim ; i++)
            values.set(i,i,1/Math.max(values.get(i,i), maxEigen*0.001)) ;
        c.invDir =  (vectors.times(values).times(vectors.transpose())).getArrayCopy() ;
    }
    
//var of coordinate numCoord
    private double var(DataObject[] o, double[] mean, int numCoord){
        double var = 0.0;
        for(int i = 0; i < o.length; i++)
            var += (o[i].coord[numCoord] - mean[numCoord]) * (o[i].coord[numCoord] - mean[numCoord]);
        return var/(double)o.length;
    }
    
    private double variance(DataObject[] o, double[] mean){
        double var = 0.0;
        for(int i = 0; i < o.length; i++)
            var += o[i].distance(mean) * o[i].distance(mean);
        return var/(double)o.length;
    }
    
    private double gaussCost(double x, double mean, double var, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double va = Math.max(var, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*va)+precision*ln10+Math.pow(x-mean,2)/2/va)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    private double laplaceCost(double x, double mean, double var, int precision){
        double stdv = Math.sqrt(var);
        double b = stdv/Math.sqrt(2);
        if (b<Math.pow(10, -precision)/10)
            return 0 ;
        double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
        double prob = Math.max(Double.MIN_VALUE, Math.min(1.0, pdf * Math.pow(10, -precision)));
        //return 100000;
        return Math.log(1.0/prob)/Math.log(2);
    }
    
    private double[] mean(DataObject[] o){
        double[] mean = new double[o[0].coord.length];
        for(int i = 0; i < o.length; i++)
            for(int j = 0; j < mean.length; j++){
            mean[j] += o[i].coord[j];
            }
        for(int j = 0; j < mean.length; j++)
            mean[j] /= o.length;
        return mean;
    }
    
    public double [] median(DataObject[] members) {
        int numObj= members.length ;
        int dim = members[0].coord.length ;
        double[] result = new double[dim];
        double [] sortarray = new double[numObj] ;
        for (int i=0 ; i<dim ; i++) {
            double median = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sortarray[j]=members[j].coord[i] ;
            Arrays.sort(sortarray) ;
            if (numObj==1)
                result[i]=sortarray[0] ;
            else if (numObj%2 == 0)
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
            else
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
        }
        return result ;
    }
    
    public double[][] covarByMedian(DataObject[] data, double[] mean) {
        int numObj= data.length ;
        int dim = data[0].coord.length ;
        double [] sortarray = new double[numObj] ;
        double[][] cov = new double[dim][dim];
        if (numObj <15){
            for (int i=0 ; i<dim ; i++)
                cov[i][i]=1.0 ;
            return cov ;
        }
        for(int l1 = 0; l1 < dim; l1++) {
            for(int l2 = 0; l2 < dim; l2++) {
                for(int j = 0; j < numObj; j++)
                    sortarray[j] = (data[j].coord[l1]-mean[l1])*(data[j].coord[l2]-mean[l2]) ;
                
                Arrays.sort(sortarray) ;
                if (numObj%2 == 0)
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3
                            ;
            }
            
        }
        double maxdiff = 0.0 ;
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                sum += Math.abs(cov[i][j]) ;
            double diff = 1.1*sum - 2.1*cov[i][i] ;
            if (diff>maxdiff)
                maxdiff = diff ;
        }
        if (maxdiff > 0.0)
            for (int i=0 ; i<dim ; i++)
                cov[i][i] += maxdiff ;
        
        return cov ;
    }
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
}

