/*
 * TopDownKM.java
 *
 * Created on 18. April 2006, 15:19
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package clusterer;
import java.util.*;
import Jama.*;
import eval.Clustering;
import eval.SortPack;
import eval.Cluster;
import eval.ClusteringEvaluation;
import compr.DataObject;
import compr.DB;
import filterCluster.FilterKM;
import visu.VisuSimple;
import visu.VisuKMeans;
import visu.Visu2D;
import visu.VisuInfo;
import io.IO;

/**
 * Split all the objects belonging to one cluster and the corresponding noise cluster.
 * @author claudia
 */
public class TopDownKMWithoutGrid {
    DataObject[] data;
    Vector<Cluster> clusters; //current clusters
    Vector<Cluster> noiseClusters;
    Cluster[] bestClusters; //cheapest clusters
    Cluster[] bestNoiseClusters;
    int[] aktClusterID; //current cluster IDs;
    DB db;
    double bestCost;
    int iter;
    int numObj;
    int dim;
    int precision;
    int maxClusterID;      //Max ID of the clusters (not noise!)
    int numTry = 0;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    static int gauss = 1;
    static int laplace = 2;
    
    /** Creates a new instance of TopDownKM */
    public TopDownKMWithoutGrid(DataObject[] data, int precision) {
        this.data = data;
        db = new DB(data);
        numObj = data.length;
        iter = 0;
        dim = data[0].coord.length;
        maxClusterID = 0;
        this.precision = precision;
        this.bestCost = 0.0;
        clusters = new Vector<Cluster>();
        noiseClusters = new Vector<Cluster>();
    }
    
    
    /**
     * Returns the cluster with the best model: Rotated or not, different distributions. Cost per Cluster
     * @param obj DataObject[]: objects belonging to the cluster
     * @return Cluster: if rotated with rotation matrix.
     */
    private Cluster bestModel(DataObject[] obj, boolean verbose){
        verbose = false;
        double gaussCost = 0.0;
        double laplaceCost = 0.0;
        if(verbose)
            System.out.println("bestModel: " + obj.length + " objects.");
        Cluster best = new Cluster();
        double addVar = 0.0;
        Cluster clNotRot = new Cluster(dim, 0);
        clNotRot.mean = mean(obj);
        clNotRot.numObj = obj.length;
        clNotRot.noiseCluster = false;
        for(int i = 0; i < dim; i++){
            clNotRot.var[i] = var(obj, clNotRot.mean, i);
            addVar += clNotRot.var[i];
        }
        for(int i = 0; i < dim; i++){
            for(int j = 0; j < obj.length; j++){
                gaussCost += gaussCost(obj[j].coord[i], clNotRot.mean[i], clNotRot.var[i]);
                laplaceCost += laplaceCost(obj[j].coord[i], clNotRot.mean[i], clNotRot.var[i]);
            }
        }
        clNotRot.cost += Math.min(gaussCost, laplaceCost);
        if(gaussCost < laplaceCost){
            clNotRot.distr = gauss;
            if(verbose)
                System.out.println("gaussian");
        } else{
            clNotRot.distr = laplace;
            if(verbose)
                System.out.println("laplacian");
        }
        //Parameter ////////////////////////
        clNotRot.cost += dim * 32 * 2 + dim * dim * 32;
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("Cost not rotated " + clNotRot.cost);
        
        gaussCost = 0.0;
        laplaceCost = 0.0;
        Cluster clRot = new Cluster(dim, 0);
        clRot.mean = median(obj);
        clRot.numObj = obj.length;
        clRot.noiseCluster = false;
        robustEstimate(clRot, obj, false);
        //transform objects to rotated space
        DataObject[] red = new DataObject[obj.length];
        Matrix m = new Matrix(clRot.dir);
        for(int l = 0; l < red.length; l++){
            double[] redCoord = new double[dim];
            Matrix x = rowVector(obj[l].coord);
            Matrix akt = x.times(m);
            for(int j = 0; j < dim; j++){
                redCoord[j] = akt.get(0,j);
            }
            DataObject redObj = new DataObject(redCoord, obj[l].number);
            redObj.clusterID = obj[l].clusterID;
            red[l] = redObj;
        }
        double[] redMeanCoord = new double[dim];
        Matrix mx = rowVector(clRot.mean);
        Matrix akt = mx.times(m);
        for(int j = 0; j < dim; j++){
            redMeanCoord[j] = akt.get(0,j);
        }
        for(int i = 0; i < dim; i++)
            clRot.var[i] = var(red, redMeanCoord, i);
        for(int i = 0; i < dim; i++){
            for(int j = 0; j < obj.length; j++){
                gaussCost += gaussCost(red[j].coord[i], redMeanCoord[i], clRot.var[i]);
                laplaceCost += laplaceCost(red[j].coord[i], redMeanCoord[i], clRot.var[i]);
            }
        }
        clRot.cost += Math.min(gaussCost, laplaceCost);
        if(gaussCost < laplaceCost)
            clRot.distr = gauss;
        else
            clRot.distr = laplace;
        //Parameter ////////////////////////
        clRot.cost += dim*32*2 + dim*dim * 32;
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("Cost rotated " + clRot.cost);
        
        //all objects noise
        if(verbose)
            System.out.println("---all objects noise---");
        Cluster allNoise = new Cluster(dim, 0);
        allNoise.noiseCluster = false;
        allNoise.mean = clNotRot.mean;
        allNoise.numObj = obj.length;
        for(int i = 0; i < dim; i++)
            allNoise.var[i] = addVar;
        for(int i = 0; i < obj.length; i++)
            for(int j = 0; j < dim; j++)
                //allNoise.cost += gaussCost(obj[i].coord[j], allNoise.mean[j], allNoise.var[j], precision);
                //TEST: echte Varianz verwenden
                allNoise.cost += gaussCost(obj[i].coord[j], allNoise.mean[j], clNotRot.var[j]);
        //TEST
        //////////////////////////
        allNoise.cost += dim * 32 * 2 + dim*dim*32;
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("Cost noise " + allNoise.cost);
        if(clRot.cost < clNotRot.cost && clRot.cost < allNoise.cost)
            return clRot;
        if(clNotRot.cost < clRot.cost && clNotRot.cost < allNoise.cost || allNoise.cost == clNotRot.cost)
            return clNotRot;
        else
            return allNoise;
    }
    
    
    /**
     * Returns the filtered clusters with the best cost. Rotated or not, splitted in cluster and noise cluster or not. Cost of resulting clusters: Cost with parameters but without ID cost.
     * @param data DataObject[]: objects to be filtered
     * @return Cluster[2]: result[0]: filtered cluster, result[1]: filtered noise, can be empty (numObj = 0).
     */
    private Cluster[] getBestFilteredClusters(DataObject[] data, boolean verbose){
        if(verbose)
            System.out.println("getBestFilteredClusters: numObj " + data.length);
        if(data.length == 1){
            if(verbose)
                System.out.println("Singleton cluster");
            Cluster[] result = new Cluster[2];
            result[0] = new Cluster(0);
            result[0].cost = dim * 32 * 2;
            result[0].numObj = 1;
            result[1] = new Cluster(1);
            //clusterID
            int[] ids = new int[1];
            ids[0] = 0;
            aktClusterID = ids;
            return result;
        }
        Cluster c = bestModel(data, true);
        //TEST
        //c.cost = Double.MAX_VALUE;
        //TEST
        if(verbose)
            System.out.println("--- All objects in cluster: " + c.cost);
        if(verbose)
            System.out.println("--filtered cluster and noise: gausssian---");
//get Filtered Clusters gauss and laplace, rotated and not rotated
        DataObject[] filteredRotatedData_g = new DataObject[data.length];
        for(int i = 0; i < data.length; i++)
            filteredRotatedData_g[i] = new DataObject(data[i]);
        KMCostWithoutGrid kmkr_g = new KMCostWithoutGrid(filteredRotatedData_g, 1, precision, true, 1);
        kmkr_g.run();
        Cluster filteredRotatedCluster_g = kmkr_g.getClusters()[0];
        filteredRotatedCluster_g.noiseCluster = false;
        Cluster filteredRotatedNoise_g = kmkr_g.getNoiseClusters()[0];
        filteredRotatedNoise_g.noiseCluster = true;
        double costFilteredRotated_g = filteredRotatedCluster_g.cost + filteredRotatedNoise_g.cost;
        double costIDClusterRot_g = 0.0;
        double costIDNoiseRot_g = 0.0;
        if(filteredRotatedCluster_g.numObj > 0 && filteredRotatedNoise_g.numObj > 0){
            costIDClusterRot_g = filteredRotatedCluster_g.numObj * Math.log((double)data.length/(double)filteredRotatedCluster_g.numObj)/Math.log(2);
            costIDNoiseRot_g = filteredRotatedNoise_g.numObj * Math.log((double)data.length/(double)filteredRotatedNoise_g.numObj)/Math.log(2);
            if(verbose)
                System.out.println("costIDClusterRot_g " + costIDClusterRot_g + " costIDNoiseRot_g " + costIDNoiseRot_g);
        }
        //Parameter ////////////////////////
        costFilteredRotated_g += dim * 32 * 2 + dim * dim * 32  + costIDClusterRot_g + costIDNoiseRot_g;
        filteredRotatedCluster_g.cost += dim * 32 * 2 + dim * dim * 32;
        //filteredRotatedNoise.cost += dim * 32 * 2;
        if(verbose)
            System.out.println("cost cluster_g: " + filteredRotatedCluster_g.cost); //Codierungskosten und Parameterkosten
        if(verbose)
            System.out.println("cost noise_g: " + filteredRotatedNoise_g.cost);//fr Noise keine Parameterkosten!!!
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("Cost filteredRotated_g " + costFilteredRotated_g); //mit ID-Kosten
        
        DataObject[] filteredData_g = new DataObject[data.length];
        for(int i = 0; i < filteredData_g.length; i++)
            filteredData_g[i] = new DataObject(data[i]);
        //FilterKM kmk_g= new FilterKM(filteredData_g, 1, precision, false, 1);
        KMCostWithoutGrid kmk_g= new KMCostWithoutGrid(filteredData_g, 1, precision, false, 1);
        kmk_g.run();
        Cluster filteredCluster_g = kmk_g.getClusters()[0];
        filteredCluster_g.noiseCluster = false;
        Cluster filteredNoise_g = kmk_g.getNoiseClusters()[0];
        filteredNoise_g.noiseCluster = true;
        double costFilteredNotRotated_g = filteredCluster_g.cost + filteredNoise_g.cost;
        double costIDCluster = 0.0;
        double costIDNoise = 0.0;
        if(filteredCluster_g.numObj > 0 && filteredNoise_g.numObj > 0){
            costIDCluster = filteredCluster_g.numObj * Math.log((double)data.length/(double)filteredCluster_g.numObj)/Math.log(2);
            costIDNoise = filteredNoise_g.numObj * Math.log((double)data.length/(double)filteredNoise_g.numObj)/Math.log(2);
            if(verbose)
                System.out.println("CostIDCluster_g " + costIDCluster + " costIDNoise_g " + costIDNoise);
        }
        //Parameter////////////////////////
        costFilteredNotRotated_g += dim * 32 * 2 + dim * dim * 32 + costIDCluster + costIDNoise;
        
        filteredCluster_g.cost += dim * 32 * 2;
        //filteredNoise_g.cost += dim * 32 * 2;
        if(verbose)
            System.out.println("cost cluster_g: " + filteredCluster_g.cost);
        if(verbose)
            System.out.println("cost noise_g: " + filteredNoise_g.cost);
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("CostfilteredNotRotated_g " + costFilteredNotRotated_g);
        
        /////////////////////////////////////////////////////////////////////////////////////////////////////////////
        if(verbose)
            System.out.println("--filtered cluster and noise: laplacian---");
        DataObject[] filteredRotatedData_l = new DataObject[data.length];
        for(int i = 0; i < data.length; i++)
            filteredRotatedData_l[i] = new DataObject(data[i]);
        //FilterKM kmkr_l = new FilterKM(filteredRotatedData_l, 1, precision, true, 2);
        KMCostWithoutGrid kmkr_l = new KMCostWithoutGrid(filteredRotatedData_l, 1, precision, true, 2);
//        //TESTKMCost
//        kmkr_l.verbose = true;
//        //TEST
        kmkr_l.run();
        Cluster filteredRotatedCluster_l = kmkr_l.getClusters()[0];
        filteredRotatedCluster_l.noiseCluster = false;
        Cluster filteredRotatedNoise_l = kmkr_l.getNoiseClusters()[0];
        filteredRotatedNoise_l.noiseCluster = true;
        double costFilteredRotated_l = filteredRotatedCluster_l.cost + filteredRotatedNoise_l.cost;
        double costIDClusterRot_l = 0.0;
        double costIDNoiseRot_l = 0.0;
        if(filteredRotatedCluster_l.numObj > 0 && filteredRotatedNoise_l.numObj > 0){
            costIDClusterRot_l = filteredRotatedCluster_l.numObj * Math.log((double)data.length/(double)filteredRotatedCluster_l.numObj)/Math.log(2);
            costIDNoiseRot_l = filteredRotatedNoise_l.numObj * Math.log((double)data.length/(double)filteredRotatedNoise_l.numObj)/Math.log(2);
            if(verbose)
                System.out.println("costIDClusterRot_l " + costIDClusterRot_l + " costIDNoiseRot " + costIDNoiseRot_l);
        }
        //Parameter ////////////////////////
        costFilteredRotated_l += dim * 32 * 2 + dim * dim * 32  + costIDClusterRot_l + costIDNoiseRot_l;
        filteredRotatedCluster_l.cost += dim * 32 * 2 + dim * dim * 32;
        //filteredRotatedNoise.cost += dim * 32 * 2;
        if(verbose)
            System.out.println("cost cluster_l: " + filteredRotatedCluster_l.cost);
        if(verbose)
            System.out.println("cost noise_l: " + filteredRotatedNoise_l.cost);
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("Cost filteredRotated " + costFilteredRotated_l);
        
        DataObject[] filteredData_l = new DataObject[data.length];
        for(int i = 0; i < filteredData_l.length; i++)
            filteredData_l[i] = new DataObject(data[i]);
        //FilterKM kmk_l = new FilterKM(filteredData_l, 1, precision, false, 2);
        KMCostWithoutGrid kmk_l = new KMCostWithoutGrid(filteredData_l, 1, precision, false, 2);
//        //TEST
//        kmk_l.verbose = true;
//        //TEST
        kmk_l.run();
        Cluster filteredCluster_l = kmk_l.getClusters()[0];
        filteredCluster_l.noiseCluster = false;
        Cluster filteredNoise_l = kmk_l.getNoiseClusters()[0];
        filteredNoise_l.noiseCluster = true;
        double costFilteredNotRotated_l = filteredCluster_l.cost + filteredNoise_l.cost;
        double costIDCluster_l = 0.0;
        double costIDNoise_l = 0.0;
        if(filteredCluster_l.numObj > 0 && filteredNoise_l.numObj > 0){
            costIDCluster_l = filteredCluster_l.numObj * Math.log((double)data.length/(double)filteredCluster_l.numObj)/Math.log(2);
            costIDNoise_l = filteredNoise_l.numObj * Math.log((double)data.length/(double)filteredNoise_l.numObj)/Math.log(2);
            if(verbose)
                System.out.println("CostIDCluster " + costIDCluster_l + " costIDNoise " + costIDNoise_l);
        }
        //Parameter////////////////////////
        costFilteredNotRotated_l += dim * 32 * 2 + costIDCluster_l + costIDNoise_l;
        filteredCluster_l.cost += dim * 32 * 2;
        //filteredNoise_l.cost += dim * 32 * 2;
        if(verbose)
            System.out.println("cost cluster: " + filteredCluster_l.cost);
        if(verbose)
            System.out.println("cost noise: " + filteredNoise_l.cost);
        ////////////////////////////////////////////
        if(verbose)
            System.out.println("CostfilteredNotRotated " + costFilteredNotRotated_l);
        
        
        double minFiltered_g = Math.min(costFilteredRotated_g, costFilteredNotRotated_g);
        double minFiltered_l = Math.min(costFilteredRotated_l, costFilteredNotRotated_l);
        double minFiltered = Math.min(minFiltered_g, minFiltered_l);
        Cluster[] result = new Cluster[2];
        int[] ids = new int[data.length];
        if(minFiltered < c.cost){
            //save current cluster IDs
            if(minFiltered_g < minFiltered_l){
                if(costFilteredRotated_g < costFilteredNotRotated_g){
                    //save current cluster IDS
                    for(int i = 0; i < ids.length; i++)
                        ids[i] = filteredRotatedData_g[i].clusterID;
                    aktClusterID = ids;
                    result[0] = filteredRotatedCluster_g;
                    result[0].distr = gauss;
                    result[1] = filteredRotatedNoise_g;
                } else{
                    for(int i = 0; i < ids.length; i++)
                        ids[i] = filteredData_g[i].clusterID;
                    aktClusterID = ids;
                    result[0] = filteredCluster_g;
                    result[0].distr = gauss;
                    result[1] = filteredNoise_g;
                }
                return result;
            } else{
                if(costFilteredRotated_l < costFilteredNotRotated_l){
                    //save current cluster IDS
                    ids = new int[filteredRotatedData_l.length];
                    for(int i = 0; i < ids.length; i++)
                        ids[i] = filteredRotatedData_l[i].clusterID;
                    aktClusterID = ids;
                    result[0] = filteredRotatedCluster_l;
                    result[0].distr = laplace;
                    result[1] = filteredRotatedNoise_l;
                } else{
                    ids = new int[filteredData_l.length];
                    for(int i = 0; i < ids.length; i++)
                        ids[i] = filteredData_l[i].clusterID;
                    aktClusterID = ids;
                    result[0] = filteredCluster_l;
                    result[0].distr = laplace;
                    result[1] = filteredNoise_l;
                }
                return result;
            }
        } else{
            //all objects in cluster, save current cluster IDS
            for(int i = 0; i < ids.length; i++)
                ids[i] = 0;
            aktClusterID = ids;
            result[0] = c;
            result[1] = new Cluster(1);
            return result;
        }
    }
    
    public void testRun(){
        Cluster[] result = getBestFilteredClusters(data, true);
        VisuSimple vk = new VisuSimple(data);
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
    }
    
    public void run(){
        Cluster[] result = getBestFilteredClusters(data, true);
        //adjust clusterIDs
        for(int i = 0; i < data.length; i++)
            data[i].clusterID = aktClusterID[i];
        bestClusters = new Cluster[1];
        bestClusters[0] = result[0];
        bestNoiseClusters = new Cluster[1];
        bestNoiseClusters[0] = result[1];
        double costFiltered = currentCost();
        double minCost = costFiltered;
        System.out.println("filtered data set " + costFiltered);
        int[] filteredID = new int[data.length];
        for(int i = 0; i < data.length; i++)
            filteredID[i] = data[i].clusterID;
        boolean improvement = true;
        //improvement = splitCluster();
        int aktTry = 0;
        Cluster[] lastLocalMinClusters = new Cluster[1];        //save Clusters and ClusterIDs of the last local minimum
        //lastLocalMinClusters[0] = bestClusters[0];
        Cluster[] lastLocalMinNoise = new Cluster[1];
        //lastLocalMinNoise[0] = bestNoiseClusters[0];
        int[] lastLocalMinIDs = new int[numObj];
        /*for(int i = 0; i < numObj; i++)
            lastLocalMinIDs[i] = data[i].clusterID;
         */
        while(improvement || aktTry < numTry){
            //improvement = splitCluster(aktTry);
            System.out.println("-------------------------------------");
            double costUnsplit = currentCost();
            int numClUnsplit = numClustersNumNoiseClusters()[0];
            int numNUnsplit = numClustersNumNoiseClusters()[1];
            System.out.println("SYMMETRIC SPLIT NUMBER: " + iter + " CURRENT COST " + costUnsplit);
            improvement = symmetricSplit(aktTry);
            if(improvement){
                minCost = currentCost();
                System.out.println("GLOBAL IMPROVEMENT: " + " CURRENT COST " + minCost + " COST OLD: " + costUnsplit);
                lastLocalMinClusters = bestClusters;
                lastLocalMinNoise = bestNoiseClusters;
                for(int i = 0; i < numObj; i++)
                    lastLocalMinIDs[i] = data[i].clusterID;
                aktTry = 0;
            } else{
                aktTry++;
                System.out.println("NO GLOBAL IMPROVEMENT: CURRENT COST " + currentCost() + " COST OLD: " + costUnsplit + " TRYING FOR ANOTHER " + (numTry - aktTry) + " TIMES");
            }
            System.out.println("OLD: " + numClUnsplit + " clusters, " + numNUnsplit + " noiseClusters.");
            System.out.println("NEW: " + numClustersNumNoiseClusters()[0] + " clusters " + numClustersNumNoiseClusters()[1] + " noiseClusters.");
            System.out.println("-------------------------------------");
            iter++;
        }
        System.out.println("done. " + iter + " iterations");
        
        //adjust clusters and IDs as from the last local minimum
        if(lastLocalMinClusters.length > 1 || lastLocalMinNoise.length > 1){
            bestClusters = lastLocalMinClusters;
            bestNoiseClusters = lastLocalMinNoise;
            for(int i = 0; i < data.length; i++)
                data[i].clusterID = lastLocalMinIDs[i];
            System.out.println("BEST: " + bestClusters.length + " clusters ");
        }
        //boolean impr = finalSplit();
        //mergeClusters();
       simpleGreedyMerge(0);
        //greedyMerge();
        minCost = currentCost();
        System.out.println("BEST COST: " + minCost);
        //testMerge();
        //FinalAssignment f = new FinalAssignment(data, bestClusters, 3);
        //f.assign();
        
//        if(costFiltered < minCost)
//            for(int i = 0; i < data.length; i++)
//                data[i].clusterID = filteredID[i];
//        /*
//        //print data of specified clusters
//        IO ea = new IO();
//        ea.printClusters(data, 4);
//        ea.printClusters(data, 2);
//        ea.printClusters(data, 0);
//        ea.printClusters(data, 10);
//         
        
////evaluate clustering result
//        int[] clusterIDs = new int[bestClusters.length];
//        for(int i = 0; i < bestClusters.length; i++)
//            clusterIDs[i] = bestClusters[i].ID;
//        ClusteringEvaluation e = new ClusteringEvaluation(data, clusterIDs);
//        e.evaluate();
//        
        VisuInfo vk = new VisuInfo(data, "result");
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
        
    }
    
    
    private int[] numClustersNumNoiseClusters(){
        int numCl = 0;
        int numN = 0;
        for(int i = 0; i < bestClusters.length; i++)
            if(bestClusters[i].numObj > 0)
                numCl++;
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                numN++;
        int[] result = new int[2];
        result[0] = numCl;
        result[1] = numN;
        return result;
    }
    
    private boolean symmetricSplit(int aktTry){
        //aktTry = numTry;
        Cluster[] currentClusters = bestClusters.clone();
        boolean[] improvement = new boolean[currentClusters.length];
        for(int i = 0; i < currentClusters.length; i++){
            improvement[i] = splitCluster(false, i);
        }
        boolean impr = false;
        for(int i = 0; i < improvement.length; i++)
            if(improvement[i])
                impr = true;
        if(!impr && aktTry < numTry){
            for(int i = 0; i < currentClusters.length; i++)
                improvement[i] = splitCluster(true, i);
        }
        return impr;
    }
    
    ///TEST
    private void checkIDs(){
        Vector ids = new Vector();
        for(int i = 0; i < data.length; i++)
            if(!ids.contains(data[i].clusterID))
                ids.add(data[i].clusterID);
        System.out.println("numIDS " + ids.size());
    }
    //TEST
    
    /**
     * Determines the next cluster to split, splits the cluster and returns if improvment of cost results from this split.
     * @param aktTry aktTry: number of trys after minimum cost has been reached
     * @return boolean: improvement of cost
     */
    private boolean splitCluster(boolean splitAnyway, int splitIndex){
        if(bestClusters[splitIndex].numObj == 1)
            return false;
        int[] currentID1 = new int[numObj]; //clusterIDs for cluster and noise cluster if filtered
        int[] currentID2;
        boolean improvement = false;
        boolean filtered1 = false; //k-means cluster 1 filtered?
        boolean filtered2 = false;
        //int splitIndex = chooseSplitCluster();
        //get all objects of the cluster to split
        double costOld = bestClusters[splitIndex].cost + bestNoiseClusters[splitIndex].cost;
        DataObject[] cobj = db.getObjects(bestClusters[splitIndex].ID); // hier obj kopieren und ClusterID auf 0 setzen!
        DataObject[] nobj = db.getObjects(bestNoiseClusters[splitIndex].ID);
        int numAll = cobj.length+nobj.length;
        DataObject[] all = new DataObject[numAll];
        for(int i = 0; i < numAll; i++)
            if(i < cobj.length)
                all[i] = cobj[i];
            else
                all[i] = nobj[i-cobj.length];
        DB kmDb = new DB(all);
        KMeans k = new KMeans(2, kmDb);
        DataObject[] newIDs = k.run(); //ClusterIDs 1 and 2
        
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = kmDb.getObjects(1);
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for(int i = 0; i < id_1.length; i++){
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap h_1 = new HashMap(id_1.length);
        for(int i = 0; i < id_1.length; i++)
            h_1.put(internalIndex_1[i], numbers_1[i]);
        for(int i = 0; i < id_1.length; i++)
            id_1[i].number = internalIndex_1[i];
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        Cluster[] clusters1 = getBestFilteredClusters(id_1, true);
        currentID1 = aktClusterID;
        clusters1[0].ID = bestClusters[splitIndex].ID;
        clusters1[1].ID = bestNoiseClusters[splitIndex].ID;
        //objects assigned to clusterID 2 by kMeans
        DataObject[] id_2 = kmDb.getObjects(2);
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for(int i = 0; i < id_2.length; i++){
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap h_2 = new HashMap(id_2.length);
        for(int i = 0; i < id_2.length; i++)
            h_2.put(internalIndex_2[i], numbers_2[i]);
        for(int i = 0; i < id_2.length; i++)
            id_2[i].number = internalIndex_2[i];
        Cluster[] clusters2 = getBestFilteredClusters(id_2, true);
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;
        clusters2[1].ID = maxClusterID + 3;
        //update current clusters
        for(int i = 0; i < bestClusters.length; i++)
            if(i != splitIndex)
                clusters.add(bestClusters[i]);
            else
                clusters.add(clusters1[0]);
        clusters.add(clusters2[0]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(i != splitIndex)
                noiseClusters.add(bestNoiseClusters[i]);
            else
                noiseClusters.add(clusters1[1]);
        noiseClusters.add(clusters2[1]);
        //boolean improvement = true;
        //boolean improvement = evaluateCost(splitIndex, cl, noise);
        if(splitAnyway)
            improvement = true;
        else
            improvement = improvement(false);
        if(improvement){
            //set new clusters and new noiseClusters as best clusters.
            Cluster[] newClusters = new Cluster[clusters.size()];
            Cluster[] newNoiseClusters = new Cluster[clusters.size()];
            for(int i = 0; i < clusters.size(); i++){
                newClusters[i] = (Cluster)clusters.elementAt(i);
                int aktID = newClusters[i].ID;
                for(int j = 0; j < noiseClusters.size(); j++){
                    if(((Cluster)noiseClusters.elementAt(j)).ID == aktID+1)
                        newNoiseClusters[i] = (Cluster)noiseClusters.elementAt(j);
                }
                if(newNoiseClusters[i] == null){
                    newNoiseClusters[i] = new Cluster();
                    newNoiseClusters[i].ID = aktID+1;
                }
            }
            for(int i = 0; i < noiseClusters.size(); i++)
                newNoiseClusters[i] = (Cluster)noiseClusters.elementAt(i);
            bestClusters = newClusters;
            bestNoiseClusters = newNoiseClusters;
            //reset bestCost
            bestCost = 0.0;
            for(int i = 0; i < bestClusters.length; i++)
                bestCost += bestClusters[i].cost;
            for(int i = 0; i < bestNoiseClusters.length; i++)
                bestCost += bestNoiseClusters[i].cost;
            
            maxClusterID = maxClusterID+2;
            int splittedID = bestClusters[splitIndex].ID;
            
            for(int i = 0; i < currentID1.length; i++){
                int index = ((Integer)h_1.get(i)).intValue();
                if(currentID1[i] == 0)
                    data[index].clusterID = splittedID;
                else
                    data[index].clusterID = splittedID+1;
            }
            for(int i = 0; i < currentID2.length; i++){
                int index = ((Integer)h_2.get(i)).intValue();
                if(currentID2[i] == 0)
                    data[index].clusterID = maxClusterID;
                else
                    data[index].clusterID = maxClusterID + 1;
            }
        }
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        if(improvement)
            return true;
        else{
            return false;
        }
    }
    
    /**
     * Try once to split all clusters and all noise clusters individually. Called after top-down has reached first cost minimum and before merging.
     * @return boolean: improvement for at least one cluster or noiseCluster
     */
    private boolean finalSplit(){
        boolean impr = false;
        Vector<Cluster> nonEmpty = new Vector<Cluster>();
        for(int i = 0; i < bestClusters.length; i++)
            if(bestClusters[i].numObj > 0)
                nonEmpty.add(bestClusters[i]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                nonEmpty.add(bestNoiseClusters[i]);
        boolean[] improvement = new boolean[nonEmpty.size()];
        Cluster[] actClusters = new Cluster[nonEmpty.size()];
        for(int i = 0; i < actClusters.length; i++)
            actClusters[i] = (Cluster)nonEmpty.elementAt(i);
        //relable noiseClusters and generate dummy noiseClusters to use the method splitCluster()
        int clusterCount = 0;
        for(int i = 0; i < actClusters.length; i++){
            int oldClusterID = actClusters[i].ID;
            if(oldClusterID != clusterCount){
                actClusters[i].ID = clusterCount;
                adjustClusterIDs(oldClusterID, clusterCount);
            }
            clusterCount = clusterCount+2;
        }
        maxClusterID = clusterCount;
        /*
        VisuSimple vk = new VisuSimple(data);
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
         */
        
        Cluster[] actNoise = new Cluster[actClusters.length];
        for(int i = 0; i < actNoise.length; i++)
            actNoise[i] = new Cluster(actClusters[i].ID + 1);
        bestClusters = actClusters;
        bestNoiseClusters = actNoise;
        for(int i = 0; i < improvement.length; i++){
            improvement[i] = splitCluster(false, i);
            if(improvement[i])
                impr = true;
        }
        if(impr){
            System.out.println("IMPROVEMENT DUE TO FINAL SPLIT. NEW: " + bestClusters.length + " clusters");
        }
        return impr;
    }
    
    ///TEST/////////////////////////////////////////////////////////
    private void testMerge(){
        Vector<Cluster> aktClusters = new Vector<Cluster>(); //current clusters and non-empty noiseClusters
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        for(int i = 0; i < bestClusters.length; i++)
            aktClusters.add(bestClusters[i]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                aktClusters.add(bestNoiseClusters[i]);
        clusters = aktClusters;
        //all clusters and noiseClusters to be merged in bestClusters, dummy noiseCluster.
        Cluster[] nonEmpty = new Cluster[aktClusters.size()];
        for(int i = 0; i < nonEmpty.length; i++)
            nonEmpty[i] = (Cluster)aktClusters.elementAt(i);
        bestClusters = nonEmpty;
        Cluster[] newNoise = new Cluster[1];
        newNoise[0] = new Cluster(1);
        newNoise[0].numObj = 0;
        bestNoiseClusters = newNoise; //dummy noisecluster
        Cluster m = merge(bestClusters[1], bestClusters[2]);
        clusters.remove(bestClusters[1]);
        clusters.remove(bestClusters[2]);
        clusters.add(m);
        boolean imp = improvement(true); //ok
        setNewBestClusters();
        adjustClusterIDs(4, 2);
        /*
        Cluster m1 = merge(bestClusters[3], bestClusters[4]);
        clusters.remove(bestClusters[3]);
        clusters.remove(bestClusters[4]);
        clusters.add(m1);
        boolean impr = improvement(true); //ok
        setNewBestClusters();
        adjustClusterIDs(5, 1);
        Cluster m2 = merge(bestClusters[0], bestClusters[2]);
        clusters.remove(bestClusters[0]);
        clusters.remove(bestClusters[2]);
        clusters.add(m2);
        boolean imprr = improvement(true); //ok
         
         */
        //System.out.println("mist");
    }
    ///TEST
    
    private void setNewBestClusters(){
        Cluster[] newBest = new Cluster[clusters.size()];
        Cluster[] newBestNoise = new Cluster[clusters.size()];
        for(int i = 0; i < clusters.size(); i++){
            newBest[i] = (Cluster)clusters.elementAt(i);
            newBestNoise[i] = new Cluster(-1);    //dummy noise fr merge
        }
        bestNoiseClusters = newBestNoise;
        bestClusters = newBest;
    }
    
    private void adjustClusterIDs(int oldID, int newID){
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == oldID)
                data[i].clusterID = newID;
        
    }
    
    private void mergeClusters(){
        int numRandom = 1000;
        Random r = new Random(2);
        Cluster c;
        Cluster d;
        Vector<Cluster> aktClusters = new Vector<Cluster>(); //current clusters and non-empty noiseClusters
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        for(int i = 0; i < bestClusters.length; i++)
            aktClusters.add(bestClusters[i]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                aktClusters.add(bestNoiseClusters[i]);
        clusters = aktClusters;
        //all clusters and noiseClusters to be merged in bestClusters, dummy noiseCluster.
        Cluster[] nonEmpty = new Cluster[aktClusters.size()];
        for(int i = 0; i < nonEmpty.length; i++)
            nonEmpty[i] = (Cluster)aktClusters.elementAt(i);
        bestClusters = nonEmpty;
        Cluster[] newNoise = new Cluster[1];
        newNoise[0] = new Cluster(1);
        newNoise[0].numObj = 0;
        bestNoiseClusters = newNoise; //dummy noisecluster
        for(int i = 0; i < numRandom; i++){
            if(bestClusters.length > 1){
                //merge random clusters and noiseClusters
                int nextMerge1 = r.nextInt(bestClusters.length);
                int nextMerge2 = r.nextInt(bestClusters.length);
                while(nextMerge2 == nextMerge1)
                    nextMerge2 = r.nextInt(bestClusters.length);
                //if(bestClusters[nextMerge2].ID == bestClusters[nextMerge1].ID+1)       //do not try to merge cluster with its own noise cluster.
                //  nextMerge2++;
                //objects with this clusterID need to be relabeled, since the merged cluster gets the minimumID of c and d.
                c = bestClusters[nextMerge1];
                d = bestClusters[nextMerge2];
                int maxID = Math.max(c.ID, d.ID);
                int minID = Math.min(c.ID, d.ID);
                Cluster m = merge(c, d);
                clusters.remove(c);
                clusters.remove(d);
                clusters.add(m);
                if(improvement(false)){
                    //adjust clusterIDs
                    adjustClusterIDs(maxID, minID);
                    //set new clusters as best clusters
                    setNewBestClusters();
                } else{
                    clusters.remove(m);
                    clusters.add(c);
                    clusters.add(d);
                }
            }
            //clusters.removeAllElements();
            //noiseClusters.removeAllElements();
        }
    }
    
    public void simpleGreedyMerge(int numTry){
        System.out.println("Merging clusters...");
        Vector<Cluster> aktClusters = new Vector<Cluster>(); //current clusters and non-empty noiseClusters
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        for(int i = 0; i < bestClusters.length; i++)
            aktClusters.add(bestClusters[i]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0){
            //bestNoiseClusters[i].cost += dim * 32 * 2 + dim * dim * 32;
            bestNoiseClusters[i].cost += dim * 32 * 2;
            aktClusters.add(bestNoiseClusters[i]);
            }
        clusters = aktClusters;
        bestCost = currentCost();
        boolean improvement = true;
        int aktTry = -1;
        //save Clusters and IDs of the last cost minimum
        int[] lastLocalMinIDs = new int[numObj];
        for(int i = 0; i < numObj; i++)
            lastLocalMinIDs[i] = data[i].clusterID;
        Cluster[] lastLocalMinClusters = new Cluster[clusters.size()];
        for(int i = 0; i < lastLocalMinClusters.length; i++)
            lastLocalMinClusters[i] = (Cluster)clusters.elementAt(i);
        while(improvement && clusters.size() > 1){
            Cluster[] startClusters = new Cluster[clusters.size()];
            for(int i = 0; i < startClusters.length; i++)
                startClusters[i] = (Cluster)clusters.elementAt(i);
            //determine the clusters with the maximum saved Cost
            double maxSaved = -Double.MAX_VALUE;
            Cluster[] toMerge = new Cluster[3];
            for (int i = 0; i < clusters.size(); i++){
                for(int j = 0; j < clusters.size(); j ++){
                    if(i != j){
                        Cluster a = (Cluster)clusters.elementAt(i);
                        Cluster b = (Cluster)clusters.elementAt(j);
                        Cluster m = merge(a, b);
                        clusters.remove(a);
                        clusters.remove(b);
                        clusters.add(m);
                        double aktSaved = savedCost(true);
                        System.out.println("try to merge clusters " + a.ID + " and " + b.ID + " resulting cluster " + m.ID + " savedCost: " + aktSaved);
                        if(aktSaved > maxSaved){
                            toMerge[0] = a;
                            toMerge[1] = b;
                            toMerge[2] = m;
                            maxSaved = aktSaved;
                        }
                        clusters.removeAllElements();
                        for(int k = 0; k < startClusters.length; k++)
                            clusters.addElement(startClusters[k]);
                    }
                }
            }
            clusters.removeAllElements();
            for(int k = 0; k < startClusters.length; k++)
                clusters.addElement(startClusters[k]);
            if(maxSaved > 0){
                System.out.println("merging clusters " + toMerge[0].ID + " " +  toMerge[1].ID + " saved cost: " + maxSaved);
                int maxID = Math.max(toMerge[0].ID, toMerge[1].ID);
                int minID = Math.min(toMerge[0].ID, toMerge[1].ID);
                adjustClusterIDs(maxID, minID);
                clusters.remove(toMerge[0]);
                clusters.remove(toMerge[1]);
                clusters.add(toMerge[2]);
                setNewBestClusters();
                bestCost = currentCost();
                aktTry = 0;
                //save IDs
                for(int i = 0; i < numObj; i++)
                    lastLocalMinIDs[i] = data[i].clusterID;
                lastLocalMinClusters = bestClusters;
            } else{
                improvement = false;
                if(aktTry < numTry){
                    System.out.println("merging clusters " + toMerge[0].ID + " " +  toMerge[1].ID + " saved cost: " + maxSaved);
                    int maxID = Math.max(toMerge[0].ID, toMerge[1].ID);
                    int minID = Math.min(toMerge[0].ID, toMerge[1].ID);
                    adjustClusterIDs(maxID, minID);
                    clusters.remove(toMerge[0]);
                    clusters.remove(toMerge[1]);
                    clusters.add(toMerge[2]);
                    aktTry++;
                    //TEST
                    setNewBestClusters();
                    bestCost = currentCost();
                }
            }
        }
        for(int i = 0; i < data.length; i++)
            data[i].clusterID = lastLocalMinIDs[i];
        bestClusters = lastLocalMinClusters;
    }
    
    //nicht fertig
    public void greedyMerge(){
        Vector<Cluster> aktClusters = new Vector<Cluster>(); //current clusters and non-empty noiseClusters
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        for(int i = 0; i < bestClusters.length; i++)
            aktClusters.add(bestClusters[i]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                aktClusters.add(bestNoiseClusters[i]);
        clusters = aktClusters;
        boolean improvement = true;
        //determine cost table
        Cluster[] intermediate = new Cluster[aktClusters.size()];
        double[][] savedCost = new double[aktClusters.size()][aktClusters.size()];
        for (int i = 0; i < aktClusters.size(); i++){
            for(int j = 0; j < aktClusters.size(); j ++){
                if(i != j){
                    for(int k = 0; k < intermediate.length; k++)
                        intermediate[k] = (Cluster)aktClusters.elementAt(k);
                    Cluster a = (Cluster)aktClusters.elementAt(i);
                    Cluster b = (Cluster)aktClusters.elementAt(j);
                    Cluster m = merge(a, b);
                    aktClusters.remove(a);
                    aktClusters.remove(b);
                    aktClusters.add(m);
                    savedCost[i][j] = savedCost(false);
                    aktClusters.removeAllElements();
                    for(int k = 0; k < intermediate.length; k++)
                        aktClusters.add(intermediate[k]);
                } else
                    savedCost[i][j] = 0.0;
            }
        }
        while(improvement && aktClusters.size() > 1){
            //determine the index of the clusters to be merged next, i.e. the index with maximum saved cost.
            double maxSaved = -Double.MAX_VALUE;
            int i_index = -1;
            int j_index = -1;
            for(int i = 0; i < savedCost.length; i++){
                for(int j = 0; j < savedCost.length; j++){
                    if(i != j && savedCost[i][j] > maxSaved){
                        maxSaved = savedCost[i][j];
                        i_index = i;
                        j_index = j;
                    }
                }
            }
            if(maxSaved < 0)
                improvement = false;
            else{
                //merge these clusters and update aktClusters and the cost table: insert new cluster at the position of minIndex in aktClusters and in the cost table
                Cluster a = (Cluster)(aktClusters.elementAt(i_index));
                Cluster b = (Cluster)(aktClusters.elementAt(j_index));
                Cluster m = merge(a, b);
                aktClusters.remove(a);
                aktClusters.remove(b);
                aktClusters.add(Math.min(i_index, j_index), m);
                //adjust clusterIDs
                int minID = Math.min(i_index, j_index);
                int maxID = Math.max(i_index, j_index);
                adjustClusterIDs(maxID, minID);
                setNewBestClusters();
                //determine cost of the new cluster to all other clusters
                double[] mCosts = new double[aktClusters.size()+1];
                Cluster[] temp = new Cluster[aktClusters.size()]; //save current Clusters
                for(int i = 0; i < temp.length; i++)
                    temp[i] = (Cluster)(aktClusters.elementAt(i));
                int indexNew = Math.min(i_index, j_index);
                int indexRemoved = Math.max(i_index, j_index);
                for(int i = 0; i < aktClusters.size(); i++){
                    Cluster akt = (Cluster)aktClusters.elementAt(i);
                    Cluster merged = merge(m, akt);
                    aktClusters.remove(m);
                    aktClusters.remove(akt);
                    aktClusters.add(merged);
                    mCosts[i] = savedCost(false);
                    aktClusters.removeAllElements();
                    for(int j = 0; j < temp.length; j++)
                        aktClusters.addElement(temp[j]);
                }
                mCosts[indexRemoved] = -Double.MAX_VALUE; //this cluster does not exist anymore
                //update cost array
                
                for(int i = 0; i < aktClusters.size(); i++)
                    for(int j = 0; j < aktClusters.size(); j++){
                    savedCost[indexRemoved][j] = -Double.MAX_VALUE;
                    savedCost[j][indexRemoved] = -Double.MAX_VALUE;
                    savedCost[indexNew][j] = mCosts[j];
                    savedCost[j][indexNew] = mCosts[j];
                    }
            }
        }
    }
    
    private Cluster merge(Cluster c, Cluster d){
        boolean verbose = false;
        if(verbose)
            System.out.println("merge clusters: " + c.ID + " " + d.ID);
        //determine best model: rotated or not and TODO: best distribution
        DataObject[] mObj = getMergedObjects(c, d);
        int minID = Math.min(c.ID, d.ID);
        int maxID = Math.max(c.ID, d.ID);
        if(minID == maxID)
            System.err.println("trying to merge clusters with the same id: " + minID);
        Cluster m = bestModel(mObj, true);
        m.ID = minID;
        return m;
    }
    
    
    private DataObject[] getMergedObjects(Cluster c, Cluster d){
        //DataObject[] members = new DataObject[c.numObj + d.numObj];
        DataObject[] members = new DataObject[countObj(c.ID, d.ID)];
        int counter = 0;
        for(int j = 0; j < data.length; j++)
            if(data[j].clusterID == c.ID || data[j].clusterID == d.ID){
            members[counter] = data[j];
            counter++;
            }
        return members;
    }
    
    public int countObj(int cId, int dId){
        int counter = 0;
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == cId || data[i].clusterID == dId)
                counter++;
        return counter;
    }
    
    /**
     * Returns the current cost of bestClusters and bestNoiseClusters including ID cost
     * @return double: cost
     */
    private double currentCost(){
        boolean verbose = true;
        if(verbose)
            System.out.println("currentCost");
        double cost = 0.0;
        for(int i = 0; i < bestClusters.length; i++){
            Cluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Cluster " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                cost += costID;
                cost += bestClusters[i].cost;
            }
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            Cluster akt = bestNoiseClusters[i];
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Noise " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                cost += costID;
                cost += bestNoiseClusters[i].cost;
            }
        }
        return cost;
    }
    
    /**
     * Checks if the costs are lower after one split. Cost for clusterID are counted here.
     * @return boolean: improvement of cost
     */
    private boolean improvement(boolean verbose){
        double costNew = 0.0;
        double costOld = 0.0;
        double costIDNew = 0.0;
        double costIDOld = 0.0;
        int clusterCountNew = 0;
        int clusterCountOld = 0;
        if (verbose)
            System.out.println("NEW:");
        for(int i = 0; i < clusters.size(); i++){
            Cluster akt = (Cluster)clusters.elementAt(i);
            if(clusters.elementAt(i).numObj > 0){
                clusterCountNew ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                costNew += akt.cost;
            } else
                if (verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < noiseClusters.size(); i++){
            Cluster akt = (Cluster)noiseClusters.elementAt(i);
            if(noiseClusters.elementAt(i).numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                clusterCountNew ++;
                costNew += akt.cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        if(verbose)
            System.out.println("OLD:");
        for(int i = 0; i < bestClusters.length; i++){
            Cluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                clusterCountOld ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            Cluster akt = bestNoiseClusters[i];
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                clusterCountOld ++;
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestNoiseClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        if(costNew < costOld){
            if(verbose)
                System.out.println("improvement : cost old: " + costOld + " " + clusterCountOld + " clusters, cost new: " + " " + costNew + " " + clusterCountNew + " clusters.");
            return true;
        }
        
        else{
            if(verbose)
                System.out.println("no improvement : cost old: " + costOld + " " + clusterCountOld + " clusters, cost new: " + " " + costNew + " " + clusterCountNew + " clusters.");
            return false;
        }
    }
    
    
    private double savedCost(boolean verbose){
        double costNew = 0.0;
        double costOld = 0.0;
        double costIDNew = 0.0;
        double costIDOld = 0.0;
        int clusterCountNew = 0;
        int clusterCountOld = 0;
        double savedCost;
        if (verbose)
            System.out.println("NEW:");
        for(int i = 0; i < clusters.size(); i++){
            Cluster akt = (Cluster)clusters.elementAt(i);
            if(clusters.elementAt(i).numObj > 0){
                clusterCountNew ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                costNew += akt.cost;
            } else
                if (verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < noiseClusters.size(); i++){
            Cluster akt = (Cluster)noiseClusters.elementAt(i);
            if(noiseClusters.elementAt(i).numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                clusterCountNew ++;
                costNew += akt.cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        if(verbose)
            System.out.println("OLD:");
        for(int i = 0; i < bestClusters.length; i++){
            Cluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                clusterCountOld ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            Cluster akt = bestNoiseClusters[i];
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                clusterCountOld ++;
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestNoiseClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        savedCost = costOld - costNew;
        return savedCost;
    }
    
    
    //Index von zu splittendem Cluster
    private int chooseSplitCluster(){
        double maxCost = 0.0;
        int max = bestClusters.length;
        for(int i = 0; i < bestClusters.length; i++){
            double aktCost = bestClusters[i].cost + bestNoiseClusters[i].cost;
            double aktNum = bestClusters[i].numObj + bestNoiseClusters[i].numObj;
            double costPerObject = aktCost/aktNum;
            if(costPerObject > maxCost){
                maxCost = costPerObject;
                max = i;
            }
        }
        System.out.println("split cluster " + bestClusters[max].ID);
        return max;
    }
    
    //BIC
    private double paramCost(Cluster c){
        double cost = 0.0;
        //if(c.numObj == 0)
        //  return 1.0;
        if(c.dir != null)
            return ((dim + dim*dim)/2) * Math.log10(c.numObj);
        else
            return (dim/2) * Math.log10(c.numObj);
    }
    
    private void robustEstimate(Cluster c, DataObject[] data, boolean verbose){
        if(verbose)
            System.out.println("Cluster " + c.ID);
        DataObject[] obj = data;
        double[] median = median(obj);
        c.mean = median;
        double[][] covarm = covarByMedian(data, median);
        Matrix covar = new Matrix(covarm);
        EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
        Matrix vectors = eig.getV();
        Matrix values = eig.getD();
        if(verbose){
            System.out.println("EW");
            values.print(5, 3);
            System.out.println("EV");
            vectors.print(5, 3);
        }
        c.dir = vectors.getArrayCopy();
        double maxEigen = -Double.MAX_VALUE ;
        for (int i=0 ; i<dim ; i++)
            maxEigen = Math.max(maxEigen, values.get(i,i));
        for (int i=0 ; i<dim ; i++)
            values.set(i,i,1/Math.max(values.get(i,i), maxEigen*0.001)) ;
        c.invDir =  (vectors.times(values).times(vectors.transpose())).getArrayCopy() ;
    }
    
    //var of coordinate numCoord
    private double var(DataObject[] o, double[] mean, int numCoord){
        double var = 0.0;
        for(int i = 0; i < o.length; i++)
            var += (o[i].coord[numCoord] - mean[numCoord]) * (o[i].coord[numCoord] - mean[numCoord]);
        return var/(double)o.length;
    }
    
    private double variance(DataObject[] o, double[] mean){
        double var = 0.0;
        for(int i = 0; i < o.length; i++)
            var += o[i].distance(mean) * o[i].distance(mean);
        return var/(double)o.length;
    }
    
    private double gaussCost(double x, double mean, double var){
        double stdv = Math.sqrt(var);
        //double pdf = 0.5*Math.log(2*Math.PI*var)*ln10+Math.pow(x-mean,2)/2/var/ln2;
        double pdf = (0.5*stdv*Math.sqrt(2*Math.PI)) * Math.exp(-0.5*Math.pow(((x-mean)/stdv),2));
          // return Math.max(0.0,(0.5*Math.log(2*Math.PI*va)+precision*ln10+Math.pow(x-mean,2)/2/va)/ln2 ); //alt
        //double pdf =  (0.5*Math.log(2*Math.PI*var)*ln10+Math.pow(x-mean,2)/2/var)/ln2;
        return -Math.log(pdf)/Math.log(2);
    }
    
    private double laplaceCost(double x, double mean, double var){
        double stdv = Math.sqrt(var);
        double b = stdv/Math.sqrt(2);
        double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
        return -Math.log(pdf)/Math.log(2);
    }
    
    private double[] mean(DataObject[] o){
        double[] mean = new double[o[0].coord.length];
        for(int i = 0; i < o.length; i++)
            for(int j = 0; j < mean.length; j++){
            mean[j] += o[i].coord[j];
            }
        for(int j = 0; j < mean.length; j++)
            mean[j] /= o.length;
        return mean;
    }
    
    public double [] median(DataObject[] members) {
        int numObj= members.length ;
        int dim = members[0].coord.length ;
        double[] result = new double[dim];
        double [] sortarray = new double[numObj] ;
        for (int i=0 ; i<dim ; i++) {
            double median = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sortarray[j]=members[j].coord[i] ;
            Arrays.sort(sortarray) ;
            if (numObj==1)
                result[i]=sortarray[0] ;
            else if (numObj%2 == 0)
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
            else
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
        }
        return result ;
    }
    
    public double[][] covarByMedian(DataObject[] data, double[] mean) {
        int numObj= data.length ;
        int dim = data[0].coord.length ;
        double [] sortarray = new double[numObj] ;
        double[][] cov = new double[dim][dim];
        if (numObj <15){
            for (int i=0 ; i<dim ; i++)
                cov[i][i]=1.0 ;
            return cov ;
        }
        for(int l1 = 0; l1 < dim; l1++) {
            for(int l2 = 0; l2 < dim; l2++) {
                for(int j = 0; j < numObj; j++)
                    sortarray[j] = (data[j].coord[l1]-mean[l1])*(data[j].coord[l2]-mean[l2]) ;
                
                Arrays.sort(sortarray) ;
                if (numObj%2 == 0)
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3
                            ;
            }
            
        }
        double maxdiff = 0.0 ;
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                sum += Math.abs(cov[i][j]) ;
            double diff = 1.1*sum - 2.1*cov[i][i] ;
            if (diff>maxdiff)
                maxdiff = diff ;
        }
        if (maxdiff > 0.0)
            for (int i=0 ; i<dim ; i++)
                cov[i][i] += maxdiff ;
        
        return cov ;
    }
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
}

