/*
 * Clustering.java
 *
 * Created on 17 October 2005, 10:51
 */

package compr;

/**
 *
 * @author Administrator
 */
public class Clustering {
    Cluster[] clusters;
    DataObject[] data;
    int numObj;
    int numClusters;
    int dim;
    double cost;
    
    /** Creates a new instance of Clustering */
    public Clustering() {
    }
    
    public Clustering(Cluster[] clusters, DataObject[] data){
        this.clusters = clusters;
        this.cost = cost;
        this.numClusters = clusters.length;
        this.data = data;
        numObj = data.length;
        dim = data[0].coord.length;
    }
    
    public double cost(){
        double sum = 0.0;
        for(int i = 0; i < clusters.length; i++)
            sum+= setCost(clusters[i], getObjects(clusters[i]));
        //Kosten fr Outlier : Outlier haben ClusterID 0
        for(int i = 0; i < data.length; i++){
            if (data[i].clusterID == 0)
                for(int j = 0; j < dim; j++)
                    sum += logStar(data[i].coord[j], 1);
        }
        
        return sum;
    }
    
    public void clusterTest(){
        for(int i = 0; i < clusters.length; i++)
            test(clusters[i], getObjects(clusters[i]));
    }
    
    public void test(Cluster c, DataObject[] members){
        double[] mean = new double [c.dim];
        double[] std = new double[c.dim];
        double sumGauss;
        double sumNorm;
        double[] ub = new double[dim];
        for(int i = 0; i < ub.length; i++)
            ub[i] = 0.0;
        double[] lb = new double[dim];
        for(int i = 0; i < lb.length; i++)
            lb[i] = Double.MAX_VALUE;
        if(c.subspaceCluster){
            double[][] transCoord = new double[c.numObj][dim];
            for(int i = 0; i < members.length; i++){
                for(int j = 0; j < dim; j++){
                    for(int k = 0; k < dim; k++){
                        transCoord[i][j] += (members[i].coord[j]-c.mean[j]) * c.eigenVector[j][k];
                    }
                }
            }
            
            for(int i = 0; i < members.length; i++)
                for(int j = 0; j < c.dim; j++)
                    mean[j] += transCoord[i][j];
            for(int i = 0; i < mean.length; i++)
                mean[i]/= members.length;
            double[] var = new double[dim];
            for(int i = 0; i < members.length; i++){
                for(int j = 0; j < dim; j++)
                var[j] += (transCoord[i][j] - mean[j]) * (transCoord[i][j] - mean[j]);
                
            }
            for(int i = 0; i < dim; i++)
                std[i] = Math.sqrt(var[i]/members.length);
            for(int i = 0; i < members.length; i++){
                for(int j = 0; j < c.dim; j++){
                    if(transCoord[i][j] > ub[j])
                        ub[j] = transCoord[i][j];
                    if(transCoord[i][j] < lb[j])
                        lb[j] = transCoord[i][j];
                }
            
            }
            System.out.println("Cluster " + c.ID);
            for(int i = 0; i < c.dim; i++){
                sumGauss = 0.0;
                for(int j = 0; j < members.length; j++)
                    sumGauss += gaussCost(transCoord[j][i], mean[i], std[i], 2);
                System.out.println("Koordinate " + i + " gaussCost " + sumGauss);
            }
            for(int i = 0; i < c.dim; i++){
                sumNorm = 0.0;
                for(int j = 0; j < members.length; j++)
                    sumNorm += uniformCost(transCoord[j][i], lb[i], ub[i], 2);
                System.out.println("Koordinate " + i + " normCost " + sumNorm);
            }
            System.out.println();
            
        } else{
            for(int i = 0; i < members.length; i++)
                for(int j = 0; j < c.dim; j++)
                    mean[j] += members[i].coord[j];
            for(int i = 0; i < mean.length; i++)
                mean[i]/= members.length;
            double[] var = new double[dim];
            for(int i = 0; i < members.length; i++){
                for(int j = 0; j < dim; j++)
                var[j] += (members[i].coord[j] - mean[j]) * (members[i].coord[j] - mean[j]);
                
            }
            for(int i = 0; i < dim; i++){
                std[i] = Math.sqrt(var[i]/members.length);
         System.out.println("Cluster " + c.ID + " Koordinate " + i + " stdv " + std[i]);   
            }
            for(int i = 0; i < members.length; i++){
                for(int j = 0; j < c.dim; j++){
                    if(members[i].coord[j] > ub[j])
                        ub[j] = members[i].coord[j];
                    if(members[i].coord[j] < lb[j])
                        lb[j] = members[i].coord[j];
                }
            }
            System.out.println("Cluster " + c.ID);
            for(int i = 0; i < c.dim; i++){
                sumGauss = 0.0;
                for(int j = 0; j < members.length; j++)
                    sumGauss += gaussCost(members[j].coord[i], mean[i], std[i], 5);
                System.out.println("Koordinate " + i + " gaussCost " + sumGauss);
            }
            for(int i = 0; i < c.dim; i++){
                sumNorm = 0.0;
                for(int j = 0; j < members.length; j++)
                    sumNorm += uniformCost(members[j].coord[i], lb[i], ub[i], 5);
                System.out.println("Koordinate " + i + " normCost " + sumNorm);
            }
            System.out.println();
        }
        
    }
    
    private double gaussCost(double x, double mean, double std, int precision){
        double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;
        return Math.log(1/prob)/Math.log(2);
    }
    
    private double uniformCost(double x, double lb, double ub, int precision){
        double prob ;
        if (Math.abs(ub-lb) < Math.pow(10, -precision))
            prob = 1 ;
        else
            prob = Math.pow(10, -precision) / Math.abs(ub - lb) ;
        return Math.log(1/prob)/Math.log(2);
    }
    
    private double setCost(Cluster c, DataObject[] members){
        double cost = 0.0;
        //Kosten fr Mittelwert
        double costMean = 0.0;
        for(int i = 0; i < c.mean.length; i++)
            costMean += logStar(c.mean[i], 1);
        cost += costMean;
        // Kosten fr ClusterID
        cost += members.length * Math.log(clusters.length)/Math.log(2);
        // Kosten normaler Cluster
        if(!c.subspaceCluster){
            //Kosten fr Koordinaten
            for(int i = 0; i < members.length; i++)
                for(int j = 0; j < dim; j++)
                    cost += logStar((members[i].coord[j]-c.mean[j]), 1);
        }
        //Kosten subspace cluster
        else{
            //Kosten fr EV
            for(int i = 0; i < dim; i++){
                for(int j = 0; j < dim; j++)
                    if(c.eigenVector != null && c.eigenVector[i][j] != 0.0)
                        cost+= logStar(c.eigenVector[i][j], 3);
            }
            
            double[][] transCoord = new double[c.numObj][dim];
            for(int i = 0; i < members.length; i++){
                for(int j = 0; j < dim; j++){
                    for(int k = 0; k < dim; k++){
                        transCoord[i][j] += (members[i].coord[j]-c.mean[j]) * c.eigenVector[j][k];
                    }
                }
            }
            //Kosten fr Koordinaten
            for(int i = 0; i < transCoord.length; i++)
                for(int j = 0; j < transCoord[i].length; j++)
                    cost += logStar(transCoord[i][j], 1);
        }
        System.out.println("Cluster " +  c.ID + " , subspace Cluster " + c.subspaceCluster +  " , Kosten " + cost);
        c.cost = cost;
        return cost;
    }
    
    
    public double logStar(double cc, int precision){
        double c = Math.pow(10.0, precision) * Math.abs(cc) ;
        double result = 0.0;
        // while (c > 0){
        c = Math.log(c)/Math.log(2);
        if(c > 0)
            result += c;
        //}
        return result;
    }
    
    private DataObject[] getObjects(Cluster c){
        DataObject[] members = new DataObject[c.numObj];
        int counter = 0;
        for(int j = 0; j < data.length; j++)
            if(data[j].clusterID == c.ID){
            members[counter] = data[j];
            counter++;
            }
        return members;
    }
    
    private double distance(double[] coord, double[] mean){
        double sum = 0.0;
        for (int i = 0; i < coord.length; i++){
            double dist = ((coord[i] - mean[i])*(coord[i] - mean[i]));
            sum = sum + dist;
        }
        return Math.sqrt(sum);
    }
}
