/*
 * CovSim.java
 *
 * Created on 19 October 2005, 11:10
 */

package compr;
import Jama.*;
import java.util.*;

/**
 *
 * @author Administrator
 */
public class CovSim {
    DataObject[] data;
    int numObj;
    int dim;
    Matrix[] covMatrix;
    int sampleSize;
    int k;
    
    /** Creates a new instance of CovSim */
    public CovSim(DataObject[] data) {
        this.data = data;
        this.numObj = data.length;
        this.dim = data[0].coord.length;
        sampleSize = data.length/4;
        k = 15;
        this.covMatrix = new Matrix[sampleSize];
        for(int i = 0; i < sampleSize; i++)
            covMatrix[i] = new Matrix(dim, dim);
    }
    
    public void setSimMatrix(){
        DataObject[] sampleData = new DataObject[sampleSize];
        Random r = new Random();
        for(int i = 0; i < sampleSize; i++){
            sampleData[i] = data[r.nextInt(numObj)];
            DataObject[] knnS = knn(data, sampleData[i], k);
            DataObject[] neighbors = new DataObject[k+1];
            for(int j = 0; j < k; j++)
                neighbors[j] = knnS[j];
            neighbors[k] = sampleData[i];
            double[] median = median(neighbors);
            double[][] covar = covarByMedian(neighbors, median);
            Matrix cm = new Matrix(covar);
            System.out.println("DO " + sampleData[i].number);
            //cm.print(5, 3);
            covMatrix[i] = cm;
        }
        for(int i = 0; i < sampleSize; i++){
            for(int j = 0; j < sampleSize; j++){
                if(i != j)
                    System.out.println("DO " + sampleData[i].number + " DO " + sampleData[j].number + " " + getDistance(covMatrix[i], covMatrix[j]));
            }
            System.out.println();
        }
        int[] clusterID = clusterMatrix(10);
        for(int i = 0; i < sampleSize; i++){
            System.out.println("DO " + sampleData[i].number + " " + clusterID[i]);
        }
    }
        
        private int[] clusterMatrix(int numCluster){
            System.out.println("Clustering Matrix");
            int[] clusterID = new int[sampleSize];
            Matrix[] means = new Matrix[numCluster];
            for(int i = 0; i < numCluster; i++)
                means[i] = new Matrix(dim, dim);
            Random r = new Random();
            for(int i = 0; i < numCluster; i++)
                means[i] = covMatrix[r.nextInt(sampleSize)];
            for(int i = 0; i < sampleSize; i++){
                int id = assignObject(covMatrix[i], means);
                clusterID[i] = id;
            }
            means = updateMeans(clusterID, numCluster);
            boolean converged = false;
            int iter = 0;
            while(!converged){
               
                iter++;
                boolean clusterChanged = false;
                for(int i = 0; i < sampleSize; i++){
                    int id = assignObject(covMatrix[i], means);
                    if(id != clusterID[i])
                        clusterChanged = true;
                    clusterID[i] = id;
                    means = updateMeans(clusterID, numCluster);
                }
                if(!clusterChanged)
                    converged = true;
            }
            System.out.println("iter " + iter);
            return clusterID;
        }
        
        private int assignObject(Matrix akt, Matrix[] means){
            //System.out.println("assign Object");
            double minDist = Double.MAX_VALUE;
            int minIndex = means.length;
            for(int i = 0; i < means.length; i++){
                double aktDist = getDistance(akt, means[i]);
                if(aktDist < minDist){
                    minDist = aktDist;
                    minIndex = i;
                }
            }
            return minIndex;
        }
        
        private Matrix[] updateMeans(int[] clusterID, int numCluster){
            //System.out.println("update Means");
            Matrix[] means = new Matrix[numCluster];
            for(int i = 0; i < numCluster; i++)
                means[i] = new Matrix(dim, dim);
            for(int i = 0; i < numCluster; i++){
                for(int j = 0; j < sampleSize; j++)
                    if (clusterID[j] == i)
                        means[i] = means[i].plus(covMatrix[j]);
            }
            for(int i = 0; i < numCluster; i++)
                means[i].timesEquals(1.0/(double)sampleSize);
            return means;
        }
        
        
        private double getDistance(Matrix ma, Matrix mb){
            double[][] a = ma.getArrayCopy();
            double[][] b = mb.getArrayCopy();
            double[][] cora = new double[dim][dim];
            double[][] corb = new double[dim][dim];
            for(int l = 0; l < dim; l++)
                for(int m = 0; m < dim; m++){
                cora[l][m]= a[l][m]/(Math.sqrt(a[l][l] * a[m][m]));
                corb[l][m]= b[l][m]/(Math.sqrt(b[l][l] * b[m][m]));
                
                }
            double dist = 0.0;
            for(int i = 0; i < dim; i++)
                for(int j = 0; j < dim; j++)
                    dist += Math.pow(cora[i][j]-corb[i][j], 2);
            return dist;
        }
        
        private DataObject[] knn(DataObject[] data, DataObject d, int k){
            double[] erg = new double[k];        //Array der Grsse MinPts anlegen
            for(int i=0; i<k; i++) {
                erg[i] = Double.MAX_VALUE;  //alle auf Max_Value setzen.
            }
            DataObject[] result = new DataObject[k];
            for(int i=0; i < data.length; i++) {   //Durchlauf alle Datenobjekte
                DataObject akt = data[i];
                if(akt.number != d.number){
                    double dist = d.distance(data[i]); //Distanz zum i-ten Datenobjekt
                    int c = k-1;
                    while(c>=0 && dist<erg[c]) { //MinPts >= 0 && Distanz zum i-ten Datenobjekt < erg[MinPts-1]
                        if(c<k-1){
                            erg[c+1] = erg[c];
                            result[c+1] = result[c];
                        }
                        erg[c] = dist;
                        result[c] = data[i];
                        c--;                  //MinPts um 1 reduzieren
                    }
                }
            }//i
            return result;
        }
        
        private double[] median(DataObject[] members) {
            int num= members.length ;
            int d = members[0].coord.length ;
            double[] med = new double[d];
            double [] sortarray = new double[num] ;
            for (int i=0 ; i<d ; i++) {
                double median = 0.0 ;
                for (int j=0 ; j<num ; j++)
                    sortarray[j]=members[j].coord[i] ;
                Arrays.sort(sortarray) ;
                if (num==1)
                    med[i]=sortarray[0] ;
                else if (num%2 == 0)
                    med[i]=(sortarray[num/2-1]+sortarray[num/2])/2 ;
                else
                    med[i]=(sortarray[num/2-1]+sortarray[num/2]+sortarray[num/2+1])/3 ;
            }
            return med;
        }
        
        private double[][] covarByMedian(DataObject[] data, double[] median) {
            int num= data.length ;
            int d = data[0].coord.length ;
            double [] sortarray = new double[num] ;
            double[][] cov = new double[d][d];
            if (num < k){
                for (int i=0 ; i<d ; i++)
                    cov[i][i]=1.0 ;
                return cov ;
            }
            for(int l1 = 0; l1 < d; l1++) {
                for(int l2 = 0; l2 < d; l2++) {
                    for(int j = 0; j < num; j++)
                        sortarray[j] = (data[j].coord[l1]-median[l1])*(data[j].coord[l2]-median[l2]) ;
                    
                    Arrays.sort(sortarray) ;
                    if (num%2 == 0)
                        
                        cov[l1][l2]=(sortarray[num/2-1]+sortarray[num/2])/2 ;
                    else
                        
                        cov[l1][l2]=(sortarray[num/2-1]+sortarray[num/2]+sortarray[num/2+1])/3;
                }
                
            }
            double maxdiff = 0.0 ;
            for (int i=0 ; i<d ; i++){
                double sum = 0.0 ;
                for (int j=0 ; j<d ; j++)
                    sum += Math.abs(cov[i][j]) ;
                double diff = 1.1*sum - 2.1*cov[i][i] ;
                if (diff>maxdiff)
                    maxdiff = diff ;
            }
            if (maxdiff > 0.0)
                for (int i=0 ; i<d ; i++)
                    cov[i][i] += maxdiff ;
            
            return cov ;
        }
        
        private double[][] setCovariance(DataObject[] data, double[] mean){
            double[][] cov = new double[dim][dim];
            for(int l = 0; l < data.length; l++)
                for(int i = 0; i < dim; i++)
                    for(int j = 0; j < dim; j++)
                        cov[i][j]+= (data[l].coord[j]-mean[j])*(data[l].coord[i]-mean[i]);
            for(int i = 0; i < dim; i++)
                for(int j = 0; j < dim; j++)
                    cov[i][j]/= data.length;
            return cov;
        }
    }
