package de.lmu.outlier.utils;

import java.util.ArrayList;
import java.util.Hashtable;

public class DataObject
  implements Cloneable
{
  public double[] coordinates;
  public int clusterID;
  public int classID;
  public int hierarchyLevel;
  public double codingCost;
  public Integer number;
  boolean isCore;
  public ArrayList<Integer> NNList = new ArrayList();

  public ArrayList<Integer> NoNNList = new ArrayList();

  public Hashtable<Integer, Double> clusterSize2Costs = new Hashtable();

  public Hashtable<Integer, Double> clusterCostMin = new Hashtable();

  public Hashtable<Integer, Double> clusterCostDeviation = new Hashtable();

  public DataObject(double[] k, int num)
  {
    this.coordinates = k;
    this.number = Integer.valueOf(num);
    this.clusterID = 0;
    this.isCore = false;
  }

  public DataObject(DataObject d)
  {
    this.coordinates = d.coordinates;
    this.clusterID = d.clusterID;
    this.classID = d.classID;
    this.number = d.number;
    this.isCore = false;
  }

  public DataObject(double[] k, int num, int classID) {
    this.classID = classID;
    this.coordinates = k;
    this.number = Integer.valueOf(num);
    this.clusterID = 0;
    this.isCore = false;
  }

  public DataObject(double[] k, int num, int classID, int clusterID) {
    this.classID = classID;
    this.clusterID = clusterID;
    this.coordinates = k;
    this.number = Integer.valueOf(num);
    this.isCore = false;
  }

  public DataObject(double[] k)
  {
    this.coordinates = k;
    this.isCore = false;
  }

  public Object clone() {
    try {
      return super.clone();
    } catch (CloneNotSupportedException e) {
      throw new InternalError();
    }
  }

  public double[] getCoord()
  {
    return this.coordinates;
  }

  public String getObjectInfo()
  {
    StringBuffer sb = new StringBuffer();
    sb.append("<html> data object: " + this.number + "<br> coordinates: (");
    for (int i = 0; i < this.coordinates.length - 1; ++i)
      sb.append(this.coordinates[i] + ", ");
    sb.append(this.coordinates[(this.coordinates.length - 1)] + ") <br>");
    sb.append("classID " + this.classID + "<br>");
    sb.append("clusterID " + this.clusterID + "<br>");
    sb.append("hierarchy level: " + this.hierarchyLevel + "<br>");
    sb.append("coding cost: " + this.codingCost + "<br>");
    sb.append("</font>");
    sb.append("</html>");
    return sb.toString();
  }

  public double logStar(int cc)
  {
    double c = Math.abs(cc);
    double result = 0.0D;

    c = Math.log(c) / Math.log(2.0D);
    if (c > 0.0D) {
      result += c;
    }
    return result;
  }

  public double distance(double[] mean)
  {
    double sum = 0.0D;
    for (int i = 0; i < this.coordinates.length; ++i) {
      double dist = (this.coordinates[i] - mean[i]) * (this.coordinates[i] - mean[i]);
      sum += dist;
    }
    return Math.sqrt(sum);
  }

  public double distance(DataObject b)
  {
    DataObject other = b;
    DataObject thisDo = this;
    double sum = 0.0D;
    for (int i = 0; i < b.coordinates.length; ++i) {
      double dist = Math.abs(other.coordinates[i] - thisDo.coordinates[i]);
      sum += dist;
    }
    return sum;
  }

  public double euclideanDistance(DataObject b) {
    DataObject other = b;
    DataObject thisDo = this;
    double sum = 0.0D;
    for (int i = 0; i < b.coordinates.length; ++i) {
      double dist = (other.coordinates[i] - thisDo.coordinates[i]) * (other.coordinates[i] - thisDo.coordinates[i]);
      sum += dist;
    }
    return Math.sqrt(sum);
  }

  public double vectorLength()
  {
    double length = 0.0D;
    double[] tCoord = getCoord();
    double sum = 0.0D;
    for (int i = 0; i < tCoord.length; ++i) {
      sum += tCoord[i] * tCoord[i];
    }
    return Math.sqrt(sum);
  }

  public String toString()
  {
    StringBuffer str = new StringBuffer("[");
    for (double coordi : this.coordinates) str.append(coordi + ",");
    str.append("] " + this.number);
    return str.toString();
  }

  public int hashCode()
  {
    int prime = 31;
    int result = 1;
    result = 31 * result + this.number.intValue();
    return result;
  }

  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (super.getClass() != obj.getClass())
      return false;
    DataObject other = (DataObject)obj;

    return (this.number == other.number);
  }
}