package de.lmu.synchronization.outlier;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

import de.lmu.outlier.utils.*;


/**
 * Outlier Detection based on local synchronization
 * Extensive Kuramoto model based synchronization
 * 
 * @author Junming Shao
 * @param None
 * Created on 16. Feb. 2010, 
 */


public class Outlier{
	
    public static final double C = 1;
	private static final String ArrayList = null;
    public double A;
    public static String fname;
    public int num;
    public int dim;
    public boolean normFlag;
    public boolean noiseFlag;
    public ArrayList src; 
    int loopNum;
    int[][] biGraph;
    double[] kernelP;
    int ID;
    double[][] datan;
    double[][] data;
    int nf;
    int f;  //loop k
    double[] mdl;
    int lp;
    double minMDL;
    int nc;
    int[] countN; 
    int[][] count;
    int[] idx;
    double[][] order;
    
    
    public Outlier() {
    	A = 0.0;
    	num = 0;
    	dim = 1; 
    	normFlag = false;
    	noiseFlag = false;
    	nf = 0;
    	lp  = 0;
    	nc = 1;
    	src = new ArrayList();
    	mdl = new double[100];
    	for(int i=0;i<100;i++){
    		mdl[i] = 0.0;
    	}
    	minMDL = Double.MAX_VALUE;
    }
    /**
     * Set the related parameters
     * @param filename, the threshold for nearest neighbors of one object
     * @return none
     */  
    public void setParams(String filename, double e){ //, boolean Flag,boolean noise){
    	fname = filename;
    	A = e;	
    	//normFlag = Flag;
    	//noiseFlag = noise;
    }
    /**
     * Load the database from a given filename
     * @param Filename
     * @return none
     */
    public void loadData(String fn){
            
        try {
        	File mFile = new File(fn);
        	FileReader fr = new FileReader(mFile);
        	BufferedReader br = new BufferedReader(fr);
        	String line;
        	
        	while((line=br.readLine())!=null){
        		line = line.trim();        		
        		String[] str = line.split("\\s+");
        		//System.out.println(num);
        		dim = str.length;
        		ArrayList temp = new ArrayList(dim);    		
        		for(int i=0;i<dim;i++){
        			temp.add(Double.parseDouble(str[i]));
        		}
    			src.add(temp);
    			num = num + 1;
        	}
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        
        //Normalize the data set
        data = new double[num][dim];
        countN = new int[num];
        count = new int[num][50];
        for(int i=0;i<num;i++){
        	countN[i] = 0;
      	  for(int j=0;j<dim;j++){	  
      		data[i][j] = (Double)((ArrayList)src.get(i)).get(j); 
      	  }
         }
    }

    public void norm1(ArrayList data){
       int n = data.size();
       int d = ((ArrayList)data.get(0)).size();
       double [] max = new double[d];
       //Initialize
       for(int i=0;i<d;i++){
    	   max[i]= -1e10;
       }
       //Find the maximum value for each variance

       for(int j=0;j<d;j++){
    	  for(int i=0;i<n;i++){
    		  double temp = (Double)((ArrayList)data.get(i)).get(j);
    		  if(temp > max[j]){
    			  max[j]= temp;
    		  }    		  
    	  }
       }
       //Normalize the data set
       for(int i=0;i<n;i++){
     	  for(int j=0;j<d;j++){
     		  double temp = (Double)((ArrayList)data.get(i)).get(j);
     		  temp = temp/max[j];
     		 ((ArrayList)data.get(i)).set(j, temp);
     	  }
        }	
       //saveData(data, "normx.txt");
    }    
    
    /**
     * save the clustering result
     * @param ArrayList data (result)
     * @param String fn (filename)
     * @return none
     */
    public void norm(ArrayList data){
       int n = data.size();
       int d = ((ArrayList)data.get(0)).size();
       double [] max = new double[d];
       double [] min = new double[d];
       //Initialize
       for(int i=0;i<d;i++){
    	   max[i]= -Double.MAX_VALUE;
    	   min[i]= Double.MAX_VALUE;
       }
       //Find the maximum value for each variance

       for(int j=0;j<d;j++){
    	  for(int i=0;i<n;i++){
    		  double temp = (Double)((ArrayList)data.get(i)).get(j);
    		  if(temp > max[j]){
    			  max[j]= temp;
    		  }    		  
    		  if(temp < min[j]){
    			  min[j]= temp;
    		  } 
    	  }
       }
       //Normalize the data set
       datan = new double[n][dim];
       for(int i=0;i<n;i++){
     	  for(int j=0;j<d;j++){
     		  double temp = (Double)((ArrayList)data.get(i)).get(j);
     		  temp = (temp-min[j])/(max[j]-min[j]);     		  
     		 ((ArrayList)data.get(i)).set(j, temp);
     		datan[i][j] = temp;
     	  }
        }	
              
    }   
    
    /**
     * save the clustering result
     * @param ArrayList data (result)
     * @param String fn (filename)
     * @return none
     */
    public void saveData(ArrayList data, String fn){
        
    	int lens = data.size();
        try{
			FileOutputStream fout = new FileOutputStream(new File(fn));     
			for(int i=0;i<lens;i++){   
				for(int j=0; j<dim;j++){   
					fout.write(((Double)(((ArrayList)data.get(i)).get(j))+"\t").getBytes());   
				}   
			fout.write(("\r\n").getBytes());   
			}   
        } catch (IOException ex) {
            ex.printStackTrace();
        }    		
    }
    
	public void save1D(int[] data, String fn){
	    
		int row = data.length;
		
	    try{
			FileOutputStream fout = new FileOutputStream(new File(fn)); 
			//fout.write(("Data ID\t Cluster ID").getBytes());
			//fout.write(("\r\n").getBytes());
			for(int i=0;i<row;i++){   
			fout.write(((Integer)(data[i])+"\t").getBytes());    
			fout.write(("\r\n").getBytes());   
			}   
	    } catch (IOException ex) {
	        ex.printStackTrace();
	    }    		
	}
	
	public void save1D(double[] data, String fn){
	    
		int row = data.length;
		
	    try{
			FileOutputStream fout = new FileOutputStream(new File(fn)); 
			//fout.write(("Data ID\t Cluster ID").getBytes());
			//fout.write(("\r\n").getBytes());
			for(int i=0;i<row;i++){   
			fout.write(((Double)(data[i])+"\t").getBytes());    
			fout.write(("\r\n").getBytes());   
			}   
	    } catch (IOException ex) {
	        ex.printStackTrace();
	    }    		
	}
	
	public void save2D(double[][] data, String fn){
	    
		int row = data.length;
		
	    try{
			FileOutputStream fout = new FileOutputStream(new File(fn)); 
			//fout.write(("Data ID\t Cluster ID").getBytes());
			//fout.write(("\r\n").getBytes());
			for(int i=0;i<row;i++){ 
				for(int j=0;j<data[0].length;j++){
					fout.write(((Double)(data[i][j])+"\t").getBytes());  
				}
			fout.write(("\r\n").getBytes());   
			}   
	    } catch (IOException ex) {
	        ex.printStackTrace();
	    }    		
	}
	
	public void save2D(int[][] data, String fn){
	    
		int row = data.length;
		
	    try{
			FileOutputStream fout = new FileOutputStream(new File(fn)); 
			//fout.write(("Data ID\t Cluster ID").getBytes());
			//fout.write(("\r\n").getBytes());
			for(int i=0;i<row;i++){ 
				for(int j=0;j<data[0].length;j++){
					fout.write(((Integer)(data[i][j])+"\t").getBytes());  
				}
			fout.write(("\r\n").getBytes());   
			}   
	    } catch (IOException ex) {
	        ex.printStackTrace();
	    }    		
	}	
    /**
     * compute the distance
     * @param double dis
     * @return none
     */    
    public double distance(double[] dis){
    	
    	double val = 0.0;
    	for(int i=0;i<dis.length;i++){
    		val = val + dis[i]*dis[i];
    	}   
    	double dist = Math.sqrt(val);
    	return dist;
    }
    
    public double[][] ArrayListtoDouble(ArrayList src){
    	int num = src.size();
    	int dim = ((ArrayList)src.get(0)).size();
    	
    	double[][] data = new double[num][dim];
    	
    	for(int i=0;i< num; i++){
    		for(int j=0;j<dim;j++){
    			data[i][j] = (Double)(((ArrayList)src.get(i)).get(j));
    		}
    	}
    	
    	return data;
    }
    
    
    
    public void cluster(){
    	
    	int len = src.size();
    	int dimm = dim;
    	System.out.println("Number of Objects:" + len + "; Dimensionality: " + dimm);
    	boolean loop = true;
    	loopNum = 0;
    	double localOrder = 0.0;
    	double allorder = 0.0;
    	ArrayList x = src;
    	saveData(x, "synC0.txt");
    	ArrayList prex = new ArrayList();
    	double[] rr = new double[50];
//    	double[][] pp = new double[x.size()][50];
//    	for(int i=0;i<50;i++)
//    		rr[i] = 0.0;
    	
    	//saveData(x, "synC"+String.valueOf(loopNum)+".txt");
    	//Copy data set
    	
    	for(int i=0;i<len;i++){
    		ArrayList temp1 = new ArrayList(dimm);    		
    		for(int j=0;j<dimm;j++){
    			temp1.add(((ArrayList)x.get(i)).get(j));
//    			pp[i][loopNum] = (Double)((ArrayList)x.get(i)).get(j);
    		}
			prex.add(temp1);
			
    	} 

    	//System.out.println(((ArrayList)x.get(0)).get(0));
    	order = new double[50][len];
    	while(loop){

    		
    		localOrder = 0.0;
    		allorder = 0.0;
    		
    		loopNum = loopNum + 1;
    		
	    	for(int i=0;i<len;i++){	    		
	    		double[] sinValue = new double[dimm]; 
    			double[] diss = new double[dimm]; 
    			double[] temp = new double[dimm];
    			
    			double dis = 0.0;
    			double kij = 0.0;
    			double allk = 0.0;
    			
	    		int num = 0; double sita = 0.0;  double ls =0.0;
    			ArrayList diff = new ArrayList();
			
	    		for(int j=0;j<len;j++){	    			
	    			dis = 0.0; kij = 0.0; 
	    			for(int d=0;d<dimm;d++){
	    				//diss[d] = (Double)((ArrayList)(x.get(j))).get(d)-(Double)((ArrayList)(x.get(i))).get(d);
	    				diss[d] = (Double)((ArrayList)(prex.get(j))).get(d)-(Double)((ArrayList)(prex.get(i))).get(d);
	    				
	    			}
	    			
	    			dis = distance(diss);  
	    			//System.out.println(dis);
	    			
	    			if(dis < A){	    		
	    				num = num + 1;
	    				//Calculating the coupling strength
		    			for(int d=0;d<dimm;d++){
		    				//temp[d] = (diss[d]+ 1e-10)/((Double)((ArrayList)(x.get(j))).get(d)+ 1e-10);
		    				temp[d] = (diss[d]+ 1e-10)/((Double)((ArrayList)(prex.get(j))).get(d)+ 1e-10);
		    			}	    	
		    			// Weighted coupling strength of each object
	    				//kij = Math.pow(distance(temp),1);
	    				//allk = allk + kij;	
	    				
		    			for(int d=0;d<dimm;d++){
		    				//sinValue[d] = sinValue[d] + kij*Math.sin(diss[d]);		 
		    				//sinValue[d] = sinValue[d] + Math.sin(diss[d]);
		    				sinValue[d] = sinValue[d] + Math.abs(diss[d]);
		    			}	 
	    				sita = sita + Math.exp(-dis);
	    				ls = ls + Math.cos(dis);
	    			}
	    		}
	    		countN[i] = countN[i]+ (num-1);
	    		count[i][loopNum-1] = num - 1;
	    		if(num > 2){
	    			for(int d=0;d<dimm;d++){
	    				double[] x1 = new double[dimm];
	    				x1[d] = ((Double)((ArrayList)(prex.get(i))).get(d))+ ((C/num)*sinValue[d]);
		    			((ArrayList)(x.get(i))).set(d, x1[d]);    			
	    			}
	    			//order[loopNum-1][i] = sita/num;
	    			order[loopNum-1][i] = ls/num;
	    			//System.out.println("order i"+ order[loopNum-1][i]);
	    		}else{
	    			order[loopNum-1][i] = 0;
	    		}
	    	}
	    	
	    	
	    	
	    	for(int k = 0; k < len; k++){
	    		allorder = allorder + order[loopNum-1][k];	    		
	    	}
	    	//Local order parameter
	    	localOrder = allorder/len;
	    	rr[loopNum-1] = localOrder;
	    	
	    	saveData(x, "synC"+String.valueOf(loopNum)+".txt");
	    	
	    	
	    	//System.out.println(Double.toString(localOrder));
	    	if( localOrder > 1 - (1e-3)|| loopNum >= 50){
	    		loop = false;
	    		System.out.println("LOCAL ORDER PARAMETER:"+Double.toString(localOrder));
	    		System.out.println("ALL LOOPS:"+Double.toString(loopNum));
	    	}
	    	// Save the previous value of the objects of the data set
	    	
	    	for(int i=0;i<len;i++){
	    		ArrayList temp1 = new ArrayList(dimm);    		
	    		for(int j=0;j<dimm;j++){
	    			temp1.add(((ArrayList)x.get(i)).get(j));
//	    			pp[i][loopNum] = (Double)((ArrayList)x.get(i)).get(j);
	    		}
				prex.set(i, temp1);
	    	}	   	
	    	
	    	
//	    	saveData(x, "synC.txt");
	    	this.save1D(rr, "r.txt");
//	    	this.save2D(pp, "pp.txt");
    	}
    	this.save2D(order, "LSF.txt");
    	idx = findSynCluster(x);
    	    	
    }
    
    public int[] findSynCluster(ArrayList x){
    	int len = x.size();
    	int[] id = new int[len];
    	int[] f = new int[len];
    	for(int i=0;i<len;i++){
    		id[i] = -1;
    		f[i] = 0;
    	}
    	ID = 0;
    	nf = 0;
    	for(int i=0;i<len;i++){
    		if(f[i]==0){   
    			//ID = ID + 1;
    			double[] diss = new double[dim];
    			int num  =0;
    			ArrayList al = new ArrayList();
	    		for(int j=0;j<len;j++){
	    			if(i!=j){
			    		for(int d=0;d<dim;d++){
			    			diss[d] = (Double)((ArrayList)(x.get(j))).get(d)-(Double)((ArrayList)(x.get(i))).get(d);			    			
			    		}
		    			double dis = distance(diss);
		    			//System.out.println("Distance:"+dis);
		    			if(Math.abs(dis)<(Math.sqrt(dim)*1e-3)){	   				    				
		    				num = num + 1;
		    				al.add(j);
		    			}	    			
//		    			if(num>0 && Math.abs(dis)<(Math.sqrt(dim)*1e-2)){
//		    				f[j] = 1;
//		    				id[j] = ID;	    				   						    			
//		    			}
	    			}
	    		}
	    		if(num>2){	    			
	    			f[i] = 1;
	    			for(int k=0;k<al.size();k++){
	    				f[(Integer)al.get(k)] = 1;
	    				id[(Integer)al.get(k)] = ID; 
	    			}
	    			id[i] = ID;
	    			ID = ID + 1;
	    		}else{
	    			nf = 1;
	    			if(num>1){
		    			for(int k=0;k<al.size();k++){
		    				//f[(Integer)al.get(k)] = 1;
		    				id[(Integer)al.get(k)] = -1; 
		    			}
	    			}
	    			id[i] = -1;
	    		}
    		}
    	}
    		
    	return id;
    	
    }
    
        
    public double[] std(double[][] data){
    	double[] avg = new double[data[0].length];
    	double[] std = new double[data[0].length];
    	for(int i=0;i<data[0].length;i++){
    		avg[i] = 0;
    	}
    	
    	for(int j=0;j<data[0].length;j++){
    		for(int i=0;i<data.length;i++){    		
    			avg[j] = avg[j] + data[i][j];
    		}
    		avg[j] = avg[j]/data.length;
    		double s = 0.0;
    		for(int i=0;i<data.length;i++){    		
    			s = s + (data[i][j]-avg[j])*(data[i][j]-avg[j]);
    		}
    		std[j] = Math.sqrt(((s+0.0)/(data.length-1)));

    	}
    	
    	return std;
    }
    

    public double dist(ArrayList al1, ArrayList al2){
    	double res = 0.0;
    	
    	if(al1.size()!=al2.size()) return 0.0;
    	else{
    		double[] diss = new double[al1.size()];
			for(int d=0;d<al1.size();d++){
				diss[d] = (Double)(al1.get(d))-(Double)(al2.get(d));		
			}
			res = distance(diss);
    	}
    	
    	return res;
    	
    }
    public double kNN(String fn, int k){
    	
    	  double res = 0.0;
    	 

    	  
    	  // check inputs for validity
    	  if(src.size() == 0 || k<=0) return 0.0; // bail on bad input
    	 
    	  for(int s=0;s<src.size();s++){  
    		  
        	  double[] d = new double[k];
        	  int n = 0; // number of element in the res
    		  
	    	  double dd = dist((ArrayList)src.get(s),(ArrayList)src.get(0)); // load first one into list
	    	  d[n++] = dd;
	    	 
	    	  // go through all other data points
	    	  for(int i = 1; i<src.size(); i++){
	    		dd = dist((ArrayList)src.get(s),(ArrayList)src.get(i));  
	    	    if( n<k || d[k-1] > dd){ //add one
	    	      if(n<k){
	    	    	  d[n++] = dd;
	    	      }
	    	      int j = n-1;
	    	      while(j>0 && dd < d[j-1]){ // slide big data up
	    	        d[j] = d[j-1]; 
	    	        j--;
	    	      }
	    	      d[j] = dd;
	    	    }
	    	  }
	    	  
	    	  res = res + d[k-1];
	    	  
    	  }
    	  return res/src.size();
    	}

    double NN(ArrayList al){
    	double s = 0.0;
    	int n = al.size();
    	if(n<1)
    		return s;
    	
    	for(int i=0;i<n;i++){
    		double[] t1 = datan[(Integer)al.get(i)];
    		for(int j=i+1;j<n;j++){
    			double[] t2 = datan[(Integer)al.get(j)];
    			double[] t = new double[dim];
    			for(int d=0;d<dim;d++){
    				t[d] = t1[d]-t2[d];
    			}
    			s = s + distance(t);
    		}
    	}
    	
    	return s;
    	
    }
    
    
    /**
     * main function - Dynamical clustering
     */
    public static void main(String[] args) {
    	
    	long   start   =   System.currentTimeMillis(); 
    	//for(int ll=40;ll<=40;ll++){
    		
    	String fn = "nba.txt"; 
    	Outlier myCluster = new Outlier(); 
    	myCluster.loadData(fn);
    	
    	//myCluster.estimationP(myCluster.data); // Test
    	
    	//====Normalize the data
    	myCluster.norm(myCluster.src);
    	
    	int kk = 40;
    	
    	double K0 = myCluster.kNN(fn,kk);
    	    	
    	myCluster.f = 1;
    	double K = K0;
    	//int n = 0;
		System.out.println("=============Plerform the clustering on K = " + K +" ============");
    	myCluster.setParams(fn,K);	    	
    	myCluster.cluster(); 
    	int n = myCluster.num;
    	double[] lsf = new double[n];
    	double[] len = new double[myCluster.ID];
    	int[] nn  = new int[n];
    	int[] id = new int[n];

    	//if(ll==20){
    	myCluster.save1D(id, "id.txt");	
    	myCluster.save1D(myCluster.idx, "idx.txt");	
    	myCluster.save1D(myCluster.countN, "countN.txt");
    	System.out.println(n);
    	//}
//    	int[] nx = new int[myCluster.ID];
////    	for(int i=0;i<n;i++){
////    		for(int j=0;j<=myCluster.ID;j++){
////    			if(myCluster.idx[i] == j){
////    				nx[j]++;
////    			}
////    		}
////    		
////    		for(int k=0;k<n;k++){
////    			if(myCluster.idx[k]==myCluster.idx[i]){
////    				double[] dd = new double[myCluster.dim];
////    				for(int d=0;d<myCluster.dim;d++){
////    					dd[d] = myCluster.datan[i][d]-myCluster.datan[k][d];
////    				}
////    				len[myCluster.idx[k]] += myCluster.distance(dd);
////    			}
////    		}
////    		
////    		System.out.println("len= " + len[i]);
////    		
////    	}
//    	
//    	
//		for(int j=0;j<myCluster.ID;j++){	
//			ArrayList temp = new ArrayList();
//			for(int i=0;i<n;i++){
//				if(myCluster.idx[i] == j){
//					temp.add(i);
//				}
//			}
//			nx[j] = temp.size();
//			len[j]=2*myCluster.NN(temp)/(temp.size());
//			System.out.println("len =" + len[j]+ "; size= " + nx[j]);
//		}
//			
//		
//		
//    	
//    	
//    	
//    	for(int i=0;i<n;i++){
//    		int numI = 0;
//    		int np = 0;
//    		for(int j=0;j<myCluster.ID;j++){
//    			if(myCluster.idx[i] == j){
//    				numI = nx[j];
//    			}
//    		}
//    		for(int j=0;j<myCluster.loopNum;j++){
//    			if(myCluster.count[i][j]<(numI-1)){
//    				count[i] += myCluster.count[i][j];
//    				np++;
//    			}
//    		}
//    		System.out.println("NUMI =" + numI);
//    		if(numI>0)
//    			count[i] /=(len[myCluster.idx[i]]*np);
//    		
//    		else
//    			count[i] = 0;
//    	}
//    	
//    	
////    	for(int i=0;i<n;i++){    		
////    		for(int j=i+1;j<n;j++){
////    			int temp=0;
////    			if(count[i]<count[j]){
////	    			temp = count[i];
////	    			count[i]=count[j];
////	    			count[j] = temp;
////    			}
////    		}
////    	}
    	
    	//double minF = Double.MAX_VALUE;
    	for(int i=0;i<n;i++){
    		int l = 0;
    		for(int j=0;j<myCluster.order.length;j++){
    			if(myCluster.order[j][i]<1.0){
    				lsf[i]+= myCluster.order[j][i];
    				l=l+1;
    			}
    				
    		}
    		lsf[i] /=l;
    		//if(count[i]<minF) minF = count[i];
    	}
    	
    	//myCluster.save1D(lsf, "LSF_K" + ll*2 + ".txt");
    	myCluster.save1D(lsf, "LSF_nba_K" + kk + ".txt");
    	
    	//Order the lsf value    
    	double[][] lsforder = new double[n][2];
    	int[] ind = new int[n];
    	
		for(int i=0;i<n;i++){ 
			ind[i] = i;
		}
		for(int i=0;i<n;i++){    
			for(int j=i+1;j<n;j++){
				double temp=0;
				int temp1 = 0;
				if(lsf[i]>lsf[j]){
					temp = lsf[i];
					lsf[i]=lsf[j];
					lsf[j] = temp;
					temp1 =ind[i];
					ind[i] = ind[j];
					ind[j] = temp1;
				}
			}
		}
    	
		for(int i=0;i<n;i++){   
			lsforder[i][0] = ind[i];
			lsforder[i][1] = lsf[i];
		}
    	
		myCluster.save2D(lsforder, "LSForder.txt");	
    	myCluster.save2D(myCluster.count,"count.txt");
    	
    	//System.out.println("num:"+myCluster.data.length+" , dim" + myCluster.data[0].length);
    	
    	OutlierPlot plot = new OutlierPlot(lsf,nn,id,myCluster.data,myCluster.idx,myCluster.ID);
    	plot.setTitle("Outlier Factor Plot");
        plot.setSize(1300, 850);
        plot.setLocation(10, 10);
        plot.setVisible(true);     
 
    	
  	  	long   end   =   System.currentTimeMillis(); 
  	  	System.out.println("All running time:"+Long.toString((end-start)/1000)+" s.");      
  	  	System.out.println("Ending");	
    }
}
