/*
 * Epdg.java
 *
 * Created on 15. Juni 2007, 13:37
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */
package epdgenerator;

import java.util.*;
import umontreal.iro.lecuyer.util.Num;

/**
 *
 * @author boehm
 */
public class Epdg {

    Random r;
    double[] quantiles;
    int numQ = 100;

    /** Creates a new instance of Epdg */
    public Epdg(double beta) {
        double step = 1.0 / 100000;
        r = new Random();
        quantiles = new double[numQ];
        double x = 0.0;
        double sum = 0.0;
        int curQ = 1;
        while (sum < 0.5 - 0.5 / numQ) {
            sum += gg(x, 0.0, 1.0, beta) * step;
            if (sum > (double) curQ / numQ / 2) {
                quantiles[curQ] = x;
                curQ++;
            //System.out.println (curQ+", "+x);
            }
            x += step;
        }
    }

    public Epdg() {

    }

    double nextEpd() {
        double unif = r.nextDouble();
        return icdf(unif);
    }

    double icdf(double x) {
        double sign = x > 0.5 ? 1.0 : -1.0;
        double abs = Math.abs(x - 0.5);
        int qIndex = (int) (abs * 2 * numQ);
        if (qIndex >= numQ - 1) {
            qIndex = numQ - 2;
        }
        double weight = (abs * 2 * numQ) - qIndex;
        return sign * ((1.0 - weight) * quantiles[qIndex] + weight * quantiles[qIndex + 1]);

    }

    double gg(double x, double mu, double sigma, double beta) {
        double t2;
        double t10;

        t2 = 1 / sigma;
        t10 = Math.pow(Math.abs((x - mu) * t2), 2 / (1 + beta));
        return (omega(beta) * t2 * Math.exp(-c(beta) * t10));
    }

    /* The options were    : operatorarrow */
    double omega(double beta) {
        double t9;
        double t3;

        t3 = Math.pow(Math.exp(Num.lnGamma(0.15E1 + 0.15E1 * beta)), 0.5);
        t9 = Math.pow(Math.exp(Num.lnGamma(0.5 + 0.5 * beta)), -0.15E1);
        return (t3 / (1 + beta) * t9);
    }

    double c(double beta) {
        return (Math.pow(Math.exp(Num.lnGamma(0.15E1 + 0.15E1 * beta)) / Math.exp(Num.lnGamma(0.5 + 0.5 * beta)), 1 / (1 +
                beta)));
    }

    public static double[][] writeCluster(int n, double[] beta, double[][] matrix, double[] mu) {
        int dim = beta.length;
        double[][] minMax = new double[dim][2]; //save for each dimension minimum and maximum for noise generation.
        for (int i = 0; i < dim; i++) {
            minMax[i][0] = Double.MAX_VALUE; //minimum
            minMax[i][1] = -Double.MAX_VALUE; //maximum
        }
        if (matrix.length != dim || mu.length != dim) {
            System.out.println("Error.");
        }
        Epdg[] pdf = new Epdg[dim];
        for (int i = 0; i < dim; i++) {
            pdf[i] = new Epdg(beta[i]);
        }
        for (int i = 0; i < n; i++) {
            double[] vec = new double[dim];
            for (int j = 0; j < dim; j++) {
                double x = pdf[j].nextEpd();
                for (int k = 0; k < dim; k++) {
                    vec[k] += matrix[j][k] * x;
                }
            }
            for (int j = 0; j < dim; j++) {
                vec[j] += mu[j];
            }
            for (int j = 0; j < dim; j++) {
                System.out.print(vec[j] + " ");
                if (vec[j] < minMax[j][0]) {
                    minMax[j][0] = vec[j];
                }
                if (vec[j] > minMax[j][1]) {
                    minMax[j][1] = vec[j];
                }
            }
            System.out.println();
        }
        return minMax;
    }

    public static void main(String[] args) {
//SYNTHETIC EXAMPLE SIGMOD 08
//        int dim = 2;
//        double[][] minMax = new double[dim][2]; //save for each dimension minimum and maximum for noise generation.
//        for (int i = 0; i < dim; i++) {
//            minMax[i][0] = Double.MAX_VALUE; //minimum
//            minMax[i][1] = -Double.MAX_VALUE; //maximum
//        }
//        //3 Cluster
//        double[] beta1 = {-0.85, 3.8};
//        double[][] m1 = {{2.5, 2.0}, {0.5, 2.0}};
//        double[] mu1 = {30.0, 25.0};
//        double[][] c = writeCluster(500, beta1, m1, mu1);
//        //double[][] c = writeCluster(16000, beta1, m1, mu1) ;
//        for (int j = 0; j < dim; j++) {
//            if (c[j][0] < minMax[j][0]) {
//                minMax[j][0] = c[j][0];
//            }
//            if (c[j][1] > minMax[j][1]) {
//                minMax[j][1] = c[j][1];
//            }
//        }
//        double[] beta2 = {-0.5, 0.5};
//        double[][] m2 = {{2.0, 0.0}, {0.8, 2.0}};
//        double[] mu2 = {50.0, 40.0};
//        c = writeCluster(500, beta2, m2, mu2);
//        //c = writeCluster(16000, beta2, m2, mu2) ;
//        for (int j = 0; j < dim; j++) {
//            if (c[j][0] < minMax[j][0]) {
//                minMax[j][0] = c[j][0];
//            }
//            if (c[j][1] > minMax[j][1]) {
//                minMax[j][1] = c[j][1];
//            }
//        }
//        double[] beta3 = {4.5, 5.3};
//        double[][] m3 = {{2.0, 1.0}, {0.2, 2.0}};
//        double[] mu3 = {50.0, 15.0};
//        c = writeCluster(500, beta3, m3, mu3);
//        //c = writeCluster(16000, beta3, m3, mu3) ;
//        for (int j = 0; j < dim; j++) {
//            if (c[j][0] < minMax[j][0]) {
//                minMax[j][0] = c[j][0];
//            }
//            if (c[j][1] > minMax[j][1]) {
//                minMax[j][1] = c[j][1];
//            }
//        }
//        //noise
//        //System.out.println("NOISE");
//        double[] betaNoise = {-0.97, -0.22297};
//        double varNoise = 0.0;
//        for (int i = 0; i < dim; i++) {
//            varNoise += (minMax[i][1] - minMax[i][0]) * (minMax[i][1] - minMax[i][0]);
//        }
//        varNoise /= 120; //alt: 120
//        double[][] mNoise = {{varNoise, 0.0}, {0.0, varNoise}};
//        double[] muNoise = new double[dim];
//        for (int i = 0; i < dim; i++) {
//            muNoise[i] = (minMax[i][0] + minMax[i][1]) / 2.0;
//        }
//        c = writeCluster(50, betaNoise, mNoise, muNoise);
//        c = writeCluster(1600, betaNoise, mNoise, muNoise);
//        System.out.println("varNoise: " + varNoise);
//        //varNoise: 24.16350657261328
//        System.out.println("muNoise: " + muNoise[0] + " " + muNoise[1]);
    //muNoise: 42.9155639254264 25.80626367633451
//SYNTHETIC EXAMPLE SIGMOD 08
//
//        Epdg e = new Epdg();
//        e.SigmodDimensionality(6);

        
        //EXAMPLE DATA SET 1 BTW2009
                int dim = 2;
        double[][] minMax = new double[dim][2]; //save for each dimension minimum and maximum for noise generation.
        for (int i = 0; i < dim; i++) {
            minMax[i][0] = Double.MAX_VALUE; //minimum
            minMax[i][1] = -Double.MAX_VALUE; //maximum
        }
        //3 Cluster
        double[] beta1 = {-0.85, 3.8};
        double[][] m1 = {{2.5, 2.0}, {0.5, 2.0}};
        double[] mu1 = {75.0, 20.0};
        double[][] c = writeCluster(1444, beta1, m1, mu1);
        //double[][] c = writeCluster(16000, beta1, m1, mu1) ;
        for (int j = 0; j < dim; j++) {
            if (c[j][0] < minMax[j][0]) {
                minMax[j][0] = c[j][0];
            }
            if (c[j][1] > minMax[j][1]) {
                minMax[j][1] = c[j][1];
            }
        }
//        double[] beta2 = {-0.5, 0.5};
//        double[][] m2 = {{2.0, 0.0}, {0.8, 2.0}};
//        double[] mu2 = {50.0, 40.0};
//        c = writeCluster(500, beta2, m2, mu2);
//        //c = writeCluster(16000, beta2, m2, mu2) ;
//        for (int j = 0; j < dim; j++) {
//            if (c[j][0] < minMax[j][0]) {
//                minMax[j][0] = c[j][0];
//            }
//            if (c[j][1] > minMax[j][1]) {
//                minMax[j][1] = c[j][1];
//            }
//        }
//        double[] beta3 = {4.5, 5.3};
//        double[][] m3 = {{2.0, 1.0}, {0.2, 2.0}};
//        double[] mu3 = {50.0, 15.0};
//        c = writeCluster(500, beta3, m3, mu3);
//        //c = writeCluster(16000, beta3, m3, mu3) ;
//        for (int j = 0; j < dim; j++) {
//            if (c[j][0] < minMax[j][0]) {
//                minMax[j][0] = c[j][0];
//            }
//            if (c[j][1] > minMax[j][1]) {
//                minMax[j][1] = c[j][1];
//            }
//        }
//        //noise
//        //System.out.println("NOISE");
//        double[] betaNoise = {-0.97, -0.22297};
//        double varNoise = 0.0;
//        for (int i = 0; i < dim; i++) {
//            varNoise += (minMax[i][1] - minMax[i][0]) * (minMax[i][1] - minMax[i][0]);
//        }
//        varNoise /= 120; //alt: 120
//        double[][] mNoise = {{varNoise, 0.0}, {0.0, varNoise}};
//        double[] muNoise = new double[dim];
//        for (int i = 0; i < dim; i++) {
//            muNoise[i] = (minMax[i][0] + minMax[i][1]) / 2.0;
//        }
//        c = writeCluster(50, betaNoise, mNoise, muNoise);
//        c = writeCluster(1600, betaNoise, mNoise, muNoise);
//        System.out.println("varNoise: " + varNoise);
//        //varNoise: 24.16350657261328
//        System.out.println("muNoise: " + muNoise[0] + " " + muNoise[1]);
    //muNoise: 42.9155639254264 25.80626367633451       
        
       //EXAMPLE DATA SET 1 BTW2009
        
        
        

    //2 Cluster
//        double[] beta1 = {-0.95,-0.95} ;
//        double[][] m1 = {{25.0, 2.0},{5.0, -2.0}} ;
//        double[] mu1 = {15.0, 25.0} ;
//        writeCluster(300, beta1, m1, mu1) ;
//        double[] beta2 = {-0.95,-0.95} ;
//        double[][] m2 = {{25.0, 2.0},{5.0, -2.0}} ;
//        double[] mu2 = {30.0, 50.0} ;
//        writeCluster(300, beta2, m2, mu2) ;


    //2 Cluster
//        double[] beta1 = {-0.95,-0.95,-0.95} ;
//        double[][] m1 = {{25.0, 2.0, 5.0},{5.0, -2.0, 3.0},{-3.0, 2.0, -1.0}} ;
//        double[] mu1 = {15.0, 25.0, 15.0} ;
//        writeCluster(300, beta1, m1, mu1) ;
//        double[] beta2 = {2.0,2.0,2.0} ;
//        double[][] m2 = {{25.0, 2.0, 3.0},{5.0, -2.0, 6.0}, {4.0, 2.0, 3.0}} ;
//        double[] mu2 = {30.0, 50.0, 70.0} ;
//        writeCluster(300, beta2, m2, mu2) ;
//



    }

    private void SigmodDimensionality(int dim) {
        double[][] minMax = new double[dim][2]; //save for each dimension minimum and maximum for noise generation.
        for (int i = 0; i < dim; i++) {
            minMax[i][0] = Double.MAX_VALUE; //minimum
            minMax[i][1] = -Double.MAX_VALUE; //maximum
        }
        double[] beta1 = {-0.85, 3.8};
        double[][] m1 = {{2.5, 2.0}, {0.5, 2.0}};
        double[] mu1 = {30.0, 25.0};
        double[] beta2 = {-0.5, 0.5};
        double[][] m2 = {{2.0, 0.0}, {0.8, 2.0}};
        double[] mu2 = {50.0, 40.0};
        double[] beta3 = {4.5, 5.3};
        double[][] m3 = {{2.0, 1.0}, {0.2, 2.0}};
        double[] mu3 = {50.0, 15.0};
        double[] betaNoise = {-0.97, -0.22297};
//mean and beta
        int counter = 0;
        double[] betaDim1 = new double[dim];
        double[] betaDim2 = new double[dim];
        double[] betaDim3 = new double[dim];
        double[] muDim1 = new double[dim];
        double[] muDim2 = new double[dim];
        double[] muDim3 = new double[dim];
        double[] betaDimNoise = new double[dim];
        for (int i = 0; i < dim; i++) {
            betaDim1[i] = beta1[counter];
            betaDim2[i] = beta2[counter];
            betaDim3[i] = beta3[counter];
            muDim1[i] = mu1[counter];
            muDim2[i] = mu2[counter];
            muDim3[i] = mu3[counter];
            betaDimNoise[i] = betaNoise[counter];
            counter++;
            if (counter == 2) {
                counter = 0;
            }
        }
//rotation matrices for clusters: random?
        Random r = new Random();
        double[][] m1Dim = new double[dim][dim];
        double[][] m2Dim = new double[dim][dim];
        double[][] m3Dim = new double[dim][dim];
        for (int i = 0; i < dim; i++) {
            for (int j = 0; j < dim; j++) {
                m1Dim[i][j] = r.nextDouble() * 2.0;
                m2Dim[i][j] = r.nextDouble() * 2.0;
                m3Dim[i][j] = r.nextDouble() * 2.0;
            }
        }


        //ROTATION MATRICES FOR CLUSTERS: not random
//        int counter_i = 0;
//        int counter_j = 0;
//        double[][] m1Dim = new double[dim][dim];
//        double[][] m2Dim = new double[dim][dim];
//        double[][] m3Dim = new double[dim][dim];
//        boolean secondRead = false;
//        for (int i = 0; i < dim; i++) {
//            for (int j = 0; j < dim; j++) {
//                m1Dim[i][j] = m1[counter_i][counter_j];
//                m2Dim[i][j] = m2[counter_i][counter_j];
//                m3Dim[i][j] = m3[counter_i][counter_j];
//                counter_j++;
//                if (counter_j == 2) {
//                    counter_j = 0;
//                    if (!secondRead) {
//                        secondRead = true;
//                    } else {
//                        counter_i++;
//                        secondRead = false;
//                    }
//                }
//            }
//        }
       
        double[][] c = writeCluster(5000, betaDim1, m1Dim, muDim1);
       for (int j = 0; j < dim; j++) {
            if (c[j][0] < minMax[j][0]) {
                minMax[j][0] = c[j][0];
            }
            if (c[j][1] > minMax[j][1]) {
                minMax[j][1] = c[j][1];
            }
        }
        c = writeCluster(5000, betaDim2, m2Dim, muDim2);
        for (int j = 0; j < dim; j++) {
            if (c[j][0] < minMax[j][0]) {
                minMax[j][0] = c[j][0];
            }
            if (c[j][1] > minMax[j][1]) {
                minMax[j][1] = c[j][1];
            }
        }
        c = writeCluster(5000, betaDim3, m3Dim, muDim3);
        for (int j = 0; j < dim; j++) {
            if (c[j][0] < minMax[j][0]) {
                minMax[j][0] = c[j][0];
            }
            if (c[j][1] > minMax[j][1]) {
                minMax[j][1] = c[j][1];
            }
        }
     //NOISE
         double varNoise = 0.0;
        for (int i = 0; i < dim; i++) {
            varNoise += (minMax[i][1] - minMax[i][0]) * (minMax[i][1] - minMax[i][0]);
        }
        varNoise /= 120; //alt: 120
        double[] muNoise = new double[dim];
        for (int i = 0; i < dim; i++) {
            muNoise[i] = (minMax[i][0] + minMax[i][1]) / 2.0;
        }

        double[][] mNoiseDim = new double[dim][dim];
        for (int i = 0; i < dim; i++) {
            for (int j = 0; j < dim; j++) {
                if (i == j) {
                    mNoiseDim[i][j] = varNoise;
                } else {
                    mNoiseDim[i][j] = 0.0;
                }
            }
        }
        c = writeCluster(500, betaDimNoise, mNoiseDim, muNoise);

    }
}
