package eval;
import Jama.*;
import compr.*;
import ica.Ica;
import io.*;
import java.util.*;
import visu.VisuSimple;
import visu.VisuInfo;
import visu.VisuVoronoi;
import visu.VisuCluster;
import visu.VisuInfo;

public class Clustering {
    Cluster[] initialClusters;
    Vector cl;
    DataObject[] data;
    int numObj;
    int dim;
    int numCluster;
    int precision;
    int maxID;     //max. clusterID
    double cost;
    OldClustering[] old;
    static int numTries = 0; //wie oft Verschmelzen versuchen
    static int numDistributions = 3; //gauss, uniform, laplace
    static int gauss = 0;
    static int uniform = 1;
    static int laplace = 2;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    static int magic = 10;
    
    //for use with clustering method
    public Clustering(DataObject[] data, int precision){
        this.data = data;
        this.precision = precision;
    }
    
    /** Creates a new instance of Clustering */
    public Clustering(DataObject[] data, Cluster[] clusters, int precision) {
        this.data = data;
        initialClusters = clusters;
        this.precision = precision;
        numObj = data.length;
        numCluster = clusters.length;
        maxID = maxClusterID() + 1;
        old = new OldClustering[numTries];
        for(int i = 0; i < numTries; i++)
            old[i] = new OldClustering();
        dim = data[0].coord.length;
        cl = new Vector();
        for(int i = 0; i < initialClusters.length; i++)
            cl.addElement(initialClusters[i]);
    }
    
    public void setCost(){
        cost = 0.0;
        for (int i = 0; i < cl.size(); i++)
            cost += ((Cluster)cl.elementAt(i)).cost;
    }
    
    public double currentCost(){
        double ccost = 0.0;
        for (int i = 0; i < cl.size(); i++)
            ccost += ((Cluster)cl.elementAt(i)).cost;
        return ccost;
    }
    
    public void test(){
        for(int i = 0; i < initialClusters.length; i++)
            // filterClusterKMeans(initialClusters[i]);
            filterCluster(initialClusters[i]);
        /*
        ClusterInit cli = new ClusterInit(data);
        Cluster[] aktClusters = getClusters();
        cli.setMeans(aktClusters);
         
         
        VisuVoronoi vs = new VisuVoronoi(data, aktClusters);
        vs.setSize(600,600);
        vs.setLocation(200,0);
        vs.setVisible(true);
         */
        testCost();
        VisuSimple vs = new VisuSimple(data);
        vs.setSize(600,600);
        vs.setLocation(200,0);
        vs.setVisible(true);
        
    }
    
    public Cluster[] getClusters(){
        Cluster[] aktClusters = new Cluster[cl.size()];
        for(int i = 0; i < cl.size(); i++)
            aktClusters[i] = (Cluster)cl.elementAt(i);
        return aktClusters;
    }
    
    public void testCost(){
        for(int i = 0; i < cl.size(); i++)
            modelChr((Cluster)cl.elementAt(i), getObjects((Cluster)cl.elementAt(i)));
        setCost();
        printPercentage();
        System.out.println("Cost of clustering " + cost);
        /*VisuSimple vs = new VisuSimple(data);
        vs.setSize(600,600);
        vs.setLocation(200,0);
        vs.setVisible(true);*/
    }
    
    
    
    public void evalModelNew(){
        for(int i = 0; i < initialClusters.length; i++)
            filterCluster(initialClusters[i]);
        System.out.println("ClID " );
        for(int i = 0; i < cl.size(); i++)
            System.out.print(((Cluster)cl.elementAt(i)).ID + " ");
        System.out.println();
        
        for(int i = 0; i < cl.size(); i++)
            model((Cluster)cl.elementAt(i), getObjects((Cluster)cl.elementAt(i)));
        
        double[][] clDist = new double[cl.size()][cl.size()];
        for(int i = 0; i < cl.size(); i++){
            for(int j = 0; j < cl.size(); j++){
                Cluster c_i = (Cluster)cl.elementAt(i);
                Cluster c_j = (Cluster)cl.elementAt(j);
                if(c_i.ID != c_j.ID){
                    DataObject[] m = getMergedObjects(c_i, c_j);
                    int minID = Math.min(c_i.ID, c_j.ID);
                    int maxID = Math.max(c_i.ID, c_j.ID);
                    double notMergedCost = c_i.cost + c_j.cost;
                    Cluster merged = new Cluster(minID, m.length, dim);
                    model(merged, m);
                    double aktSavedCost = notMergedCost - merged.cost;
                    clDist[i][j] = aktSavedCost;
                } else{
                    clDist[i][j] = 0.0;
                }
            }
        }
        /*
        for(int i = 0; i < cl.size(); i++)
            for(int j = 0; j < cl.size(); j++)
                clDist[i][i]+= Math.abs(clDist[i][j]);
         */
        Matrix m = new Matrix(clDist);
        m.print(6, 4);
        Matrix mm = m.times(m.transpose());
        EigenvalueDecomposition eig = new EigenvalueDecomposition(mm);
        eig.getD().print(6, 4);
        eig.getV().print(6, 4);
        
//        VisuSimple vs = new VisuSimple(data);
//        vs.setSize(600,600);
//        vs.setLocation(200,0);
//        vs.setVisible(true);
        
        VisuInfo vs = new VisuInfo(data, "result");
        vs.setSize(600,600);
        vs.setLocation(200,0);
        vs.setVisible(true);
    }
    
    public void evalModel(){
        /*
        VisuSimple v = new VisuSimple(data);
        v.setSize(600,600);
        v.setLocation(200,0);
        v.setVisible(true);
        if(true)
            return;
         */
        System.out.println("Filtering...");
        for(int i = 0; i < initialClusters.length; i++)
            filterCluster(initialClusters[i]);
        /*
        VisuSimple v = new VisuSimple(data);
        v.setSize(600,600);
        v.setLocation(200,0);
        v.setVisible(true);
        if(true)
            return;
         */
        
        for(int i = 0; i < cl.size(); i++)
            // model((Cluster)cl.elementAt(i), getObjects((Cluster)cl.elementAt(i)));
            
            //TEST
            //modelBIC((Cluster)cl.elementAt(i), getObjects((Cluster)cl.elementAt(i)), false);
            //TEST
            
            
            //TEST
            modelChr((Cluster)cl.elementAt(i), getObjects((Cluster)cl.elementAt(i)));
        //TEST
        
        System.out.println("Merging...");
        boolean improvement = true;
        while(improvement && cl.size() > 1){
            double savedCost = 0.0;
            Cluster[] nextm = new Cluster[2];
            nextm[0] = new Cluster();
            nextm[1] = new Cluster();
            for(int i = 0; i < cl.size(); i++){
                for(int j = 0; j < cl.size(); j++){
                    Cluster c_i = (Cluster)cl.elementAt(i);
                    Cluster c_j = (Cluster)cl.elementAt(j);
                    if(c_i.ID != c_j.ID){
                        DataObject[] m = getMergedObjects(c_i, c_j);
                        int minID = Math.min(c_i.ID, c_j.ID);
                        int maxID = Math.max(c_i.ID, c_j.ID);
                        double notMergedCost = c_i.cost + c_j.cost;
                        Cluster merged = new Cluster(minID, m.length, dim);
                        //model(merged, m);
                        
                        //TEST
                        //modelBIC(merged, m, true);
                        //TEST
                        
                        //TEST
                        modelChr(merged, m);
                        //TEST
                        
                        double aktSavedCost = notMergedCost - merged.cost;
                        //System.out.println("Cluster " + c_i.ID + " cluster " + c_j.ID + " aktSavedC " + aktSavedCost);
                        if(aktSavedCost > 0 && aktSavedCost > savedCost){
                            savedCost = aktSavedCost;
                            nextm[0] = c_i;
                            nextm[1] = c_j;
                        }
                    }
                }
            }
            if(savedCost >= 0 && nextm[0].ID != 0 && nextm[1].ID != 0){
                //System.out.println("savedCost " + savedCost);
                merge(nextm[0], nextm[1]);
                
                //TEST
                //updateClusterCost(Math.min(nextm[0].ID, nextm[1].ID));
                //TEST
                
            } else
                improvement = false;
        }
        //setCost();
        continueMerging(getCurrentClustering());
        setCost();
        System.out.println("Cost of clustering " + cost);
        //evaluate result
        int[] ids = new int[cl.size()];
        for(int i = 0; i < ids.length; i++)
            ids[i] = ((Cluster)cl.elementAt(i)).ID;
        ClusteringEvaluation e = new ClusteringEvaluation(data, ids);
        //e.evaluate();
        e.evaluateKl();
        IO ea = new IO();
        ea.printResult(data);
        
        printPercentage();
        
//        VisuSimple vs = new VisuSimple(data);
//        vs.setSize(600,600);
//        vs.setLocation(200,0);
//        vs.setVisible(true);
        /*
        VisuCluster vc = new VisuCluster(data, 2);
        vc.setSize(600,600);
        vc.setLocation(200,0);
        vc.setVisible(true);
         */
        
        VisuInfo vs = new VisuInfo(data, "result");
        vs.setSize(600,600);
        vs.setLocation(200,0);
        vs.setVisible(true);
    }
    
    public void continueMerging(OldClustering current){
        //System.out.println("continue merging");
        int count = 0;
        while(cl.size() > 1 && count < numTries){
            double savedCost = -Double.MAX_VALUE;
            Cluster[] nextm = new Cluster[2];
            nextm[0] = new Cluster();
            nextm[1] = new Cluster();
            for(int i = 0; i < cl.size(); i++){
                for(int j = 0; j < cl.size(); j++){
                    Cluster c_i = (Cluster)cl.elementAt(i);
                    Cluster c_j = (Cluster)cl.elementAt(j);
                    if(c_i.ID != c_j.ID){
                        DataObject[] m = getMergedObjects(c_i, c_j);
                        int minID = Math.min(c_i.ID, c_j.ID);
                        int maxID = Math.max(c_i.ID, c_j.ID);
                        double notMergedCost = c_i.cost + c_j.cost;
                        Cluster merged = new Cluster(minID, m.length, dim);
                        //model(merged, m);
                        
                        //TEST
                        //modelBIC(merged, m, true);
                        
                        //TEST
                        modelChr(merged, m);
                        
                        double aktSavedCost = notMergedCost - merged.cost;
                        //System.out.println("Cluster " + c_i.ID + " cluster " + c_j.ID + " aktSavedC " + aktSavedCost);
                        if(aktSavedCost > savedCost){
                            savedCost = aktSavedCost;
                            nextm[0] = c_i;
                            nextm[1] = c_j;
                        }
                    }
                }
            }
            if(nextm[0].ID != 0 && nextm[1].ID != 0){
                //System.out.println("savedCost " + savedCost);
                merge(nextm[0], nextm[1]);
                
                //TEST
                //updateClusterCost(Math.min(nextm[0].ID, nextm[1].ID));
                //TEST
                
            }
            
            old[count] = getCurrentClustering();
            count++;
        }
        int cheap = old.length;
        double minCost = current.cost;
        //System.out.println("Started with " + current.clusters.length + " cost " + current.cost);
        for(int i = 0; i < count; i++){
            //System.out.println("NumClusters: " + old[i].clusters.length + " cost:  " + old[i].cost);
            if(old[i].cost < minCost){
                cheap = i;
                minCost = old[i].cost;
            }
        }
        if(cheap < old.length)
            updateClustering(old[cheap]);
        else
            updateClustering(current);
        
        
        
    }
    
    private OldClustering getCurrentClustering(){
        int[] clusterID = new int[numObj];
        for(int i = 0; i < data.length; i++)
            clusterID[i] = data[i].clusterID;
        Cluster[] cclusters = new Cluster[cl.size()];
        for(int i = 0; i < cclusters.length; i++){
            Cluster akt = (Cluster)cl.elementAt(i);
            cclusters[i] = akt;
        }
        OldClustering cc = new OldClustering(clusterID, cclusters, currentCost());
        return cc;
    }
    
    private void updateClustering(OldClustering o){
        for(int i = 0; i < data.length; i++)
            data[i].clusterID = o.clusterID[i];
        this.cost = o.cost;
        cl.removeAllElements();
        for(int i = 0; i < o.clusters.length; i++)
            cl.addElement(o.clusters[i]);
    }
    
    private int maxClusterID(){
        int max = 0;
        for(int i = 0; i < numCluster; i++)
            if(initialClusters[i].ID > max)
                max = initialClusters[i].ID;
        return max;
    }
    
    private int minCost(double[] cost){
        int minIndex = cost.length+1;
        double min = Double.MAX_VALUE;
        for(int i = 0; i < cost.length; i++)
            if(cost[i] < min){
            minIndex = i;
            min = cost[i];
            }
        return minIndex;
    }
    
    //Art der Verteilung entscheiden
    public void model(Cluster c, DataObject[] obj){
        //System.out.println("Cluster " + c.ID);
        double cost_untrans = 0.0;
        //Verteilung der einzelnen Koordinaten
        int[] untrans_dist = new int[dim];
        double[] mean = mean(obj);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(obj, mean, i);
            
            int min = minCost(aktCost);
            if(min == 0){
                untrans_dist[i] = gauss;
                cost_untrans += aktCost[0];
            }
            if(min == 1){
                untrans_dist[i] = uniform;
                cost_untrans += aktCost[1];
            }
            
            if(min == 2){
                untrans_dist[i] = laplace;
                cost_untrans += aktCost[2];
            }
        }
        double cost_trans = 0.0;
        int[] trans_dist = new int[dim];
        //TEST: ICA
        // DataObject[] transData = ica(obj, c);
        
        //PCA
        DataObject[] transData = pca(obj, c);
        
        double[] transMean = mean(transData);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(transData, transMean, i);
            
            int min = minCost(aktCost);
            if(min == 0){
                trans_dist[i] = gauss;
                cost_trans += aktCost[0];
            }
            if(min == 1){
                trans_dist[i] = uniform;
                cost_trans += aktCost[1];
            }
            if(min == 2){
                trans_dist[i] = laplace;
                cost_trans += aktCost[2];
            }
        }
        //System.out.println("costTrans " + cost_trans + " costUntrans " + cost_untrans);
        if(cost_trans < cost_untrans){
            c.subspaceCluster = true;
            c.dist = trans_dist;
            c.mean = transMean;
        } else{
            c.dir = null;
            c.dist = untrans_dist;
            c.subspaceCluster = false;
            c.mean = mean;
        }
        //double hole = 0.0;
        double hole = c.numObj  * 200 * (Math.log(1/(1-1.75 * Math.pow(Math.max(2.0,c.numObj), -0.8127)))/Math.log(2));
        double costID = c.numObj * Math.log((double)data.length/(double)c.numObj)/Math.log(2);
        c.cost = Math.min(cost_trans, cost_untrans) + costID + hole;
        //System.out.println("Cluster " + c.ID + " subspace " + c.subspaceCluster + " cost " + c.cost);
        //System.out.println("cost without hole " + (c.cost-hole));
        //System.out.println("Verteilung ");
        //for(int i = 0; i < c.dist.length; i++)
        //System.out.print(c.dist[i] + " ");
    }
    
    
    //Art der Verteilung entscheiden
    public void modelChr(Cluster c, DataObject[] obj){
        //System.out.println("Cluster " + c.ID);
        double cost_untrans = 0.0;
        //Verteilung der einzelnen Koordinaten
        int[] untrans_dist = new int[dim];
        double[] mean = mean(obj);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(obj, mean, i);
            
            int min = minCost(aktCost);
            if(min == 0){
                untrans_dist[i] = gauss;
                cost_untrans += aktCost[0];
            }
            if(min == 1){
                untrans_dist[i] = uniform;
                cost_untrans += aktCost[1];
            }
            
            if(min == 2){
                untrans_dist[i] = laplace;
                cost_untrans += aktCost[2];
            }
        }
        cost_untrans += dim * 32 * 2;
        // Art Verteilung
        cost_untrans += dim * 2;
        double cost_trans = 0.0;
        int[] trans_dist = new int[dim];
        //TEST: ICA
        //DataObject[] transData = ica(obj, c);
        
        //PCA
        DataObject[] transData = pca(obj, c);
        
        double[] transMean = mean(transData);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(transData, transMean, i);
            
            int min = minCost(aktCost);
            if(min == 0){
                trans_dist[i] = gauss;
                cost_trans += aktCost[0];
            }
            if(min == 1){
                trans_dist[i] = uniform;
                cost_trans += aktCost[1];
            }
            if(min == 2){
                trans_dist[i] = laplace;
                cost_trans += aktCost[2];
            }
        }
        //Parameter zhlen
        cost_trans += dim*32*2 + dim*dim * 32;
        // Art Verteilung
        cost_trans += dim * 2;
        //System.out.println("costTrans " + cost_trans + " costUntrans " + cost_untrans);
        if(cost_trans < cost_untrans){
            c.subspaceCluster = true;
            c.dist = trans_dist;
            c.mean = transMean;
        } else{
            c.dir = null;
            c.dist = untrans_dist;
            c.subspaceCluster = false;
            c.mean = mean;
        }
        double hole = 0.0;
        //double hole = c.numObj  * 200 * (Math.log(1/(1-1.75 * Math.pow(Math.max(2.0,c.numObj), -0.8127)))/Math.log(2));
        double costID = c.numObj * Math.log((double)data.length/(double)c.numObj)/Math.log(2);
        c.cost = Math.min(cost_trans, cost_untrans) + costID + hole;
        //System.out.println("Cluster " + c.ID + " subspace " + c.subspaceCluster + " cost " + c.cost);
        //System.out.println("cost without hole " + (c.cost-hole));
        //System.out.println("Verteilung ");
        //for(int i = 0; i < c.dist.length; i++)
        //  System.out.print(c.dist[i] + " ");
    }
    
    //Art der Verteilung entscheiden
    public void modelBIC(Cluster c, DataObject[] obj, boolean merged){
        int numClusters;
        System.out.println("Cluster " + c.ID);
        double cost_untrans = 0.0;
        //Verteilung der einzelnen Koordinaten
        int[] untrans_dist = new int[dim];
        double[] mean = mean(obj);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(obj, mean, i);
            int min = minCost(aktCost);
            if(min == 0){
                untrans_dist[i] = gauss;
                cost_untrans += aktCost[0];
            }
            if(min == 1){
                untrans_dist[i] = uniform;
                cost_untrans += aktCost[1];
            }
            if(min == 2){
                untrans_dist[i] = laplace;
                cost_untrans += aktCost[2];
            }
        }
        double cost_trans = 0.0;
        int[] trans_dist = new int[dim];
        //TEST: ICA
        //DataObject[] transData = ica(obj, c);
        
        //PCA
        DataObject[] transData = pca(obj, c);
        double[] transMean = mean(transData);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(transData, transMean, i);
            int min = minCost(aktCost);
            if(min == 0){
                trans_dist[i] = gauss;
                cost_trans += aktCost[0];
            }
            if(min == 1){
                trans_dist[i] = uniform;
                cost_trans += aktCost[1];
            }
            if(min == 2){
                trans_dist[i] = laplace;
                cost_trans += aktCost[2];
            }
        }
        System.out.println("costTrans " + cost_trans + " costUntrans " + cost_untrans);
        if(cost_trans < cost_untrans){
            c.subspaceCluster = true;
            c.dist = trans_dist;
            c.mean = transMean;
        } else{
            c.dir = null;
            c.dist = untrans_dist;
            c.subspaceCluster = false;
            c.mean = mean;
        }
        double costID = c.numObj * Math.log((double)data.length/(double)c.numObj)/Math.log(2);
        if(merged)
            c.cost = Math.min(cost_trans, cost_untrans) + costID + 0.5 * (double) numParam(cl.size()-1) * Math.log(numObj)/Math.log(2);
        else
            c.cost = Math.min(cost_trans, cost_untrans) + costID + 0.5 * (double) numParam(cl.size()) * Math.log(numObj)/Math.log(2);
        System.out.println("Cluster " + c.ID + " subspace " + c.subspaceCluster + " cost " + c.cost);
        System.out.println("Verteilung ");
        for(int i = 0; i < c.dist.length; i++)
            System.out.print(c.dist[i] + " ");
    }
    
    private int numParam(int numClusters){
        return numClusters*dim;
    }
    
    private void updateClusterCost(int mergedID){
        int oldCost = numParam(cl.size()+1);
        int newCost = numParam(cl.size());
        int diff = oldCost-newCost;
        for(int i = 0; i < cl.size(); i++)
            if(((Cluster)cl.elementAt(i)).ID != mergedID)
                ((Cluster)cl.elementAt(i)).cost -= diff;
        
    }
    
    
    
    //merge cluster c and d.
    public void merge(Cluster c, Cluster d){
        //System.out.println();
        //System.out.println("Merging clusters " + c.ID + " and " + d.ID);
        double maxMergeCost = c.cost + d.cost;
        DataObject[] m = getMergedObjects(c, d);
        int minID = Math.min(c.ID, d.ID);
        int maxID = Math.max(c.ID, d.ID);
        if(minID == maxID)
            System.err.println("trying to merge clusters with the same id: " + minID);
        Cluster merged = new Cluster(minID, m.length, dim);
        //model(merged, m);
        
        //TEST
        modelChr(merged, m);
        //TEST
        
        double hole = 0.0;
        //double hole = merged.numObj * (Math.log(1/(1- 1.75 * Math.pow(merged.numObj, -0.8127)))/Math.log(2));
        //if(merged.cost - dim*hole <= maxMergeCost){
        //System.out.println("merging Clusters, saved cost: " + (maxMergeCost-merged.cost));
        cl.remove(c);
        cl.remove(d);
        merged.numObj = c.numObj + d.numObj;
        cl.add(merged);
        adjustClusterID(maxID, minID);
        //} else
        // System.out.println("merging too expensive : merged cost " + (merged.cost) + " not merged cost " + maxMergeCost);
    }
    
    
    
    public void gaussTest(Cluster c){
        DataObject[] members = getObjects(c);
        double[] mean = mean(members);
        double[] std = new double[c.dim];
        double sumGauss;
        double sumNorm;
        //System.out.println("Cluster " + c.ID);
        for(int i = 0; i < c.dim; i++){
            sumGauss = 0.0;
            double stdv = stdv(members, mean, i);
            //System.out.println("stdv " + stdv);
            for(int j = 0; j < members.length; j++)
                sumGauss += gaussCost(members[j].coord[i], mean[i], stdv, 1);
            //System.out.println("Koordinate " + i + " gaussCost " + sumGauss);
        }
        for(int i = 0; i < c.dim; i++){
            sumNorm = 0.0;
            double[] bounds = bounds(members, i);
            for(int j = 0; j < members.length; j++)
                sumNorm += uniformCost(members[j].coord[i], bounds[1], bounds[0], 1);
            //System.out.println("Koordinate " + i + " normCost " + sumNorm);
        }
        //System.out.println();
    }
    
    //adjust clusterID after merging
    private void adjustClusterID(int oldID, int newID){
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == oldID)
                data[i].clusterID = newID;
    }
    
    
    
    
    //returns cost to code the numCoord-th coordinate according to the different distributions: cost[0]: gaussian, cost[1]: uniform, cost[2]: laplacian
    public double[] cost(DataObject[] o, double[] mean, int numCoord){
        double[] cost = new double[numDistributions];
        int dim = o[0].coord.length;
        double stdv = stdv(o, mean, numCoord);
        for(int i = 0; i < o.length; i++){
            cost[0] += gaussCost(o[i].coord[numCoord], mean[numCoord], stdv, precision);
        }
        double[] bounds = bounds(o, numCoord);
        for(int i = 0; i < o.length; i++){
            cost[1] += uniformCost(o[i].coord[numCoord], bounds[1], bounds[0], precision);
        }
        for(int i = 0; i < o.length; i++){
            cost[2] += laplaceCost(o[i].coord[numCoord], mean[numCoord], stdv, precision);
        }
        return cost;
    }
    
    private double laplaceCost(double x, double mean, double std, int precision){
        double b = std/Math.sqrt(2);
        if (b<Math.pow(10, -precision)/10)
            return 0 ;
        double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
        double prob = Math.max(Double.MIN_VALUE, Math.min(1.0, pdf * Math.pow(10, -precision)));
        //return 100000;
        return Math.log(1.0/prob)/Math.log(2);
    }
    
    
    private double gaussCost(double x, double mean, double std, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double var = Math.max(std*std, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*var)+precision*ln10+Math.pow(x-mean,2)/2/var)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    private double uniformCost(double x, double lb, double ub, int precision){
        double prob ;
        if (Math.abs(ub-lb) < Math.pow(10, -precision))
            prob = 1 ;
        else
            prob = Math.pow(10, -precision) / Math.abs(ub - lb) ;
        return Math.log(1/prob)/Math.log(2);
    }
    
    public double[] mean(DataObject[] o){
        double[] mean = new double[o[0].coord.length];
        for(int i = 0; i < o.length; i++)
            for(int j = 0; j < mean.length; j++){
            mean[j] += o[i].coord[j];
            }
        for(int j = 0; j < mean.length; j++)
            mean[j] /= o.length;
        return mean;
    }
    
    //stdv of coordinate numCoord
    private double stdv(DataObject[] o, double[] mean, int numCoord){
        double stdv = 0.0;
        for(int i = 0; i < o.length; i++)
            stdv += (o[i].coord[numCoord] - mean[numCoord]) * (o[i].coord[numCoord] - mean[numCoord]);
        return Math.sqrt(stdv/o.length);
    }
    
    //bounds[0]: upper bound, bounds[1]: lower bound
    private double[] bounds(DataObject[] o, int numCoord){
        int dim = o[0].coord.length;
        double[] bounds = new double[2];
        for(int i = 0; i < bounds.length; i++){
            bounds[0] = -Double.MAX_VALUE;
            bounds[1] = Double.MAX_VALUE;
        }
        for(int i = 0; i < o.length; i++){
            for(int j = 0; j < dim; j++){
                if(o[i].coord[numCoord] > bounds[0])
                    bounds[0] = o[i].coord[j];
                if(o[i].coord[numCoord] < bounds[1])
                    bounds[1] = o[i].coord[j];
            }
        }
        return bounds;
    }
    
    private double[][] covar(DataObject[] data, double[] mean){
        double[][] cov = new double[dim][dim];
        for(int l = 0; l < data.length; l++)
            for(int i = 0; i < dim; i++)
                for(int j = 0; j < dim; j++)
                    cov[i][j]+= (data[l].coord[j]-mean[j])*(data[l].coord[i]-mean[i]);
        for(int i = 0; i < dim; i++)
            for(int j = 0; j < dim; j++)
                cov[i][j]/= data.length;
        return cov;
    }
    
    public double[][] covarByMedian(DataObject[] data, double[] mean) {
        int numObj= data.length ;
        int dim = data[0].coord.length ;
        double [] sortarray = new double[numObj] ;
        double[][] cov = new double[dim][dim];
        if (numObj <5){
            for (int i=0 ; i<dim ; i++)
                cov[i][i]=1.0 ;
            return cov ;
        }
        for(int l1 = 0; l1 < dim; l1++) {
            for(int l2 = 0; l2 < dim; l2++) {
                for(int j = 0; j < numObj; j++)
                    sortarray[j] = (data[j].coord[l1]-mean[l1])*(data[j].coord[l2]-mean[l2]) ;
                
                Arrays.sort(sortarray) ;
                if (numObj%2 == 0)
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3
                            ;
            }
            
        }
        double maxdiff = 0.0 ;
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                sum += Math.abs(cov[i][j]) ;
            double diff = 1.1*sum - 2.1*cov[i][i] ;
            if (diff>maxdiff)
                maxdiff = diff ;
        }
        if (maxdiff > 0.0)
            for (int i=0 ; i<dim ; i++)
                cov[i][i] += maxdiff ;
        
        return cov ;
    }
    
    
    public int countClusterMembers(int clusterID){
        int counter = 0;
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == clusterID)
                counter++;
        return counter;
    }
    
    public DataObject[] getObjects(Cluster c){
        //DataObject[] members = new DataObject[c.numObj];
        DataObject[] members = new DataObject[countClusterMembers(c.ID)];
        int counter = 0;
        for(int j = 0; j < data.length; j++)
            if(data[j].clusterID == c.ID){
            members[counter] = data[j];
            counter++;
            }
        return members;
    }
    
    private DataObject[] getMergedObjects(Cluster c, Cluster d){
        DataObject[] members = new DataObject[c.numObj + d.numObj];
        int counter = 0;
        for(int j = 0; j < data.length; j++)
            if(data[j].clusterID == c.ID || data[j].clusterID == d.ID){
            members[counter] = data[j];
            counter++;
            }
        return members;
    }
    
    //ICA
    public DataObject[] ica(DataObject[] untransData, Cluster c){
        int dim = untransData[0].coord.length;
        double[][] coordObj = new double[untransData.length][dim];
        for(int i = 0; i < untransData.length; i++){
            for(int j = 0; j < dim; j++)
                coordObj[i][j] = untransData[i].coord[j];
        }
        Ica ica = new Ica();
        double[][] w = ica.optimizeEntropy(coordObj);
        Matrix data = new Matrix(coordObj);
        Matrix dir = new Matrix(w);
        Matrix icaTrans = (data.times(dir.inverse())).transpose() ;
        double[][] tr_ica = icaTrans.getArrayCopy();
        DataObject[] transData_ica = new DataObject[untransData.length];
        for(int i = 0; i < transData_ica.length; i++){
            double[] co = new double[dim];
            for(int j = 0; j < dim; j++)
                co[j] = tr_ica[j][i];
            transData_ica[i] = new DataObject(co, i, c.ID);
        }
        c.dir = w;
        return transData_ica;
    }
    
    public DataObject[] pca(DataObject[] untransData, Cluster c){
        //System.out.println("PCA Cluster " + c.ID + " numObj " + c.numObj);
        double[][] coordObj = new double[untransData.length][dim];
        for(int i = 0; i < untransData.length; i++){
            for(int j = 0; j < dim; j++)
                coordObj[i][j] = untransData[i].coord[j];
        }
        DataObject[] transData_pca = new DataObject[untransData.length];
        Matrix data = new Matrix(coordObj);
        double[] mean = mean(untransData);
        c.mean = mean;
        Matrix covar = new Matrix(covar(untransData, c.mean));
        EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
        Matrix vectors = eig.getV();
        Matrix values = eig.getD();
        //System.out.println("EW");
        //values.print(5, 3);
        //System.out.println("EV");
        //vectors.print(5, 3);
        c.dir = vectors.getArrayCopy();
        double[][] transCoord = new double[c.numObj][dim];
        for(int i = 0; i < c.numObj; i++){
            for(int j = 0; j < dim; j++){
                for(int k = 0; k < dim; k++){
                    transCoord[i][k] += (untransData[i].coord[j]-c.mean[j]) * c.dir[j][k];
                }
            }
        }
        for(int i = 0; i < c.numObj; i++){
            double[] co = new double[dim];
            for(int j = 0; j < dim; j++)
                co[j] = transCoord[i][j];
            transData_pca[i] = new DataObject(co, untransData[i].number, untransData[i].clusterID);
        }
        
        return transData_pca;
    }
    
    
    
    public double logStar(double cc, int precision){
        double c = Math.pow(10.0, precision) * Math.abs(cc) ;
        double result = 0.0;
        // while (c > 0){
        c = Math.log(c)/Math.log(2);
        if(c > 0)
            result += c;
        //}
        return result;
    }
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    
    public Matrix getMahalanobis(DataObject[] data, double fraction,
            boolean robust, Matrix [] ctr, Matrix[] vv, int vNum) {
        // determines the matrix for mahalonobis distance either robust or not
        // based on a fraction of the data set which is selected by Euclidean nearest neighbors
        int numObj = data.length ;
        int numRed = (int)(numObj*fraction) ;
        double[] center = robust
                ? median(data)
                : mean(data) ;
        double[] distances = new double[numObj] ;
        for (int i=0 ; i<numObj ; i++)
            distances[i] = distance(center, data[i].coord);
        SortPack [] distcopy = SortPack.array(distances) ;
        DataObject [] dataRed = new DataObject [numRed] ;
        for (int i=0 ; i<numRed ; i++)
            dataRed[i] = data[distcopy[i].index] ;
        center = robust
                ? median(dataRed)
                : mean(dataRed) ;
        ctr[vNum] = rowVector(center) ;
        double[][] cov = robust
                ? covarByMedian(dataRed, center)
                : covar(dataRed, center) ;
        EigenvalueDecomposition e = new EigenvalueDecomposition(new Matrix(cov)) ;
        vv[vNum] = e.getV() ;
        Matrix d = e.getD() ;
        //d.print(6, 4);
        double maxEigen = -Double.MAX_VALUE ;
        for (int i=0 ; i<dim ; i++)
            maxEigen = Math.max(maxEigen, d.get(i,i)) ;
        for (int i=0 ; i<dim ; i++)
            d.set(i,i,1/Math.max(d.get(i,i), maxEigen*0.001)) ;
        return vv[vNum].times(d).times(vv[vNum].transpose()) ;
    }
    
    private double[][] getRobustEV(DataObject[] data){
        double[] center = median(data);
        double[][] cov = covarByMedian(data, center);
        EigenvalueDecomposition e = new EigenvalueDecomposition(new Matrix(cov)) ;
        return e.getV().getArrayCopy();
    }
    
    public Cluster filterCluster(Cluster cluster) {
        DataObject [] data = getObjects(cluster) ;
        int numObj = data.length ;
        if(numObj < 5)
            return null;
        Matrix[] vv = new Matrix [10] ;
        vv[0] = Matrix.identity(dim,dim) ;
        Matrix[] ctr = new Matrix [10] ;
        Matrix [] vdvt = {
            Matrix.identity(dim, dim) ,
            getMahalanobis(data, 1.0, true, ctr, vv, 1) ,
            getMahalanobis(data, 0.5, true, ctr, vv, 2) ,
            getMahalanobis(data, 1.0, false, ctr, vv, 3) ,
            getMahalanobis(data, 0.5, false, ctr, vv, 4)
        } ;
        ctr[0] = ctr[3] ;
        double minCost = Double.MAX_VALUE ;
        double minCostUnsplit = Double.MAX_VALUE ;
        int minCase = -1 ;
        int minCut = -1 ;
        
        double [] lb = data[0].coord.clone() ;
        double [] ub = data[0].coord.clone() ;
        for (int i=1 ; i<numObj ; i++) {
            for (int j=0 ; j<dim ; j++) {
                lb[j] = Math.min(lb[j], data[i].coord[j]) ;
                ub[j] = Math.max(ub[j], data[i].coord[j]) ;
            }
        }
        for (int testCase=0 ; testCase<vdvt.length ; testCase ++) {
            //System.out.println("=============================================") ;
            //System.out.println("TestCase "+testCase) ;
            double [] distances = new double [numObj] ;
            for (int i=0 ; i<numObj ; i++) {
                Matrix x = rowVector(data[i].coord).minus(ctr[testCase]) ;
                // left out here
                
                distances[i]=x.times(vdvt[testCase]).times(x.transpose()).get(0,0) ;
            }
            SortPack [] distcopy = SortPack.array(distances) ;
            double [][] reduced = new double [numObj][] ;
            for (int i=0 ; i<numObj ; i++)
                reduced[i] =
                        rowVector(data[distcopy[i].index].coord).times(vv[testCase]).getColumnPackedCopy()
                        ;
            double [] meanCum = new double [dim] ;
            double [] mean = new double [dim] ;
            double [] varMeanCum = new double [dim] ;
            double [] varMean = new double [dim] ;
            double [] var = new double [dim] ;
            double [] lbCore = reduced[0].clone() ;
            double [] ubCore = reduced[0].clone() ;
            double [] cGaussCost = new double [dim] ;
            double [] cLaplaceCost = new double [dim] ;
            double cOverallCost = Double.MAX_VALUE ;
            for (int cut = 0 ; cut < numObj ; cut ++) {
                // update now models for cluster
                cOverallCost = 0 ;
                for (int j=0 ; j<dim ; j++) {
                    meanCum [j] += reduced[cut][j] ;
                    mean [j] = meanCum [j] / (cut+1) ;
                    varMeanCum [j] += reduced[cut][j] * reduced[cut][j] ;
                    var [j] = varMeanCum [j] / (cut/*+1*/) - mean[j]*mean[j] ;
                    // SOME SANITY CHECK NEEDED HERE !!!!
                    lbCore[j] = Math.min(lbCore[j], reduced[cut][j]) ;
                    ubCore[j] = Math.max(ubCore[j], reduced[cut][j]) ;
                    cGaussCost[j]=0 ;
                    cLaplaceCost[j]=0 ;
                    if(cut > 0)
                        for (int i=0 ; i<=cut ; i++) {
                        cGaussCost[j] += gaussCost(reduced[i][j],
                                mean[j], Math.sqrt(var[j]), precision) ;
                        cLaplaceCost[j] += laplaceCost(reduced[i][j],
                                mean[j], Math.sqrt(var[j]), precision) ;
                        }
                    
                    else {
                        cGaussCost[j] = Double.MAX_VALUE;
                        cLaplaceCost[j] = Double.MAX_VALUE;
                    }
                    cOverallCost += Math.min(cLaplaceCost[j], Math.min(cGaussCost[j],
                            (cut+1)*uniformCost(0, lbCore[j], ubCore[j], precision) ) )
                            + (numObj-cut-1)*uniformCost(0, lb[j], ub[j], precision)  ;
                }
                if (cOverallCost<minCost){
                    minCost=cOverallCost ;
                    minCase = testCase ;
                    minCut = cut ;
                }
            }
            if (cOverallCost < minCostUnsplit)
                // Assume that all points are *cluster members*
                // The actual solution must outperform this situation
                minCostUnsplit = cOverallCost ;
        }
        
        // TEST TEST TEST minCostUnsplit: Assume that all points are
        // *outliers*
        minCostUnsplit = 0 ;
        for (int j=0 ; j<dim ; j++)
            minCostUnsplit += numObj*uniformCost(0, lb[j], ub[j], precision) ;
        
        if (minCost    +numObj    > minCostUnsplit||minCut<1 || minCut>numObj-2)
            return null ;
        
        // CREATE A NEW CLUSTER
        // TODO: THE NEW CLUSTER ID MUST BE DETERMINED !!!
        Cluster newCl = new Cluster(maxID, numObj-minCut-1, dim) ;
        maxID++;
        cl.add(newCl);
        //TEST
        //TEST
        //System.out.println("ClusterID " + newCl.ID + " neu erzeugt " + countClusterMembers(newCl.ID) + " obj ");
        
        int testCase = minCase ;
        double [] distances = new double [numObj] ;
        for (int i=0 ; i<numObj ; i++) {
            Matrix x = rowVector(data[i].coord).minus(ctr[testCase]) ;
            
            distances[i]=x.times(vdvt[testCase]).times(x.transpose()).get(0,0) ;
        }
        SortPack [] distcopy = SortPack.array(distances) ;
        for (int i=minCut+1 ; i<numObj ; i++)
            data[distcopy[i].index].clusterID = newCl.ID ;
        cluster.numObj = minCut+1 ;
        //TEST
        //System.out.println("Cluster " + newCl.ID + " neu erzeugt " + countClusterMembers(newCl.ID) + " obj ");
        return newCl ;
    }
    
    
    public Cluster filterClusterKMeans(Cluster cluster){
        Cluster c = new Cluster();
        boolean done = false;
        Partitioning p;
        DataObject [] data = getObjects(cluster) ;
        int numObj = data.length ;
        boolean[] isCore = new boolean[numObj];
        int[] numbers = new int[numObj];
        int[] internalIndex = new int[numObj];
        for(int i = 0; i < numObj; i++){
            numbers[i] = data[i].number;
            internalIndex[i] = i;
        }
        HashMap h = new HashMap(numObj);
        for(int i = 0; i < numObj; i++)
            h.put(numbers[i], internalIndex[i]);
        /*
        if(numObj < 5)
            return null;
        else
         */
        Partitioning initial = init(data);
        for(int i = 0; i < initial.core.length; i++){
            Integer o = (Integer)h.get(initial.core[i].number);
            int oakt = o.intValue();
            isCore[oakt] = true;
        }
        initial.update();
        double costAkt = initial.setCost();
        System.out.println("initialCost " + costAkt);
        p = initial;
        while(!done){
            Partitioning pNew = p.assignPoints();
            //test if done: when no changes then stop
            boolean change = false;
            if(pNew.core.length != p.core.length)
                change = true;
            else{
                for(int i = 0; i < pNew.core.length; i++){
                    Integer o = (Integer)h.get(pNew.core[i].number);
                    int oakt = o.intValue();
                    if(!isCore[oakt])       //new core object
                        change = true;
                }
                for(int i = 0; i < pNew.noise.length; i++){
                    Integer o = (Integer)h.get(pNew.noise[i].number);
                    int oakt = o.intValue();
                    if(isCore[oakt])     //new noise object
                        change = true;
                }
            }
            pNew.updateIter();
            pNew.setCost();
            System.out.println("aktCost " + pNew.cost);
            for(int i = 0; i < numObj; i++)
                isCore[i] = false;
            for(int i = 0; i < pNew.core.length; i++){
                Integer o = (Integer)h.get(pNew.core[i].number);
                int oakt = o.intValue();
                isCore[oakt] = true;
            }
            p = pNew;
            if(!change)
                done = true;
        }
        System.out.println("cost " + cost);
        return c;
    }
    
    
    
    
    private Partitioning init(DataObject[] data){
        int numObj = data.length;
        //CHECK THIS
        Matrix[] vv = new Matrix [10] ;
        vv[0] = Matrix.identity(dim,dim) ;
        Matrix[] ctr = new Matrix [10] ;
        Matrix dir = getMahalanobis(data, 1.0, true, ctr, vv, 1);
        // vv?
        double [] distances = new double [numObj] ;
        for (int i=0 ; i<numObj ; i++) {
            Matrix x = rowVector(data[i].coord).minus(ctr[1]) ;
            distances[i]=x.times(dir.times(x.transpose())).get(0,0);
        }
        SortPack [] distcopy = SortPack.array(distances);
        int cut = distcopy.length/2;
        double [][] reduced = new double [numObj][] ;
        for (int i=0 ; i<numObj ; i++)
            reduced[i] = rowVector(data[distcopy[i].index].coord).times(vv[1]).getColumnPackedCopy();
        //determine cost for extreme cases: allNoise and allCluster. Set upper and lower bound, mean and variance for all data.
        double allNoise = 0.0;
        double allCluster = 0.0;
        double[] lbAll = new double[dim];
        double[] lbCore = new double[dim];
        double[] var = new double[dim];
        double[] mean = new double[dim];
        for(int i = 0; i < lbAll.length; i++){
            lbAll[i] = Double.MAX_VALUE;
            lbCore[i] = Double.MAX_VALUE;
        }
        double[] ubAll = new double[dim];         //check lb, ub, stdv!!!
        double[] ubCore = new double[dim];
        for(int i = 0; i < ubAll.length; i++){
            ubAll[i] = -Double.MAX_VALUE;
            ubCore[i] = -Double.MAX_VALUE;
        }
        for(int i = 0; i < numObj; i++)
            for(int j = 0; j < dim; j++){
            var[j] += reduced[i][j] * reduced[i][j];
            mean[j] += data[i].coord[j];
            if(data[i].coord[j] < lbAll[j])
                lbAll[j] = data[i].coord[j];
            if(data[i].coord[j] > ubAll[j])
                ubAll[j] = data[i].coord[j];
            if(reduced[i][j] < lbCore[j])
                lbCore[j] = reduced[i][j];
            if(reduced[i][j] > ubCore[j])
                ubCore[j] = reduced[i][j];
            }
        for(int i = 0; i < dim; i++){
            mean[i] /= (double)numObj;
            allNoise+= uniformCost(0, lbAll[i], ubAll[i], precision);
        }
        for(int i = 0; i < numObj; i++)
            for(int j = 0; j < dim; j++)
                var[j] += reduced[i][j] * reduced[i][j];
        for(int j = 0; j < dim; j++)
            var[j]/= (double)numObj;
        allNoise *= numObj;
        for(int i = 0; i < numObj; i++)
            for(int j = 0; j < dim; j++)
                allCluster += Math.min(uniformCost(0, lbCore[j], ubCore[j], precision),Math.min(gaussCost(reduced[i][j], mean[j], Math.sqrt(var[j]), precision), laplaceCost(reduced[i][j],  mean[j], Math.sqrt(var[j]), precision)));
        //AllCluster und AllNoise noch nicht korrekt: Es fehlt Kosten fr Rotationsmatrix bei allCluster.
        //sort out 50% of the objects as noise according mahalonobis distance
        int numCore = numObj-cut;
        int numNoise = numObj-numCore;
        DataObject[] core = new DataObject[numCore];
        DataObject[] noise = new DataObject[numNoise];
        if(numObj %2 == 0){
            for(int i = 0; i < cut; i++)
                core[i] = data[distcopy[i].index];
            for(int i = cut; i < numObj; i++)
                noise[i-cut] = data[distcopy[i].index];
        } else{
            for(int i = 0; i <= cut; i++)
                core[i] = data[distcopy[i].index];
            for(int i = cut+1; i < numObj; i++)
                noise[i-(cut+1)] = data[distcopy[i].index];
            
        }
        Partitioning p  = new Partitioning(core, noise, precision);
        return p;
    }
    
    public Matrix matrix(double[][] m) {
        return new Matrix(m) ;
    }
    
    public Matrix columnVector(double[] v) {
        return new Matrix(v, v.length) ;
    }
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
    public double [] median(DataObject[] members) {
        int numObj= members.length ;
        int dim = members[0].coord.length ;
        double[] result = new double[dim];
        double [] sortarray = new double[numObj] ;
        for (int i=0 ; i<dim ; i++) {
            double median = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sortarray[j]=members[j].coord[i] ;
            Arrays.sort(sortarray) ;
            if (numObj==1)
                result[i]=sortarray[0] ;
            else if (numObj%2 == 0)
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
            else
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
        }
        return result ;
    }
    
    private double distance(double[] mean, double[] d){
        double sum = 0.0;
        for (int i = 0; i < d.length; i++){
            double dist = ((d[i] - mean[i])*(d[i] - mean[i]));
            sum = sum + dist;
        }
        return Math.sqrt(sum);
    }
    
//normalized mutual information
    private double mi(){
        double mi = 0.0;
        int numClasses = countClasses();
        for(int i = 0; i < numClasses; i++){
            int classMembers = countClassMembers(i+1);
            mi+= (double)classMembers/(double)numObj * Math.log((double)classMembers/(double)numObj)/Math.log(2);
            for(int j = 0; j < cl.size(); j++){
                int classClusterMembers = countClassClusterMembers(((Cluster)cl.elementAt(j)).ID, i+1);
                mi+= (double)classClusterMembers/(double)numObj * Math.log((double)classClusterMembers/(double)((Cluster)cl.elementAt(j)).numObj)/Math.log(2);
            }
        }
        return -mi;
    }
    
    private void printPercentage(){
        int numClasses = countClasses();
        int[] classMembers = new int[numClasses];
        for(int i = 0; i < numClasses; i++)
            classMembers[i] = countClassMembers(i+1);
        for(int i = 0; i < cl.size(); i++){
            Cluster akt = (Cluster)cl.elementAt(i);
            System.out.println("cluster " + akt.ID);
            for(int j = 0; j < numClasses; j++){
                System.out.println("class " + (j+1) + " " + ((double)countClassClusterMembers(akt.ID, j+1)/(double)classMembers[j]) + " percent ");
                System.out.println("class " + (j+1) + " " + countClassClusterMembers(akt.ID, j+1));
                System.out.println();
            //System.out.println("mi " + mi());
            }
        }
    }
    
    private int countClasses(){
        Vector clCount = new Vector();
        for(int i = 0; i < data.length; i++){
            Integer clID = (Integer)(data[i].classID);
            if(!clCount.contains(clID))
                clCount.add(clID);
        }
        return clCount.size();
    }
    
    private int countClassMembers(int classID){
        int clM =0;
        for(int i = 0; i < data.length; i++)
            if(data[i].classID == classID)
                clM++;
        return clM;
    }
    
    private int countClassClusterMembers(int clusterID, int classID){
        int clM =0;
        for(int i = 0; i < data.length; i++)
            if(data[i].classID == classID && data[i].clusterID == clusterID)
                clM++;
        return clM;
    }
}