/*
 * ClusteringEvaluation.java
 *
 * Created on November 15, 2006, 5:42 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package eval;
import java.util.Vector;
import Jama.*;
import compr.DataObject;
import clusterer.IcaCluster;
import gengauss.GenGauss;

/**
 *
 * @author Administrator
 */
public class ClusteringEvaluation {
    DataObject[] data;
    int[] clusterIDs;
    double[][] eigenvectors;
    double[][] eigenvalues;
    int dim;
    static int gauss = 0;
    static int uniform = 1;
    static int laplace = 2;
    static int numDistributions = 3; //gauss, uniform, laplace
    static int precision = 3;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    
    /** Creates a new instance of ClusteringEvaluation */
    public ClusteringEvaluation(DataObject[] data, int[] clusterIDs) {
        this.data = data;
        this.clusterIDs = clusterIDs;
        dim = data[0].coord.length;
        eigenvectors = new double[dim][dim];
        eigenvalues = new double[dim][dim];
    }
    
    public Matrix columnVector(double[] v) {
        return new Matrix(v, v.length) ;
    }
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
    public void evaluate(){
        System.out.println("Evaluating clustering...");
        double overallCost = 0.0;
        for(int i = 0; i < clusterIDs.length; i++){
            DataObject[] d = getObjects(clusterIDs[i]);
            //double costD = cost(d);
            System.out.println("Cluster " + clusterIDs[i] + " numObj " + d.length);
            double costD = costWithoutParameters(d);
            System.out.println("cost: " + costD);
            //System.out.println("Cluster " + clusterIDs[i] + " numObj " + d.length + " cost " + costD);
            overallCost+= costD;
        }
        System.out.println("COST OF CLUSTERING: " + overallCost);
    }
    
    //evaluate fit to generated pdf
    public void evaluateKl(){
        double[] overallLb = new double[dim]; //for KL-divergence
        double[] overallUb = new double[dim];
        for(int i = 0; i < dim; i++){
            overallLb[i] = Double.MAX_VALUE;
            overallUb[i] = -Double.MAX_VALUE;
        }
        for(int i = 0; i < data.length; i++){
            for(int j = 0; j < dim; j++){
                if(data[i].coord[j] < overallLb[j])
                    overallLb[j] = data[i].coord[j];
                if(data[i].coord[j] > overallUb[j])
                    overallUb[j] = data[i].coord[j];
            }
        }
        IcaCluster[] res = new IcaCluster[clusterIDs.length];
        GenGauss g = new GenGauss();
        for(int i = 0; i < clusterIDs.length; i++){
            DataObject[] d = getObjects(clusterIDs[i]);
            res[i] = bestModel(d, clusterIDs[i]);
        }
        //System.out.println("KL divergence  " + g.klEvaluation(res, overallLb, overallUb, 1000000, data.length));
        printInfo(res);
    }
    
    private void printInfo(IcaCluster[] res){
        System.out.println("PRINT INFO: ");
        for(int i = 0; i < res.length; i++){
            System.out.print("cluster " + res[i].ID + " numObj: " + res[i].numObj);
            System.out.println();
            System.out.print("realCenter: ");
            for(int j = 0; j < res[i].realCenter.length; j++)
                System.out.print(res[i].realCenter[j] +  " ");
            System.out.println();
            System.out.println("realRotation: ");
            Matrix m = new Matrix(res[i].realRotation);
            m.print(2, 3);
            System.out.println("beta: ");
            for(int j = 0; j < res[i].beta.length; j++)
                System.out.print(res[i].beta[j] +  " ");
            System.out.println();
        }
    }
    
    private DataObject[] getObjects(int ID){
        Vector<DataObject> d = new Vector<DataObject>();
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == ID)
                d.add(data[i]);
        DataObject[] obj = new DataObject[d.size()];
        for(int i = 0; i < d.size(); i++)
            obj[i] = (DataObject)d.elementAt(i);
        return obj;
    }
    
    public IcaCluster bestModel(DataObject[] obj, int clID){
        IcaCluster res = new IcaCluster();
        res.beta = new double[dim];
        res.realCenter = new double[dim];
        res.realRotation = new double[dim][dim];
        double betaG = 0.0;
        double betaL = 1.0;
        double betaU = -0.97;
        double cost_untrans = 0.0;
        double paramCostUntrans = dim * 32 * 2;
        double paramCostTrans = dim * 32 * 2 + dim * dim * 32;
        //Verteilung der einzelnen Koordinaten
        int[] untrans_dist = new int[dim];
        double[] mean = mean(obj);
        res.realCenter = mean;
        res.numObj = obj.length;
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(obj, mean, i);
            int min = minCost(aktCost);
            if(min == 0){
                untrans_dist[i] = gauss;
                cost_untrans += aktCost[0];
            }
            if(min == 1){
                untrans_dist[i] = uniform;
                cost_untrans += aktCost[1];
            }
            if(min == 2){
                untrans_dist[i] = laplace;
                cost_untrans += aktCost[2];
            }
        }
        double cost_trans = 0.0;
        int[] trans_dist = new int[dim];
        //PCA
        DataObject[] transData = pca(obj);
        double[] transMean = mean(transData);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(transData, transMean, i);
            int min = minCost(aktCost);
            if(min == 0){
                trans_dist[i] = gauss;
                cost_trans += aktCost[0];
            }
            if(min == 1){
                trans_dist[i] = uniform;
                cost_trans += aktCost[1];
            }
            if(min == 2){
                trans_dist[i] = laplace;
                cost_trans += aktCost[2];
            }
        }
        cost_trans += paramCostTrans; //parameter costs
        cost_untrans += paramCostUntrans;
        double cost = Math.min(cost_trans, cost_untrans);
        if(cost_trans < cost_untrans){
            //if(true){
            double[] uniformCorr = new double[dim];
            System.out.println("rotated ");
            double[][] sqrtEigenvalues = new double[dim][dim];
            for(int i = 0; i < dim; i++)
                if(eigenvalues[i][i] > 0.0)
                    sqrtEigenvalues[i][i] = Math.sqrt(1.0/eigenvalues[i][i]);
            Matrix sqrtEv = new Matrix(sqrtEigenvalues);
            res.realRotation = new Matrix(eigenvectors).times(sqrtEv).getArrayCopy();
            //print type of distribution and parameters
            for(int i = 0; i < dim; i++){
                System.out.println("coordinate: " + i);
                if (trans_dist[i] == gauss){
                    System.out.println(" gaussian distribution, parameters:" );
                    System.out.print(" mean: ");
                    for(int j = 0; j < dim; j++)
                        System.out.print(mean[j] + "\t");
                    System.out.println();
                    res.beta[i] = betaG;
                    uniformCorr[i] = 0.0;
                }
                if(trans_dist[i] == laplace){
                    System.out.println(" laplacian distribution, parameters:" );
                    System.out.print(" mean: ");
                    for(int j = 0; j < dim; j++)
                        System.out.print(mean[j] + "\t");
                    System.out.println();
                    res.beta[i] = betaL;
                    uniformCorr[i] = 0.0;
                }
                if(trans_dist[i] == uniform){
                    System.out.println(" uniform distribution, parameters:" );
                    double[] b = bounds(transData, i);
                    System.out.print("lower bound: " + b[0] + " upper bound " + b[1]);
                    uniformCorr[i] = (b[0]+b[1])/2.0;
                }
                System.out.println();
                double[] cc = rowVector(uniformCorr).times(new Matrix(eigenvectors).transpose()).getRowPackedCopy();
                for(int k = 0; k < dim; k++)
                    res.realCenter[k] += cc[k];
            }
        }else{
            System.out.println(" not rotated.");
            //print type of distribution and parameters
            for(int i = 0; i < dim; i++){
                System.out.print("coordinate: " + i);
                if (untrans_dist[i] == gauss){
                    System.out.println(" gaussian distribution, parameters:" );
                    System.out.print(" mean: ");
                    for(int j = 0; j < dim; j++)
                        System.out.print(mean[j] + "\t");
                    System.out.println();
                    res.beta[i] = betaG;
                    //variance, real rotation
                    double v = var(obj, mean, i);
                    if(v > 0.0)
                        res.realRotation[i][i] = 1.0/v;
                }
                if(untrans_dist[i] == laplace){
                    System.out.println(" laplacian distribution, parameters:" );
                    System.out.print(" mean: ");
                    for(int j = 0; j < dim; j++)
                        System.out.print(mean[j] + "\t");
                    System.out.println();
                    res.beta[i] = betaL;
                    //variance, real rotation
                    res.realRotation[i][i] = 1.0/var(obj, mean, i);
                }
                if(untrans_dist[i] == uniform){
                    System.out.println(" uniform distribution, parameters:" );
                    double[] b = bounds(obj, i);
                    System.out.print("lower bound: " + b[0] + " upper bound " + b[1]);
                    res.beta[i] = betaL;
                    res.realCenter[i] = (b[1]+b[0])/2.0;
                    //variance, real rotation
                    res.realRotation[i][i] = 1.0/var(obj, mean, i);
                }
                System.out.println();
            }
        }
        res.ID = clID;
        return res;
    }
    
    //to compare with ica-clustering
    private double costWithoutParameters(DataObject[] obj){
        double cost_untrans = 0.0;
        double paramCostUntrans = 0.0;
        double paramCostTrans = 0.0;
        //Verteilung der einzelnen Koordinaten
        int[] untrans_dist = new int[dim];
        double[] mean = mean(obj);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(obj, mean, i);
            int min = minCost(aktCost);
            if(min == 0){
                untrans_dist[i] = gauss;
                cost_untrans += aktCost[0];
            }
            if(min == 1){
                untrans_dist[i] = uniform;
                cost_untrans += aktCost[1];
            }
            if(min == 2){
                untrans_dist[i] = laplace;
                cost_untrans += aktCost[2];
            }
        }
        paramCostUntrans += dim * 32 * 2;
        paramCostUntrans += dim * 2;
        // Art Verteilung
        cost_untrans += paramCostUntrans;
        double cost_trans = 0.0;
        int[] trans_dist = new int[dim];
        //PCA
        DataObject[] transData = pca(obj);
        double[] transMean = mean(transData);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(transData, transMean, i);
            int min = minCost(aktCost);
            if(min == 0){
                trans_dist[i] = gauss;
                cost_trans += aktCost[0];
            }
            if(min == 1){
                trans_dist[i] = uniform;
                cost_trans += aktCost[1];
            }
            if(min == 2){
                trans_dist[i] = laplace;
                cost_trans += aktCost[2];
            }
        }
        paramCostTrans += dim*32*2 + dim*dim * 32;
        paramCostTrans += dim*2;
        //Parameter zhlen
        cost_trans += paramCostTrans;
        double costID = obj.length * Math.log((double)data.length/(double)obj.length)/Math.log(2);
        cost_trans += costID;
        cost_untrans += costID;
        double cost = Math.min(cost_trans, cost_untrans);
        if(cost_trans < cost_untrans){
            System.out.println("PCA: ");
            printEV(obj);
            //print type of distribution and parameters
            for(int i = 0; i < dim; i++){
                System.out.println("coordinate: " + i);
                if (trans_dist[i] == gauss){
                    System.out.println(" gaussian distribution, parameters:" );
                    System.out.print(" mean: ");
                    for(int j = 0; j < dim; j++)
                        System.out.print(mean[j] + "\t");
                    System.out.println();
                    System.out.print("var: " + stdv(transData, transMean, i));
                }
                if(trans_dist[i] == laplace){
                    System.out.println(" laplacian distribution, parameters:" );
                    System.out.print(" mean: ");
                    for(int j = 0; j < dim; j++)
                        System.out.print(mean[j] + "\t");
                    System.out.println();
                    System.out.print(" b: " +  stdv(transData, transMean, i)/Math.sqrt(2));
                }
                if(trans_dist[i] == uniform){
                    System.out.println(" uniform distribution, parameters:" );
                    double[] b = bounds(transData, i);
                    System.out.print("lower bound: " + b[0] + " upper bound " + b[1]);
                }
                System.out.println();
            }
        }else{
            System.out.println(" not rotated.");
            //print type of distribution and parameters
            cost += dim*dim*32;
            for(int i = 0; i < dim; i++){
                System.out.print("coordinate: " + i);
                if (untrans_dist[i] == gauss){
                    System.out.println(" gaussian distribution, parameters:" );
                    System.out.print(" mean: ");
                    for(int j = 0; j < dim; j++)
                        System.out.print(mean[j] + "\t");
                    System.out.println();
                    System.out.print("var: " + stdv(obj, mean, i));
                }
                if(untrans_dist[i] == laplace){
                    System.out.println(" laplacian distribution, parameters:" );
                    System.out.print(" mean: ");
                    for(int j = 0; j < dim; j++)
                        System.out.print(mean[j] + "\t");
                    System.out.println();
                    System.out.print(" b: " +  stdv(obj, mean, i)/Math.sqrt(2));
                }
                if(untrans_dist[i] == uniform){
                    System.out.println(" uniform distribution, parameters:" );
                    double[] b = bounds(obj, i);
                    System.out.print("lower bound: " + b[0] + " upper bound " + b[1]);
                }
                System.out.println();
            }
        }
        return cost;
    }
    
    //Kosten: Wie in eval.Clustering
    private double cost(DataObject[] obj){
        double cost_untrans = 0.0;
        double paramCostUntrans = 0.0;
        double paramCostTrans = 0.0;
        //Verteilung der einzelnen Koordinaten
        int[] untrans_dist = new int[dim];
        double[] mean = mean(obj);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(obj, mean, i);
            int min = minCost(aktCost);
            if(min == 0){
                untrans_dist[i] = gauss;
                cost_untrans += aktCost[0];
            }
            if(min == 1){
                untrans_dist[i] = uniform;
                cost_untrans += aktCost[1];
            }
            if(min == 2){
                untrans_dist[i] = laplace;
                cost_untrans += aktCost[2];
            }
        }
        paramCostUntrans += dim * 32 * 2;
        paramCostUntrans += dim * 2;
        // Art Verteilung
        cost_untrans += paramCostUntrans;
        
        double cost_trans = 0.0;
        int[] trans_dist = new int[dim];
        //PCA
        DataObject[] transData = pca(obj);
        double[] transMean = mean(transData);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(transData, transMean, i);
            int min = minCost(aktCost);
            if(min == 0){
                trans_dist[i] = gauss;
                cost_trans += aktCost[0];
            }
            if(min == 1){
                trans_dist[i] = uniform;
                cost_trans += aktCost[1];
            }
            if(min == 2){
                trans_dist[i] = laplace;
                cost_trans += aktCost[2];
            }
        }
        paramCostTrans += dim*32*2 + dim*dim * 32;
        paramCostTrans += dim*2;
        //Parameter zhlen
        cost_trans += paramCostTrans;
        //System.out.println("costTrans " + cost_trans + " costUntrans " + cost_untrans);
        double costID = obj.length * Math.log((double)data.length/(double)obj.length)/Math.log(2);
        cost_trans += costID;
        cost_untrans += costID;
        if(cost_trans < cost_untrans){
            System.out.println("rotated " + paramCostTrans);
        } else
            System.out.println("not rotated " + paramCostUntrans);
        double cost = Math.min(cost_trans, cost_untrans);
        return cost;
    }
    
    public double[] mean(DataObject[] o){
        double[] mean = new double[o[0].coord.length];
        for(int i = 0; i < o.length; i++)
            for(int j = 0; j < mean.length; j++){
            mean[j] += o[i].coord[j];
            }
        for(int j = 0; j < mean.length; j++)
            mean[j] /= o.length;
        return mean;
    }
    
    public double[] cost(DataObject[] o, double[] mean, int numCoord){
        double[] cost = new double[numDistributions];
        int dim = o[0].coord.length;
        double stdv = stdv(o, mean, numCoord);
        for(int i = 0; i < o.length; i++){
            //cost[0] += gaussCost(o[i].coord[numCoord], mean[numCoord], stdv, precision);
            cost[0] += gaussCostWithoutGrid(o[i].coord[numCoord], mean[numCoord], stdv, precision);
        }
        double[] bounds = bounds(o, numCoord);
        for(int i = 0; i < o.length; i++){
            //cost[1] += uniformCost(o[i].coord[numCoord], bounds[1], bounds[0], precision);
            cost[1] += uniformCostWithoutGrid(o[i].coord[numCoord], bounds[1], bounds[0], precision);
        }
        for(int i = 0; i < o.length; i++){
            //cost[2] += laplaceCost(o[i].coord[numCoord], mean[numCoord], stdv, precision);
            cost[2] += laplaceCostWithoutGrid(o[i].coord[numCoord], mean[numCoord], stdv, precision);
        }
        return cost;
    }
    private double uniformCost(double x, double lb, double ub, int precision){
        double prob ;
        if (Math.abs(ub-lb) < Math.pow(10, -precision))
            prob = 1 ;
        else
            prob = Math.pow(10, -precision) / Math.abs(ub - lb) ;
        return Math.log(1/prob)/Math.log(2);
    }
    
    private double uniformCostWithoutGrid(double x, double lb, double ub, int precision){
        double prob ;
        prob = Math.max(Double.MIN_VALUE, Math.abs(ub - lb)) ;
        return (-Math.log(prob)/Math.log(2));
    }
    
    private double laplaceCost(double x, double mean, double std, int precision){
        double b = std/Math.sqrt(2);
        if (b<Math.pow(10, -precision)/10)
            return 0 ;
        double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
        double prob = Math.max(Double.MIN_VALUE, Math.min(1.0, pdf * Math.pow(10, -precision)));
        return Math.log(1.0/prob)/Math.log(2);
    }
    
    private double laplaceCostWithoutGrid(double x, double mean, double std, int precision){
        double b = Math.max(std/Math.sqrt(2), Double.MIN_VALUE);
        double pdf = Math.min(1/(2*b)*Math.exp(-Math.abs(x-mean)/b), Double.MAX_VALUE);
        return (-Math.log(pdf)/Math.log(2));
    }
    
    private double gaussCost(double x, double mean, double std, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double var = Math.max(std*std, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*var)+precision*ln10+Math.pow(x-mean,2)/2/var)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    private double gaussCostWithoutGrid(double x, double mean, double std, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double var = Math.max(std*std, Double.MIN_VALUE) ;
        return (-(0.5*Math.log(2*Math.PI*var)*ln10+Math.pow(x-mean,2)/2/var))/ln2 ;
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    private double[] bounds(DataObject[] o, int numCoord){
        int dim = o[0].coord.length;
        double[] bounds = new double[2];
        for(int i = 0; i < bounds.length; i++){
            bounds[0] = -Double.MAX_VALUE; //upper bound
            bounds[1] = Double.MAX_VALUE; //lower bound
        }
        for(int i = 0; i < o.length; i++){
            if(o[i].coord[numCoord] > bounds[0])
                bounds[0] = o[i].coord[numCoord];
            if(o[i].coord[numCoord] < bounds[1])
                bounds[1] = o[i].coord[numCoord];
        }
        return bounds;
    }
    
    
    
    
    
    private int minCost(double[] cost){
        int minIndex = cost.length+1;
        double min = Double.MAX_VALUE;
        for(int i = 0; i < cost.length; i++)
            if(cost[i] < min){
            minIndex = i;
            min = cost[i];
            }
        return minIndex;
    }
    
    public void printEV(DataObject[] untransData){
        double[][] coordObj = new double[untransData.length][dim];
        for(int i = 0; i < untransData.length; i++){
            for(int j = 0; j < dim; j++)
                coordObj[i][j] = untransData[i].coord[j];
        }
        DataObject[] transData_pca = new DataObject[untransData.length];
        Matrix data = new Matrix(coordObj);
        double[] mean = mean(untransData);
        Matrix covar = new Matrix(covar(untransData, mean));
        EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
        Matrix vectors = eig.getV();
        Matrix values = eig.getD();
        System.out.println("eigenvalues");
        values.print(5, 3);
        System.out.println("eigenvectors");
        vectors.print(5, 3);
        double[][]dir = vectors.getArrayCopy();
        double[][] transCoord = new double[untransData.length][dim];
        for(int i = 0; i < untransData.length; i++){
            for(int j = 0; j < dim; j++){
                for(int k = 0; k < dim; k++){
                    transCoord[i][k] += (untransData[i].coord[j]-mean[j]) * dir[j][k];
                }
            }
        }
        for(int i = 0; i < untransData.length; i++){
            double[] co = new double[dim];
            for(int j = 0; j < dim; j++)
                co[j] = transCoord[i][j];
            transData_pca[i] = new DataObject(co, untransData[i].number, untransData[i].clusterID);
        }
    }
    
    public DataObject[] pca(DataObject[] untransData){
        double[][] coordObj = new double[untransData.length][dim];
        for(int i = 0; i < untransData.length; i++){
            for(int j = 0; j < dim; j++)
                coordObj[i][j] = untransData[i].coord[j];
        }
        DataObject[] transData_pca = new DataObject[untransData.length];
        Matrix data = new Matrix(coordObj);
        double[] mean = mean(untransData);
        Matrix covar = new Matrix(covar(untransData, mean));
        EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
        Matrix vectors = eig.getV();
        Matrix values = eig.getD();
        //System.out.println("EW");
        //values.print(5, 3);
        //System.out.println("EV");
        //vectors.print(5, 3);
        eigenvectors = vectors.getArrayCopy();
        eigenvalues = values.getArrayCopy();
        double[][]dir = vectors.getArrayCopy();
        double[][] transCoord = new double[untransData.length][dim];
        for(int i = 0; i < untransData.length; i++){
            for(int j = 0; j < dim; j++){
                for(int k = 0; k < dim; k++){
                    transCoord[i][k] += (untransData[i].coord[j]-mean[j]) * dir[j][k];
                }
            }
        }
        for(int i = 0; i < untransData.length; i++){
            double[] co = new double[dim];
            for(int j = 0; j < dim; j++)
                co[j] = transCoord[i][j];
            transData_pca[i] = new DataObject(co, untransData[i].number, untransData[i].clusterID);
        }
        
        return transData_pca;
    }
    
    private double stdv(DataObject[] o, double[] mean, int numCoord){
        double stdv = 0.0;
        for(int i = 0; i < o.length; i++)
            stdv += (o[i].coord[numCoord] - mean[numCoord]) * (o[i].coord[numCoord] - mean[numCoord]);
        return Math.sqrt(stdv/o.length);
    }
    
    private double var(DataObject[] o, double[] mean, int numCoord){
        double var = 0.0;
        for(int i = 0; i < o.length; i++)
            var += (o[i].coord[numCoord] - mean[numCoord]) * (o[i].coord[numCoord] - mean[numCoord]);
        return var/o.length;
    }
    
    private double[][] covar(DataObject[] data, double[] mean){
        double[][] cov = new double[dim][dim];
        for(int l = 0; l < data.length; l++)
            for(int i = 0; i < dim; i++)
                for(int j = 0; j < dim; j++)
                    cov[i][j]+= (data[l].coord[j]-mean[j])*(data[l].coord[i]-mean[i]);
        for(int i = 0; i < dim; i++)
            for(int j = 0; j < dim; j++)
                cov[i][j]/= data.length;
        return cov;
    }
    
}
