/*
 * Partitioning.java
 *
 * Created on 13. April 2006, 09:09
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package eval;
import compr.DataObject;
import Jama.*;
import java.util.*;


/**
 * Helper Class for Clustering.FilterClusterKMeans //TO DO: handle the case that noise or core is empty.
 * @author claudia
 */
public class Partitioning implements Cloneable{
    DataObject[] core;
    DataObject[] noise;
    int dim;
    int precision;
    double cost;
    //for noise
    double[] lb;
    double[] ub;
    //for core
    double[][] dir;
    double[] mean;
    double[] meanRot;
    double[] stdv;
    double[] stdvRot;
    double[] lbCore;
    double[] ubCore;
    double[] lbCoreRot;
    double[] ubCoreRot;
    static int numDistributions = 3; //gauss, uniform, laplace
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    
    
    /** Creates a new instance of Partitioning */
    public Partitioning() {
    }
    
    /**
     *
     * @param core
     * @param noise
     */
    public Partitioning(DataObject[] core, DataObject[] noise, int precision){
        this.core = core;
        this.noise = noise;
        this.precision = precision;
        dim = noise[0].coord.length;
        lb = new double[dim];
        ub= new double[dim];
        dir = new double[dim][dim];
        mean = new double[dim];
        meanRot = new double[dim];
        stdv = new double[dim];
        stdvRot = new double[dim];
        lbCore = new double[dim];
        ubCore = new double[dim];
        lbCoreRot = new double[dim];
        ubCoreRot = new double[dim];
        cost = 0.0;
    }
    
    public Object clone(){
        try{
            return super.clone();
        } catch (CloneNotSupportedException e){
            throw new InternalError();
        }
    }
    
    
    public void update(){
        setNoiseParameters();
        setCoreParameters();
    }
    
    public void updateIter(){
        setCoreParameters();
    }
    
    /**
     * Assigns the objects of a partitioning to core and noise and returns the resulting partitioning. NOTICE: The resulting partitioning needs to be updated.
     * @return resulting partitioning
     */
    public Partitioning assignPoints(){
        Vector<DataObject> coreObj = new Vector<DataObject>();
        Vector<DataObject> noiseObj = new Vector<DataObject>();
        DataObject akt;
        int numObj = core.length + noise.length;
        for(int i = 0; i < numObj; i++){
            if(i<core.length)
                akt = core[i];
            else
                akt = noise[i-core.length];
            if(akt.number == 0)
                System.out.println("mist");
            double noiseCost = 0.0;
            double coreCost = 0.0;
            //cost for coding akt as noise
            
            for(int j = 0; j < dim; j++){
                noiseCost += uniformCost(0, lb[j], ub[j], precision);
              //System.out.println("uk " + uniformCost(0, lb[j], ub[j], precision));
            }
            //cost for coding akt as core
            double[] aktRot = rowVector(akt.coord).times(new Matrix(dir)).getColumnPackedCopy();
            for(int j = 0; j < dim; j++){
                //double costNotRot = Math.min(Math.min(uniformCost(0, lbCore[j], ubCore[j], precision), gaussCost(akt.coord[j], mean[j], stdv[j], precision)), laplaceCost(akt.coord[j], mean[j], stdv[j], precision));
                double costNotRot = Math.min(gaussCost(akt.coord[j], mean[j], stdv[j], precision), laplaceCost(akt.coord[j], mean[j], stdv[j], precision));
                //double costRot = Math.min(Math.min(uniformCost(0, lbCoreRot[j], ubCoreRot[j], precision), gaussCost(aktRot[j], meanRot[j], stdvRot[j], precision)), laplaceCost(aktRot[j], meanRot[j], stdvRot[j], precision));
                double costRot = Math.min(gaussCost(aktRot[j], meanRot[j], stdvRot[j], precision), laplaceCost(aktRot[j], meanRot[j], stdvRot[j], precision));
                coreCost += Math.min(costNotRot, costRot);
            }
             System.out.println("nk " + noiseCost + " ck " + coreCost);
            if(noiseCost < coreCost)
                noiseObj.addElement(akt);
            else
                coreObj.addElement(akt);
        }
        DataObject[] coreNew = new DataObject[coreObj.size()];
        DataObject[] noiseNew = new DataObject[noiseObj.size()];
        for (int i = 0; i < coreObj.size(); i++)
            coreNew[i] = (DataObject)coreObj.elementAt(i);
        for(int i = 0; i < noiseObj.size(); i++)
            noiseNew[i] = (DataObject)noiseObj.elementAt(i);
        Partitioning pNew = (Partitioning)this.clone();
        pNew.core = coreNew;
        pNew.noise = noiseNew;
        return pNew;
    }
    
    public double setCost(){
        double noiseCost = 0.0;
        double coreCost = 0.0;
        double coreCostNotRot = 0.0;
        double CoreCostRot = 0.0;
        //cost for noise objects
        if(noise.length == 0)
            noiseCost = 0.0;
        else{
            for(int i = 0; i < dim; i++)
                noiseCost += uniformCost(0, lb[i], ub[i], precision);
            noiseCost *= noise.length;
        }
        //cost for core objects
        if(core.length == 0){
            coreCost = 0.0;
        } else{
            for(int i = 0; i < dim; i++){
                double[] costNotRot = cost(i, false);
                double minCostNotRot = Math.min(Math.min(costNotRot[0], costNotRot[1]), costNotRot[2]);
                double[] costRot = cost(i, true);
                double minCostRot = Math.min(Math.min(costRot[0], costRot[1]), costRot[2]);
                coreCost += Math.min(minCostNotRot, minCostRot);
            }
        }
        cost = noiseCost+coreCost;
        return cost;
    }
    
    
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    private void setNoiseParameters(){
        //lb and ub
        for(int i = 0; i < dim; i++){
            lb[i] = Double.MAX_VALUE;
            ub[i] = -Double.MAX_VALUE;
        }
        for(int i = 0; i < noise.length; i++){
            for(int j = 0; j < dim; j++){
                if(noise[i].coord[j] < lb[j])
                    lb[j] = noise[i].coord[j];
                if(noise[i].coord[j] > ub[j])
                    ub[j] = noise[i].coord[j];
            }
        }
    }
    
    private void setCoreParameters(){
        int numObj = core.length;
        //dir
        double[] center = median(core);
        double[][] cov = covarByMedian(core, center);
        EigenvalueDecomposition e = new EigenvalueDecomposition(new Matrix(cov)) ;
        dir = e.getV().getArrayCopy();
        //mean
        for(int i = 0; i < numObj; i++)
            for(int j = 0; j < dim; j++)
                mean[j]+= core[i].coord[j]/(double)numObj;
        //stdv
        for(int i = 0; i < numObj; i++)
            for(int j = 0; j < dim; j++)
                stdv[j]+= (core[i].coord[j]-mean[j]) * (core[i].coord[j]-mean[j]);
        for(int j = 0; j < dim; j++)
            stdv[j] = Math.sqrt(stdv[j])/(double)numObj;
        //lbCore and ubCore
        for(int i = 0; i < dim; i++){
            lbCore[i] = Double.MAX_VALUE;
            ubCore[i] = -Double.MAX_VALUE;
        }
        for(int i = 0; i < numObj; i++){
            for(int j = 0; j < dim; j++){
                if(core[i].coord[j] < lbCore[j])
                    lbCore[j] = core[i].coord[j];
                if(core[i].coord[j] > ubCore[j])
                    ubCore[j] = core[i].coord[j];
            }
        }
        //meanRot
        System.out.println("hier");
        double [][] rot = new double [numObj][dim] ;
        for (int i=0 ; i<numObj ; i++)
            rot[i] = rowVector(core[i].coord).times(new Matrix(dir)).getColumnPackedCopy();
        for(int i = 0; i < numObj; i++)
            for(int j = 0; j < dim; j++)
                meanRot[j]+= rot[i][j]/(double)numObj;
        //stdvRot
        for(int i = 0; i < numObj; i++)
            for(int j = 0; j < dim; j++)
                stdvRot[j]+= (rot[i][j]-meanRot[j]) * (rot[i][j]-meanRot[j]);
        for(int j = 0; j < dim; j++)
            stdvRot[j] = Math.sqrt(stdvRot[j])/(double)numObj;
        //lbCoreRot and ubCoreRot
        for(int i = 0; i < dim; i++){
            lbCoreRot[i] = Double.MAX_VALUE;
            ubCoreRot[i] = -Double.MAX_VALUE;
        }
        for(int i = 0; i < numObj; i++){
            for(int j = 0; j < dim; j++){
                if(rot[i][j] < lbCoreRot[j])
                    lbCoreRot[j] = rot[i][j];
                if(rot[i][j] > ubCoreRot[j])
                    ubCoreRot[j] = rot[i][j];
            }
        }
    }
    
    private double[] cost(int numCoord, boolean rotated){
        double[] cost = new double[numDistributions];
        for(int i = 0; i < core.length; i++){
            if(rotated){
                cost[0] += gaussCost(core[i].coord[numCoord], mean[numCoord], stdv[numCoord], precision);
                cost[1] += uniformCost(core[i].coord[numCoord], lbCore[numCoord], ubCore[numCoord], precision);
                cost[2] += laplaceCost(core[i].coord[numCoord], mean[numCoord], stdv[numCoord], precision);
            } else{
                cost[0] += gaussCost(core[i].coord[numCoord], meanRot[numCoord], stdvRot[numCoord], precision);
                cost[1] += uniformCost(core[i].coord[numCoord], lbCoreRot[numCoord], ubCoreRot[numCoord], precision);
                cost[2] += laplaceCost(core[i].coord[numCoord], mean[numCoord], stdvRot[numCoord], precision);
            }
        }
        return cost;
    }
    
    private Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
    private double [] median(DataObject[] members) {
        int numObj= members.length ;
        int dim = members[0].coord.length ;
        double[] result = new double[dim];
        double [] sortarray = new double[numObj] ;
        for (int i=0 ; i<dim ; i++) {
            double median = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sortarray[j]=members[j].coord[i] ;
            Arrays.sort(sortarray) ;
            if (numObj==1)
                result[i]=sortarray[0] ;
            else if (numObj%2 == 0)
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
            else
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
        }
        return result ;
    }
    
    private double[][] covarByMedian(DataObject[] data, double[] mean) {
        int numObj= data.length ;
        int dim = data[0].coord.length ;
        double [] sortarray = new double[numObj] ;
        double[][] cov = new double[dim][dim];
        if (numObj <5){
            for (int i=0 ; i<dim ; i++)
                cov[i][i]=1.0 ;
            return cov ;
        }
        for(int l1 = 0; l1 < dim; l1++) {
            for(int l2 = 0; l2 < dim; l2++) {
                for(int j = 0; j < numObj; j++)
                    sortarray[j] = (data[j].coord[l1]-mean[l1])*(data[j].coord[l2]-mean[l2]) ;
                
                Arrays.sort(sortarray) ;
                if (numObj%2 == 0)
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3
                            ;
            }
            
        }
        double maxdiff = 0.0 ;
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                sum += Math.abs(cov[i][j]) ;
            double diff = 1.1*sum - 2.1*cov[i][i] ;
            if (diff>maxdiff)
                maxdiff = diff ;
        }
        if (maxdiff > 0.0)
            for (int i=0 ; i<dim ; i++)
                cov[i][i] += maxdiff ;
        
        return cov ;
    }
    
    private double laplaceCost(double x, double mean, double std, int precision){
        double b = std/Math.sqrt(2);
        if (b<Math.pow(10, -precision)/10)
            return 0 ;
        double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
        double prob = Math.max(Double.MIN_VALUE, Math.min(1.0, pdf * Math.pow(10, -precision)));
        //return 100000;
        return Math.log(1.0/prob)/Math.log(2);
    }
    
    
    private double gaussCost(double x, double mean, double std, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double var = Math.max(std*std, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*var)+precision*ln10+Math.pow(x-mean,2)/2/var)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    private double uniformCost(double x, double lb, double ub, int precision){
        double prob ;
        if (Math.abs(ub-lb) < Math.pow(10, -precision))
            prob = 1 ;
        else
            prob = Math.pow(10, -precision) / Math.abs(ub - lb) ;
        return Math.log(1/prob)/Math.log(2);
    }
    
    
}


