/*
 * Evaluation.java
 *
 * Created on October 4, 2006, 12:38 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package filterCluster;


import compr.DataObject;
import visu.VisuInfo;
import visu.VisuSimple;
import Jama.*;

/**
 * Evaluation of filtering resuts: costs and percentage of correctly identified points.
 * @author Administrator
 */

public class Evaluation {
    DataObject[] data;
    DataObject[] clusterPts;
    DataObject[] noisePts;
    int numObj;
    int dim;
    int numCluster;      //number of cluster points
    int numNoise;        //number of noise points
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    static int precision = 3;
    static int numDistributions = 3;
    static int gauss = 0;
    static int uniform = 1;
    static int laplace = 2;
    
    /**
     * Creates a new instance of Evaluation
     * @param data DataObject[] with clusterIDs:
     * 0: cluster, 1: noise
     */
    public Evaluation(DataObject[] data) {
        this.data = data;
        numObj = data.length;
        dim = data[0].coord.length;
    }
    
    
    public void evaluateFiltering(){
        splitData();
        //evaluate cost
        double overallCost = 0.0;
        System.out.println("Best Models: ");
        System.out.println("CLUSTER");
        overallCost += coordCost(clusterPts);
        System.out.println("NOISE");
        overallCost += coordCost(noisePts);
        System.out.println("overall cost " + overallCost);
        System.out.println("correctly identified: " + correctlyIdentified());
//        VisuSimple vi = new VisuSimple(data);
//        vi.setSize(600,600);
//        vi.setLocation(200,0);
//        vi.setVisible(true);
        
        VisuInfo vi = new VisuInfo(data, "result");
        vi.setSize(600,600);
        vi.setLocation(200,0);
        vi.setVisible(true);
        
    }
    
    private double correctlyIdentified(){
        int clusterAssignedToNoise = 0;
        int noiseAssignedToCluster = 0;
        int clusterCounter = 0;
        int noiseCounter = 0;
        for(int i = 0; i < numObj; i++){
            if(data[i].classID == 0){
                clusterCounter++;
            if(data[i].clusterID == 1)
                clusterAssignedToNoise++;
            }
            if(data[i].classID == 1){
                noiseCounter++; 
            if(data[i] .clusterID == 0)
                noiseAssignedToCluster++;
            }
        }
        System.out.println(noiseAssignedToCluster + " of " + noiseCounter + " noise points assigned to cluster.");
        System.out.println(clusterAssignedToNoise + " of " + clusterCounter + " cluster points assigned to noise.");
        return 1.0-(double)(clusterAssignedToNoise + noiseAssignedToCluster)/(double)data.length;
    }
    
    //Art der Verteilung entscheiden
    private double coordCost(DataObject[] obj){
        double cost_untrans = 0.0;
        //distribution functions of the single coordinates
        int[] untrans_dist = new int[dim];
        double[] mean = mean(obj);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(obj, mean, i);
            int min = minCost(aktCost);
            if(min == 0){
                untrans_dist[i] = gauss;
                cost_untrans += aktCost[0];
            }
            if(min == 1){
                untrans_dist[i] = uniform;
                cost_untrans += aktCost[1];
            }
            
            if(min == 2){
                untrans_dist[i] = laplace;
                cost_untrans += aktCost[2];
            }
        }
        //add parameter cost
        cost_untrans += dim *32*2;
        double cost_trans = 0.0;
        int[] trans_dist = new int[dim];
        //PCA
        DataObject[] transData = pca(obj);
        double[] transMean = mean(transData);
        for(int i = 0; i < dim; i++){
            double[] aktCost = cost(transData, transMean, i);
            int min = minCost(aktCost);
            if(min == 0){
                trans_dist[i] = gauss;
                cost_trans += aktCost[0];
            }
            if(min == 1){
                trans_dist[i] = uniform;
                cost_trans += aktCost[1];
            }
            if(min == 2){
                trans_dist[i] = laplace;
                cost_trans += aktCost[2];
            }
        }
        //add parameter cost
        cost_trans += dim*32*2 + dim*dim * 32;
        if(cost_trans < cost_untrans){
            System.out.println("rotated cluster ");
            System.out.println("distribution functions: ");
            for(int i = 0; i < dim; i++){
                if (trans_dist[i] == gauss)
                    System.out.print("g \t");
                if (trans_dist[i] == uniform)
                    System.out.print("u \t");
                if (trans_dist[i] == laplace)
                    System.out.print("l \t");
            }
            System.out.println();
            System.out.println("cost " + cost_trans);
            return cost_trans;
        } else{
            System.out.println("not rotated ");
            System.out.println("distribution functions: ");
            for(int i = 0; i < dim; i++){
                if (untrans_dist[i] == gauss)
                    System.out.print("g \t");
                if (untrans_dist[i] == uniform)
                    System.out.print("u \t");
                if (untrans_dist[i] == laplace)
                    System.out.print("l \t");
            }
            System.out.println();
            System.out.println("cost " + cost_untrans);
            return cost_untrans;
        }
    }
    
    
    
    /**
     * Split up into cluster and noise points.
     */
    private void splitData(){
        //count cluster and noise points
        int clusterCounter = 0;
        int noiseCounter = 0;
        for(int i = 0; i < numObj; i++){
            if(data[i].clusterID == 0)
                clusterCounter++;
            if(data[i].clusterID == 1)
                noiseCounter++;
        }
        numCluster = clusterCounter;
        clusterPts = new DataObject[clusterCounter];
        numNoise = noiseCounter;
        noisePts = new DataObject[noiseCounter];
        System.out.println(clusterCounter + " cluster points.");
         System.out.println(noiseCounter + " noise points.");
        //write objects into cluster and noise array
        int indexC = 0;
        int indexN = 0;
        for(int i = 0; i < numObj; i++){
            if(data[i].clusterID == 0){
                clusterPts[indexC] = data[i];
                indexC++;
            }
            if(data[i].clusterID == 1){
                noisePts[indexN] = data[i];
                indexN++;
            }
        }
    }
    
    
    //returns cost to code the numCoord-th coordinate according to the different distributions: cost[0]: gaussian, cost[1]: uniform, cost[2]: laplacian
    public double[] cost(DataObject[] o, double[] mean, int numCoord){
        double[] cost = new double[numDistributions];
        int dim = o[0].coord.length;
        double stdv = stdv(o, mean, numCoord);
        for(int i = 0; i < o.length; i++){
            cost[0] += gaussCost(o[i].coord[numCoord], mean[numCoord], stdv, precision);
        }
        double[] bounds = bounds(o, numCoord);
        for(int i = 0; i < o.length; i++){
            cost[1] += uniformCost(o[i].coord[numCoord], bounds[1], bounds[0], precision);
        }
        for(int i = 0; i < o.length; i++){
            cost[2] += laplaceCost(o[i].coord[numCoord], mean[numCoord], stdv, precision);
        }
        return cost;
    }
    
    //bounds[0]: upper bound, bounds[1]: lower bound
    private double[] bounds(DataObject[] o, int numCoord){
        int dim = o[0].coord.length;
        double[] bounds = new double[2];
        for(int i = 0; i < bounds.length; i++){
            bounds[0] = -Double.MAX_VALUE;
            bounds[1] = Double.MAX_VALUE;
        }
        for(int i = 0; i < o.length; i++){
            for(int j = 0; j < dim; j++){
                if(o[i].coord[numCoord] > bounds[0])
                    bounds[0] = o[i].coord[j];
                if(o[i].coord[numCoord] < bounds[1])
                    bounds[1] = o[i].coord[j];
            }
        }
        return bounds;
    }
    
    
    private double laplaceCost(double x, double mean, double std, int precision){
        double b = std/Math.sqrt(2);
        if (b<Math.pow(10, -precision)/10)
            return 0 ;
        double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
        double prob = Math.max(Double.MIN_VALUE, Math.min(1.0, pdf * Math.pow(10, -precision)));
        //return 100000;
        return Math.log(1.0/prob)/Math.log(2);
    }
    
    
    private double gaussCost(double x, double mean, double std, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double var = Math.max(std*std, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*var)+precision*ln10+Math.pow(x-mean,2)/2/var)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    private double uniformCost(double x, double lb, double ub, int precision){
        double prob ;
        if (Math.abs(ub-lb) < Math.pow(10, -precision))
            prob = 1 ;
        else
            prob = Math.pow(10, -precision) / Math.abs(ub - lb) ;
        return Math.log(1/prob)/Math.log(2);
    }
    
    private double[] mean(DataObject[] o){
        double[] mean = new double[o[0].coord.length];
        for(int i = 0; i < o.length; i++)
            for(int j = 0; j < mean.length; j++){
            mean[j] += o[i].coord[j];
            }
        for(int j = 0; j < mean.length; j++)
            mean[j] /= o.length;
        return mean;
    }
    
    //stdv of coordinate numCoord
    private double stdv(DataObject[] o, double[] mean, int numCoord){
        double stdv = 0.0;
        for(int i = 0; i < o.length; i++)
            stdv += (o[i].coord[numCoord] - mean[numCoord]) * (o[i].coord[numCoord] - mean[numCoord]);
        return Math.sqrt(stdv/(double)o.length);
    }
    
    private int minCost(double[] cost){
        int minIndex = cost.length+1;
        double min = Double.MAX_VALUE;
        for(int i = 0; i < cost.length; i++)
            if(cost[i] < min){
            minIndex = i;
            min = cost[i];
            }
        return minIndex;
    }
    
    public DataObject[] pca(DataObject[] untransData){
        double[][] coordObj = new double[untransData.length][dim];
        for(int i = 0; i < untransData.length; i++){
            for(int j = 0; j < dim; j++)
                coordObj[i][j] = untransData[i].coord[j];
        }
        DataObject[] transData_pca = new DataObject[untransData.length];
        Matrix data = new Matrix(coordObj);
        double[] mean = mean(untransData);
        Matrix covar = new Matrix(covar(untransData, mean));
        EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
        Matrix vectors = eig.getV();
        Matrix values = eig.getD();
        //System.out.println("EW");
        //values.print(5, 3);
        //System.out.println("EV");
        //vectors.print(5, 3);
        double[][] dir = vectors.getArrayCopy();
        double[][] transCoord = new double[untransData.length][dim];
        for(int i = 0; i < untransData.length; i++){
            for(int j = 0; j < dim; j++){
                for(int k = 0; k < dim; k++){
                    transCoord[i][k] += (untransData[i].coord[j]-mean[j]) * dir[j][k];
                }
            }
        }
        for(int i = 0; i < untransData.length; i++){
            double[] co = new double[dim];
            for(int j = 0; j < dim; j++)
                co[j] = transCoord[i][j];
            transData_pca[i] = new DataObject(co, untransData[i].number, untransData[i].clusterID);
        }
        
        return transData_pca;
    }
    
    private double[][] covar(DataObject[] data, double[] mean){
        double[][] cov = new double[dim][dim];
        for(int l = 0; l < data.length; l++)
            for(int i = 0; i < dim; i++)
                for(int j = 0; j < dim; j++)
                    cov[i][j]+= (data[l].coord[j]-mean[j])*(data[l].coord[i]-mean[i]);
        for(int i = 0; i < dim; i++)
            for(int j = 0; j < dim; j++)
                cov[i][j]/= data.length;
        return cov;
    }
    
}
