/*
 * GaussianFilter.java
 *
 * Created on 11. September 2006, 11:26
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package filterCluster;
import compr.DataObject;
import eval.Cluster;
import utils.Statistics;
import visu.VisuSimple;
import io.IO;

/**
 * For jounal version: Cost based k-means style filtering algorithm. Input: DataObject[] of cluster points.
 * As in paper version: no costs for cluster ID, grid costs, cost unsplit: uniform cost. 
 * Initialisation: mean: median. Variance cluster: addVar/dim, variance noise: addVar * dim;
 * @author boehm
 */
public class Filter {
    DataObject[] d;
    int numObj;
    int dim;
    double[] mean; //mean of cluster and noise;
    double[] clusterVariance;
    double[] noiseVariance;
    static int numDistributions = 3; //gauss, uniform, laplace
    static int gauss = 0;
    static int uniform = 1;
    static int laplace = 2;
    int precision;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    Cluster unfiltered;
    Cluster filtered;
    Cluster noise;
    
    
    
    /**
     * Creates a new instance of Filter
     * @param d DataObjects of the cluster to filter
     */
    public Filter(DataObject[] d, int precision) {
        this.d = d;
        this.precision = precision;
        numObj = d.length;
        dim = d[0].coord.length;
        mean = new double[dim];
        clusterVariance = new double[dim];
        noiseVariance = new double[dim];
    }
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        int dim = 2;
        //Verzeichnis Ica-test
        //IO ea = new IO("icaQuadratNoise5.txt", dim);
        //IO ea = new IO("2cluster.txt", dim); //ica uniform am besten.
        //IO ea = new IO("icaTest1.txt", dim); //am besten ist laplacian ica?
        //IO ea = new IO("rotatedLpl.txt", dim); //gut: am besten ist laplacian ica
        //draussen
        //String dir = "cor1noise10.txt"; //noise wird nicht ausgesondert
        //String dir = "cor1asymmout.txt"; //ok
        //String dir = "cor1asymmoutn.txt"; //noise nicht ausreichend ausgefiltert
        //String dir = "clust50.txt"; // ok
        //String dir = "cor1.txt"; //ok
        //String dir = "cor1g1n1.txt";  //noise wird nicht ausgesondert
        //String dir = "cor1noise.txt"; //ok
        //String dir = "cor1vielnoise.txt"; //ok: unmglich
        //String dir = "cor20noise5.txt"; //ok
        //String dir = "g100.txt"; //ok
        //String dir = "g1000.txt"; //ok
        //String dir = "gauss2000.txt"; //ok
        //String dir = "laplace200.txt"; //? laplace-Verteilung aufnehmen
        //String dir = "noise500.txt"; //ok
        //String dir = "outlierTest.txt"; // ok
        String dir = "laplace2000noise100.txt";
        //String dir = "t2.txt"; // ok
        IO ea = new IO(dir, dim);
        DataObject[] d = ea.readFile();
        for(int i = 0; i < d.length; i++)
            d[i].clusterID = -1;
        Filter f = new Filter(d, 3);
        f.run();
    }
    
    /**
     * filter these objects
     */
    public void run(){
        init();
        boolean clusterChanged = true;
        int iter = 0;
        while(clusterChanged){
            System.out.println("iter " + iter);
            iter++;
            clusterChanged = assignPoints();
            update();
        }
        VisuSimple visu = new VisuSimple(d);
        visu.setSize(600,600);
        visu.setLocation(100,0);
        visu.setVisible(true);
    }
    
    
    /**
     * initialisation: mean: median, variance cluster: sum variances/dim, variances noise: sum variances * dim
     */
    public void init(){
        Statistics s = new Statistics(d.length, d[0].coord.length);
        double[] m = s.mean(d);
        double[] sigma = new double[dim];
        double addVar = 0.0;
        for(int i = 0; i < dim; i++){
            sigma[i] = s.sigma(d, m, i);
            addVar += sigma[i];
        }
        //initialize mean and cluster- and noise variances.
        mean = s.median(d);
        double varCluster = addVar/(double)dim;
        double varNoise = addVar * (double)dim;
        for(int i = 0; i < dim; i++){
            clusterVariance[i] = varCluster;
            noiseVariance[i] = varNoise;
        }
    }
    
    
    /**
     * ClusterID 1: cluster, 0: noise.
     * @return boolean clusterChanged
     */
    private boolean assignPoints(){
        boolean clusterChanged = false;
        for(int i = 0; i < numObj; i++){
            double clusterCost = 0.0;
            for(int j = 0; j < dim; j++)
                clusterCost += gaussCost(d[i].coord[j], clusterVariance[j], mean[j], precision);
            double noiseCost = 0.0;
            for(int j = 0; j < dim; j++)
                noiseCost += gaussCost(d[i].coord[j], noiseVariance[j], mean[j], precision);
            if(clusterCost < noiseCost){
                if(d[i].clusterID != 1){
                    d[i].clusterID = 1;
                    clusterChanged = true;
                }
            } else{
                if(d[i].clusterID != 0){
                    d[i].clusterID = 0;
                    clusterChanged = true;
                }
            }
        }
        return clusterChanged;
    }
    
    
    /**
     * update mean of cluster, noise cluster has the same mean as the cluster. Update cluster and noise variances.
     */
    private void update(){
        //update mean as mean of the cluster points (clusterID 1)
        double[] meanNew = new double[dim];
        int counter = 0;
        for(int i = 0; i < d.length; i++){
            if(d[i].clusterID == 1){
                for(int j = 0; j < dim; j++){
                    meanNew[j] += d[i].coord[j];
                }
                counter++;
            }
        }
        for(int j = 0; j < dim; j++)
            meanNew[j]/=(double)counter;
        mean = meanNew;
        //update variances
        //cluster points
        DataObject[] clusterPts = new DataObject[counter];
        DataObject[] noisePts = new DataObject[numObj-counter]; //Fehlt: Behandlung: eines von beidem wird leer.
        int clusterCounter = 0;
        int noiseCounter = 0;
        for(int i = 0; i < numObj; i++){
            if(d[i].clusterID == 1){
                clusterPts[clusterCounter] = d[i];
                clusterCounter++;
            } else{
                noisePts[noiseCounter] = d[i];
                noiseCounter++;
            }
        }
        double[] clusterVarianceNew = new double[dim];
        double[] noiseVarianceNew = new double[dim];
        Statistics sCluster = new Statistics(clusterCounter, dim);
        Statistics sNoise = new Statistics(noiseCounter, dim);
        for(int i = 0; i < dim; i++)
            clusterVarianceNew[i] = sCluster.sigma(clusterPts, mean, i);
        for(int i = 0; i < dim; i++)
            noiseVarianceNew[i] = sNoise.sigma(noisePts, mean, i);
    }
    
    
    /**
     * gaussian cost of one coordinate without grid. THIS METHOD IS NOT USED.
     * @param coord index of coordinate
     * @param sigma variance
     * @param mean mean
     * @return cost
     */
    private double gaussianCost(double coord, double sigma, double mean){
        double factor = 1/(Math.sqrt(2 * Math.PI * sigma));
        return -Math.log(Math.exp(-(coord - mean)*(coord - mean)/(2 * sigma)) * factor)/Math.log(2.0);
    }
    
    
    /**
     * Huffman coded cost of clusterID for all points. THIS METHOD IS NOT USED.
     * @param numClusterObj number of cluster objects
     * @param numNoiseObj number of noise objects
     * @return cost
     */
    private double costID(int numClusterObj, int numNoiseObj){
        return numClusterObj * Math.log((double)numClusterObj/(double)numObj)/Math.log(2) + numNoiseObj * Math.log((double)numNoiseObj/(double)numObj)/Math.log(2);
    }
    
    
   private double gaussCost(double x, double mean, double var, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double va = Math.max(var, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*va)+precision*ln10+Math.pow(x-mean,2)/2/va)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    private double laplaceCost(double x, double mean, double var, int precision){
        double stdv = Math.sqrt(var);
        double b = stdv/Math.sqrt(2);
        if (b<Math.pow(10, -precision)/10)
            return 0 ;
        double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
        double prob = Math.max(Double.MIN_VALUE, Math.min(1.0, pdf * Math.pow(10, -precision)));
        //return 100000;
        return Math.log(1.0/prob)/Math.log(2);
    }
    
    /**
     * method from eval.clustering
     */
      private double uniformCost(double x, double lb, double ub, int precision){
        double prob ;
        if (Math.abs(ub-lb) < Math.pow(10, -precision))
            prob = 1 ;
        else
            prob = Math.pow(10, -precision) / Math.abs(ub - lb) ;
        return Math.log(1/prob)/Math.log(2);
    }
    
    
    
    /**
     * Determine the minimum Cost for all objects in one cluster: uniform distribution only
     */
    private double costUnfiltered(){
        double uniformCost = 0.0;
        double [] lb = d[0].coord.clone() ;
        double [] ub = d[0].coord.clone() ;
        for (int i=1 ; i<numObj ; i++) {
            for (int j=0 ; j<dim ; j++) {
                lb[j] = Math.min(lb[j], d[i].coord[j]) ;
                ub[j] = Math.max(ub[j], d[i].coord[j]) ;
            }
        }
         for (int j=0 ; j<dim ; j++)
            uniformCost += numObj*uniformCost(0, lb[j], ub[j], precision) ;
        return uniformCost;
    }
}
