/*
 * RunKMeans.java
 *
 * Created on October 5, 2006, 8:15 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package filterCluster;
import compr.DataObject;
import clusterer.KMCost;

/**
 * Run filtering algorithms and choose the best result:
 * K-Means style filtering algorithms rotated/not rotated and different distribution functions, FilterSort rotated and not rotated.
 * @author Administrator
 */
public class RunFilter {
    DataObject[] data;
    int dim;
    int numObj;
    static int precision = 3;
    static int numDistributions = 3;
    static int gauss = 1;
    static int laplace = 2;
    boolean verbose = true;
    
    /**
     * Creates a new instance.
     * @param data DataObjects to be filtered
     */
    public RunFilter(DataObject[] data) {
        this.data = data;
        dim = data[0].coord.length;
    }
    
    
    
    /**
     * run k-means style filtering algorithm and select the best result.
     */
    public void runKMeans(){
        //init runtime
        long start = System.currentTimeMillis();
        //run all variants
        double minCost = Double.MAX_VALUE;
        int minIndex = -1;
        FilterKM g = new FilterKM(copyObj(), 1, precision, false, gauss);
        g.run();
        double gaussCost = g.getCost() + parameterCost(false);
        if(gaussCost < minCost){
            minCost = gaussCost;
            minIndex = 0;
        }
        FilterKM g_rot = new FilterKM(copyObj(), 1, precision, true, gauss);
        g_rot.run();
        double gaussCost_rot = g_rot.getCost() + parameterCost(true);
        if(gaussCost_rot < minCost){
            minCost = gaussCost_rot;
            minIndex = 1;
        }
        FilterKM l = new FilterKM(copyObj(), 1, precision, false, laplace);
        l.run();
        double laplaceCost = l.getCost() + parameterCost(false);
        if(laplaceCost < minCost){
            minCost = laplaceCost;
            minIndex = 2;
        }
        FilterKM l_rot = new FilterKM(copyObj(), 1, precision, true, laplace);
        l_rot.run();
        double laplaceCost_rot = l_rot.getCost() + parameterCost(true);
        if(laplaceCost_rot < minCost){
            minCost = laplaceCost_rot;
            minIndex = 3;
        }
        long end = System.currentTimeMillis();
        long ms = end-start;
        //long seconds = ms/1000;
        long seconds = ms;
        System.out.println("duration " + seconds + " s.");
        if(verbose){
            System.out.println("Costs: ");
            System.out.println("gaussCost " + gaussCost + " gaussCost_rot " + gaussCost_rot + " laplaceCost " + laplaceCost + " laplaceCost_rot " + laplaceCost_rot);
            System.out.println("selected model " + minIndex);
        }
        DataObject[] result = new DataObject[numObj];
        switch(minIndex){
            case(0): result = g.getData(); break;
            case(1): result = g_rot.getData(); break;
            case(2): result = l.getData(); break;
            case(3): result = l_rot.getData(); break;
            default: System.err.println("wrong minIndex"); break;
        }
        Evaluation e = new Evaluation(result);
        e.evaluateFiltering();
        
    }
    
    
    public void runKMCost(){
        //init runtime
        long start = System.currentTimeMillis();
        //run all variants
        double minCost = Double.MAX_VALUE;
        int minIndex = -1;
        KMCost g = new KMCost(copyObj(), 1, precision, false, gauss);
        g.run();
        double gaussCost = g.getCost() + parameterCost(false);
        if(gaussCost < minCost){
            minCost = gaussCost;
            minIndex = 0;
        }
        KMCost g_rot = new KMCost(copyObj(), 1, precision, true, gauss);
        g_rot.run();
        double gaussCost_rot = g_rot.getCost() + parameterCost(true);
        if(gaussCost_rot < minCost){
            minCost = gaussCost_rot;
            minIndex = 1;
        }
        KMCost l = new KMCost(copyObj(), 1, precision, false, laplace);
        l.run();
        double laplaceCost = l.getCost() + parameterCost(false);
        if(laplaceCost < minCost){
            minCost = laplaceCost;
            minIndex = 2;
        }
        KMCost l_rot = new KMCost(copyObj(), 1, precision, true, laplace);
        l_rot.run();
        double laplaceCost_rot = l_rot.getCost() + parameterCost(true);
        if(laplaceCost_rot < minCost){
            minCost = laplaceCost_rot;
            minIndex = 3;
        }
        long end = System.currentTimeMillis();
        long ms = end-start;
        //long seconds = ms/1000;
        long seconds = ms;
        System.out.println("duration " + seconds + " s.");
        if(verbose){
            System.out.println("Costs: ");
            System.out.println("gaussCost " + gaussCost + " gaussCost_rot " + gaussCost_rot + " laplaceCost " + laplaceCost + " laplaceCost_rot " + laplaceCost_rot);
            System.out.println("selected model " + minIndex);
        }
        DataObject[] result = new DataObject[numObj];
        switch(minIndex){
            case(0): result = g.getData(); break;
            case(1): result = g_rot.getData(); break;
            case(2): result = l.getData(); break;
            case(3): result = l_rot.getData(); break;
            default: System.err.println("wrong minIndex"); break;
        }
        Evaluation e = new Evaluation(result);
        e.evaluateFiltering();
        
    }
    
    
    
    /**
     * Run algorithm based on sorting the objects w.r.t. coding cost.
     */
    public void runFilterSort(){
          //init runtime
        long start = System.currentTimeMillis();
        FilterSort f = new FilterSort(data, precision, false, false);
        f.filter();
        double costNotRot = f.getCost() + parameterCost(false);
        FilterSort fRot = new FilterSort(data, precision, true, false);
        fRot.filter();
        double costRot = fRot.getCost() + parameterCost(true);
        long end = System.currentTimeMillis();
        long ms = end-start;
         long seconds = ms/1000;
        //long seconds = ms;
        System.out.println("duration " + seconds + " s.");
        if(costRot < costNotRot){
            Evaluation e = new Evaluation(fRot.getData());
            e.evaluateFiltering();
        } else{
            Evaluation e = new Evaluation(f.getData());
            e.evaluateFiltering();
        }
    }
    
    
    
    
    
    public void runFinalFilter(){
        FilterSort f = new FilterSort(data, precision, false, true);
        f.finalFilter(true);
        Evaluation e = new Evaluation(f.getData());
        e.evaluateFiltering();
    }
    
    private DataObject[] copyObj(){
        DataObject[] d = new DataObject[data.length];
        for(int i = 0; i < d.length; i++)
            d[i] = (DataObject)data[i].clone();
        return d;
    }
    
    private double parameterCost(boolean rotated){
        if(rotated)
            return dim*32*2 + dim*dim * 32;
        else
            return dim *32*2;
    }
    
}
