/*
 * KMCost.java
 *
 * Created on 12. Januar 2006, 13:55
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package oci;
import java.util.*;
import Jama.*;

/**
 * k-means algorithm based on coding cost. Initialisation: arbitrary mean, all points in cluster. variance: noise: added variances, cluster: real variance.
 * cluster IDs: 0: cluster, 1:noise
 * @author boehm
 */
public class FilterKM {
    int k;
    DataObject[] data;
    Cluster[] clusters;
    Cluster[] noiseClusters;
    int iter;
    int numObj;
    int dim;
    int precision;
    boolean rotated;
    int numDistribution;  //gauss: 1, laplace: 2
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    boolean verbose = false;
    
    /**
     * Creates a new instance of KMCost
     */
    public FilterKM(DataObject[] data, int k, int precision, boolean rotated, int numDistribution) {
        this.data = data;
        dim = data[0].coord.length;
        this.k = k;
        this.numDistribution = numDistribution;
        clusters = new Cluster[k];
        for(int i = 0; i < clusters.length; i++){
            clusters[i] = new Cluster();
            clusters[i].var = new double[dim];
        }
        noiseClusters = new Cluster[k];
        for(int i = 0; i < clusters.length; i++){
            noiseClusters[i] = new Cluster();
            noiseClusters[i].var = new double[dim];
        }
        numObj = data.length;
        this.precision = precision;
        this.rotated = rotated;
        iter = 0;
    }
    
    public DataObject[] getData(){
        return data;
    }
    
    
    /**
     * returns the coding cost of the coordinates. Without parameter costs.
     */
    public double getCost(){
        double cost = 0.0;
        for(int i = 0; i < k; i++){
            cost+= clusters[i].cost;
            cost+= noiseClusters[i].cost;
        }
        return cost;
    }
    
    public Cluster[] getClusters(){
        return clusters;
    }
    
    public Cluster[] getNoiseClusters(){
        return noiseClusters;
    }
    
    public void run(){
        //long start = System.currentTimeMillis();
        if(verbose)
            System.out.println("KMCost " + "k: " + k + " numObj: " + numObj);
        init();
        for(int i = 0; i < k; i++){
            if(rotated)
                robustEstimate(clusters[i], false);
            else
                setMean(clusters[i]);
            noiseClusters[i].mean = clusters[i].mean;
            //seperateNoise(clusters[i]);
            randomInitIDs();
        }
        DataObject[] meansCl = new DataObject[k];
        
        iter++;
        boolean clusterChanged = true;
        while(clusterChanged && iter < 10000){
            iter++;
            //if(iter < 7) //erste Iteration: iter = 2
            determineNoiseVariances();
            //else
            //determineNoiseVariancesIter();
            if(rotated)
                determineClusterVariances();
            else
                determineClusterVariancesNotRot();
            clusterChanged = assignPoints();
            for(int i = 0; i < k; i++){
                if(clusters[i].numObj > 0){
                    if(rotated)
                        robustEstimate(clusters[i], false);
                    else
                        setMean(clusters[i]);
                    noiseClusters[i].mean = clusters[i].mean;
                }
            }
        }
        System.out.println("iter: " + iter);
//        long end = System.currentTimeMillis();
//        long seconds = end-start/1000;
//        System.out.println("duration " + seconds + " s.");
        //evaluate and draw result
//
        //Mediane zeichnen
        for(int i = 0; i < k; i++){
            DataObject m = new DataObject(clusters[i].mean, 0);
            m.clusterID = clusters[i].ID;
            meansCl[i] = m;
        }
        
        if(verbose){
            Evaluation e = new Evaluation(data);
            e.evaluateFiltering();
            System.out.println(iter + " iterations.");
        }
    }
    
    
    
    private void init(){
        //initiale Clusterzentren sampeln
        Random r = new Random(2);
        for(int i = 0; i < k; i++){
            double[] mean = new double[dim];
            int nextMean = r.nextInt(numObj);
            for(int j = 0; j < dim; j++){
                mean[j] = data[nextMean].coord[j];
            }
            clusters[i].mean = mean;
            if(i%2 == 0)
                clusters[i].ID = i;
            else
                clusters[i].ID = i+1;
            noiseClusters[i].ID = clusters[i].ID + 1;
            noiseClusters[i].mean = mean;
        }
        //Punkte das erste mal nach eukl. Dist. zuordnen
        for(int i = 0; i < numObj; i++)
            euclAssign(data[i]);
    }
    
    
    private void robustEstimate(Cluster c, boolean verbose){
        if(verbose)
            System.out.println("Cluster " + c.ID);
        DataObject[] obj = getObjects(c);
        double[] median = median(obj);
        c.mean = median;
        double[][] covarm = covarByMedian(data, median);
        Matrix covar = new Matrix(covarm);
        EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
        Matrix vectors = eig.getV();
        Matrix values = eig.getD();
        if(verbose){
            System.out.println("EW");
            values.print(5, 3);
            System.out.println("EV");
            vectors.print(5, 3);
        }
        c.dir = vectors.getArrayCopy();
        double maxEigen = -Double.MAX_VALUE ;
        for (int i=0 ; i<dim ; i++)
            maxEigen = Math.max(maxEigen, values.get(i,i));
        for (int i=0 ; i<dim ; i++)
            values.set(i,i,1/Math.max(values.get(i,i), maxEigen*0.001)) ;
        c.invDir =  (vectors.times(values).times(vectors.transpose())).getArrayCopy() ;
    }
    
    private void setMean(Cluster c){
        DataObject[] obj = getObjects(c);
        double[] mean = mean(obj);
        c.mean = mean;
    }
    private double[] mean(DataObject[] o){
        double[] mean = new double[o[0].coord.length];
        for(int i = 0; i < o.length; i++)
            for(int j = 0; j < mean.length; j++){
            mean[j] += o[i].coord[j];
            }
        for(int j = 0; j < mean.length; j++)
            mean[j] /= o.length;
        return mean;
    }
    
    
    /**
     * Sort according to Euclidean Distance/Mahanolobis Distance: 50% cluster, 50% noise
     * @param c Cluster
     */
    private void seperateNoise(Cluster c){
        DataObject[] obj = getObjects(c);
        double[] dist = new double[obj.length];
        if(rotated){
            Matrix m = new Matrix(c.invDir);
            for(int i = 0; i < obj.length; i++){
                Matrix x = rowVector(obj[i].coord).minus(rowVector(c.mean));
                dist[i]=x.times(m).times(x.transpose()).get(0,0) ;
            }
        } else{
            for(int i = 0; i < obj.length; i++)
                dist[i] = obj[i].distance(c.mean);
        }
        SortPack [] distcopy = SortPack.array(dist) ;
        int count = 0;
        for(int i = (int)Math.ceil(0.90 * obj.length); i < obj.length; i++){
            int index = obj[distcopy[i].index].number;
            data[index].clusterID++;
            if(k==1){
                noiseClusters[data[index].clusterID-1].numObj++;
                clusters[data[index].clusterID-1].numObj--;
            }
            count++;
        }
        if(verbose)
            System.out.println("NoiseCluster " + count);
    }
    
    
    
    /**
     * Sort out 50% noise based on coding cost w.r.t the considered distribution function.
     * @param c Cluster
     */
//    private void separateNoiseCost(Cluster c){
//        DataObject[] obj = getObjects(c);
//        double[] cost = new double[obj.length]; //coding cost
//        for(int i = 0; i < c.numObj; i++)
//            cost[i] = clusterCost(obj[i]);
//
//    }
    
    
    /**
     * Random intialisation of clusterIDS: 50% noise (ID 1), 50% cluster (ID 0)
     */
    private void randomInitIDs(){
        Random r = new Random();
        int counter = data.length/2;
        boolean[] used = new boolean[data.length];
        for(int i = 0; i < counter; i++){
            int next = r.nextInt(data.length);
            while(used[next])
                next = r.nextInt(data.length);
            data[next].clusterID++;
            used[next] = true;
        }
    }
    
    private void determineNoiseVariances(){
        for(int i = 0; i < k; i++){
            double addVar = 0.0;
            DataObject[] obj = getObjects(noiseClusters[i]);
            for(int j = 0; j < dim; j++)
                addVar += var(obj, noiseClusters[i].mean, j);
            for(int j = 0; j < dim; j++){
                noiseClusters[i].var[j] = addVar;
            }
            if(verbose)
                System.out.println("Noise " + noiseClusters[i].ID + " addvar " + addVar);
        }
    }
    
    private void determineNoiseVariancesIter(){
        for(int i = 0; i < k; i++){
            DataObject[] obj = getObjects(noiseClusters[i]);
            for(int j = 0; j < dim; j++){
                noiseClusters[i].var[j] = var(obj, noiseClusters[i].mean, j);
                if(verbose)
                    System.out.println("Noise " + noiseClusters[i].ID + " coord " + j + " var " + noiseClusters[i].var[j]);
            }
        }
    }
    
    private void determineClusterVariancesNotRot(){
        for(int i = 0; i < k; i++){
            DataObject[] obj = getObjects(clusters[i]);
            for(int j = 0; j < dim; j++){
                clusters[i].var[j] = var(obj, clusters[i].mean, j);
                if(verbose)
                    System.out.println("Cluster " + clusters[i].ID + " coord " + j + " var " + clusters[i].var[j]);
            }
        }
    }
    
    
    
    private void determineClusterVariances(){
        for(int i = 0; i < k; i++){
            DataObject[] obj = getObjects(clusters[i]);
            DataObject[] red = new DataObject[obj.length];
            Matrix m = new Matrix(clusters[i].dir);
            for(int l = 0; l < red.length; l++){
                double[] redCoord = new double[dim];
                Matrix x = rowVector(obj[l].coord);
                Matrix akt = x.times(m);
                for(int j = 0; j < dim; j++){
                    redCoord[j] = akt.get(0,j);
                }
                DataObject redObj = new DataObject(redCoord, obj[l].number);
                redObj.clusterID = obj[l].clusterID;
                red[l] = redObj;
            }
            double[] redMeanCoord = new double[dim];
            Matrix mx = rowVector(clusters[i].mean);
            Matrix akt = mx.times(m);
            for(int j = 0; j < dim; j++){
                redMeanCoord[j] = akt.get(0,j);
            }
            for(int j = 0; j < dim; j++){
                clusters[i].var[j] = var(red, redMeanCoord, j);
                if(verbose)
                    System.out.println("Cluster " + clusters[i].ID + " coord " + j + " var " + clusters[i].var[j]);
            }
        }
    }
    
    //stdv of coordinate numCoord
    private double var(DataObject[] o, double[] mean, int numCoord){
        double stdv = 0.0;
        for(int i = 0; i < o.length; i++)
            stdv += (o[i].coord[numCoord] - mean[numCoord]) * (o[i].coord[numCoord] - mean[numCoord]);
        return stdv/(double)o.length;
    }
    
    
    //nehme Gaussverteilung an, bestimme auch Kosten der Cluster
    private boolean assignPoints(){
        if(verbose){
            System.out.println();
            System.out.println("assignPoints");
        }
        boolean clusterChanged = false;
        double aktCost = 0.0;
        double aktNoiseCost = 0.0;
        for(int i = 0; i < k; i++){
            clusters[i].numObj = 0;
            noiseClusters[i].numObj = 0;
            clusters[i].cost = 0.0;
            noiseClusters[i].cost = 0.0;
        }
        for(int i = 0; i < numObj; i++){
            double minCost = Double.MAX_VALUE;
            int minIndex = k+1;
            int minPos = k+1;
            for(int j = 0; j < k; j++){
                aktCost = clusterCost(data[i], clusters[j]);
                if(aktCost < minCost){
                    minCost = aktCost;
                    minIndex = clusters[j].ID;
                    minPos = j;
                }
            }
            for(int j = 0; j < k; j++){
                aktNoiseCost = noiseCost(data[i], noiseClusters[j]);
                if(aktNoiseCost < minCost){
                    minCost = aktNoiseCost;
                    minIndex = noiseClusters[j].ID;
                    minPos = j;
                }
            }
            
            if(minIndex != data[i].clusterID){
                data[i].clusterID = minIndex;
                clusterChanged = true;
            }
            if(minIndex % 2 == 0){
                clusters[minPos].cost += minCost;
                clusters[minPos].numObj++;
            } else{
                noiseClusters[minPos].cost += minCost;
                noiseClusters[minPos].numObj++;
            }
        }
        if(verbose){
            double overallCost = 0.0;
            for(int i = 0; i < k; i++){
                overallCost += clusters[i].cost;
                overallCost += noiseClusters[i].cost;
                //System.out.println("clusterCost " + clusters[i].cost);
                //System.out.println("noiseCost " + noiseClusters[i].cost);
            }
            System.out.println("overall Cost " + overallCost);
            for(int i = 0; i < k; i++){
                System.out.println("Cluster " + clusters[i].ID + " " + clusters[i].numObj);
                System.out.println("NoiseCluster " + noiseClusters[i].ID + " " + noiseClusters[i].numObj);
            }
        }
        /*
        for(int i = 0; i < k; i++){
            clusters[i].cost += paramCost(clusters[i]);
            noiseClusters[i].cost += paramCost(noiseClusters[i]);
        }
         */
        return clusterChanged;
    }
    
    ///TEST
    private void checkIDs(){
        Vector ids = new Vector();
        for(int i = 0; i < data.length; i++)
            if(!ids.contains(data[i].clusterID))
                ids.add(data[i].clusterID);
        System.out.println("numIDS " + ids.size());
    }
    //TEST
    
    //einmal fr rotated einmal fr nicht rotated: nicht rotated cluster oder noise.
    private double clusterCost(DataObject o, Cluster c){
        double sumCost = 0.0;
        //cluster
        if(c.dir != null){
            Matrix m = new Matrix(c.dir);
            Matrix x = rowVector(o.coord);
            double[] redMeanCoord = new double[dim];
            Matrix mx = rowVector(c.mean);
            Matrix akt = mx.times(m);
            for(int j = 0; j < dim; j++){
                redMeanCoord[j] = akt.get(0,j);
            }
            Matrix aktc = x.times(m);
            for(int j = 0; j < dim; j++){
                double red = aktc.get(0,j);
                if(numDistribution == 1)
                    sumCost += gaussCost(red, redMeanCoord[j], c.var[j], precision);
                else
                    sumCost += laplaceCost(red, redMeanCoord[j], c.var[j], precision);
            }
        } else{
            for(int j = 0; j < dim; j++)
                if(numDistribution == 1)
                    sumCost += gaussCost(o.coord[j], c.mean[j], c.var[j], precision);
                else
                    sumCost += laplaceCost(o.coord[j], c.mean[j], c.var[j], precision);
        }
        return sumCost;
    }
    
    
    private double noiseCost(DataObject o, Cluster c){
        double sumCost = 0.0;
        for(int j = 0; j < dim; j++)
            sumCost += gaussCost(o.coord[j], c.mean[j], c.var[j], precision);
        return sumCost;
    }
    
    //Hilfsmethoden/////////////////////////////////////////////////////////////////////////////////////////////
    
    
    
    /*
    //Kosten von Cluster und NoiseCluster Paar
    private double clusterNoiseCost(Cluster c, Cluster noise){
        double clusterCost = 0.0;
        double noiseCost = 0.0;
        DataObject[] clusterObj = getObjects(c);
        DataObject[] noiseObj = getObjects(noise);
        for(int i = 0; i < clusterObj.length; i++)
            clusterCost+= cost(clusterObj[i], c);
        for(int i = 0; i < noiseObj.length; i++)
            noiseCost+= cost(noiseObj[i], noise);
        /*
        System.out.println("clusterCost " + clusterCost);
        System.out.println("noiseCost " + noiseCost);
        System.out.println("gesamt " + (clusterCost + noiseCost));
     
        return clusterCost + noiseCost;
    }
     */
    
    
    
    
    private double gaussCost(double x, double mean, double var, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double va = Math.max(var, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*va)+precision*ln10+Math.pow(x-mean,2)/2/va)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    private double laplaceCost(double x, double mean, double var, int precision){
        double stdv = Math.sqrt(var);
        double b = stdv/Math.sqrt(2);
        if (b<Math.pow(10, -precision)/10)
            return 0 ;
        double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
        double prob = Math.max(Double.MIN_VALUE, Math.min(1.0, pdf * Math.pow(10, -precision)));
        //return 100000;
        return Math.log(1.0/prob)/Math.log(2);
    }
    
    
    public int countClusterMembers(int clusterID){
        int counter = 0;
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == clusterID)
                counter++;
        /*
        if(clusterID > 0)
            clusters[clusterID/2].numObj = counter;
        else
            clusters[clusterID].numObj = counter;
         */
        return counter;
    }
    
    public DataObject[] getObjects(Cluster c){
        DataObject[] members = new DataObject[countClusterMembers(c.ID)];
        int counter = 0;
        for(int j = 0; j < data.length; j++)
            if(data[j].clusterID == c.ID){
            members[counter] = data[j];
            counter++;
            }
        return members;
    }
    
    public double [] median(DataObject[] members) {
        int numObj= members.length ;
        int dim = members[0].coord.length ;
        double[] result = new double[dim];
        double [] sortarray = new double[numObj] ;
        for (int i=0 ; i<dim ; i++) {
            double median = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sortarray[j]=members[j].coord[i] ;
            Arrays.sort(sortarray) ;
            if (numObj==1)
                result[i]=sortarray[0] ;
            else if (numObj%2 == 0)
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
            else
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
        }
        return result ;
    }
    
    public double[][] covarByMedian(DataObject[] data, double[] mean) {
        int numObj= data.length ;
        int dim = data[0].coord.length ;
        double [] sortarray = new double[numObj] ;
        double[][] cov = new double[dim][dim];
        if (numObj <15){
            for (int i=0 ; i<dim ; i++)
                cov[i][i]=1.0 ;
            return cov ;
        }
        for(int l1 = 0; l1 < dim; l1++) {
            for(int l2 = 0; l2 < dim; l2++) {
                for(int j = 0; j < numObj; j++)
                    sortarray[j] = (data[j].coord[l1]-mean[l1])*(data[j].coord[l2]-mean[l2]) ;
                
                Arrays.sort(sortarray) ;
                if (numObj%2 == 0)
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3
                            ;
            }
            
        }
        double maxdiff = 0.0 ;
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                sum += Math.abs(cov[i][j]) ;
            double diff = 1.1*sum - 2.1*cov[i][i] ;
            if (diff>maxdiff)
                maxdiff = diff ;
        }
        if (maxdiff > 0.0)
            for (int i=0 ; i<dim ; i++)
                cov[i][i] += maxdiff ;
        
        return cov ;
    }
    
    
    private void euclAssign(DataObject d){
        double minDist = Double.MAX_VALUE;
        int minIndex = Integer.MAX_VALUE;
        for(int i = 0; i < k; i++){
            double[] aktMean = new double[dim];
            for(int j = 0; j < dim; j++)
                aktMean[j] = clusters[i].mean[j];
            double aktDist = distance(d, aktMean);
            if(aktDist < minDist){
                minDist = aktDist;
                minIndex = i;
            }
        }
        d.clusterID = clusters[minIndex].ID;
        clusters[minIndex].numObj++;
    }
    
    private double distance(DataObject d, double[] mean){
        double sum = 0.0;
        for (int i = 0; i < d.coord.length; i++){
            double dist = ((d.coord[i] - mean[i])*(d.coord[i] - mean[i]));
            sum = sum + dist;
        }
        return Math.sqrt(sum);
    }
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
}

