package oci;
import java.util.*;
import java.io.*;
import Jama.*;

/**
 * My long version of the KNN were each element is added to the cluster and the clustercost is estimated and only the element
 * with the lowest cost is taken
 * @author khaegler
 */
public class KNN_long {
    
    public int minKNN;
    public int dim;
    public int numberObjects; // the number of points in the initial data set
    
    public KNN_long(int minKNN){
        this.minKNN = minKNN;
    }
    
    public void getAllNN(DataObject[] o, int dimension){

        // Start the runtime check of the program
        long startTime = System.currentTimeMillis();
       
        this.dim = dimension;
        numberObjects = o.length;
        // fills the initialNN variable with the NN derived by the EuclideanDistance
        setInitialNNs(o);
        // The outlier cost has to be calculated for all nearest neighbors > 20
        // the first minKNN points are already determined and the point itself doesn't have to be considered 
        
        for (int i = 0; i < numberObjects; i++){
            ArrayList<Integer> noNN = new ArrayList<Integer>();
            for (int j = 0; j < numberObjects; j++){
                boolean iNN = false;
                for (int k = 0; k < (minKNN+1); k++){
                    if (o[j].number == o[i].NNList.get(k)){
                        iNN = true;
                        break;
                    }
                }
                if (!iNN){
                    // The objects not yet assigned to the NNList
                    noNN.add(o[j].number); 
                }
            }
            o[i].NoNNList = (ArrayList<Integer>)noNN.clone(); 
        }
        
        // for loop should be going from 0 until numberObjects
        for (int i = 0; i < numberObjects;i++){
            o[i].kNNCost = new ArrayList<Double>();
            // while loop should be going until NoNNList.size() is 0
            while (o[i].NoNNList.size() > 0){
                System.out.println("Object "+o[i].number+" "+o[i].NoNNList.size()+" NN left.");
                int NNid = 0;
                double clusterCost = Double.MAX_VALUE; 
                for (int j = 0; j < o[i].NoNNList.size(); j++){    

                    IcaCluster resultNoRot = new IcaCluster ();
                    IcaCluster resultRot = new IcaCluster ();
                    //int NNListSize = o[i].NNList.size()+1;
                    //DataObject[] testKNN = new DataObject[NNListSize];
                    double[][] coord = new double[o[i].NNList.size()+1][];
                    for (int k = 0; k < o[i].NNList.size();k++){
                        //testKNN[k] = (DataObject)o[o[i].NNList.get(k)].clone();
                        coord[k] = o[o[i].NNList.get(k)].coord.clone();
                    }
                    coord[o[i].NNList.size()] = o[o[i].NoNNList.get(j)].coord.clone();

                    int NewObjectOfInterest = o[i].NoNNList.get(j);
                    ArrayList<Integer> NNList = (ArrayList<Integer>)o[i].NNList.clone();
                    NNList.add(NewObjectOfInterest);
                    //testKNN[NNListSize-1] = (DataObject)o[NewObjectOfInterest].clone(); 
                    
                    resultNoRot = mergedClusterNoRot(NNList, coord);
                    resultRot = mergedCluster(NNList, coord);
                    
                    if (resultNoRot.cost < resultRot.cost && (Double)resultNoRot.cost < clusterCost) {
                        clusterCost = resultNoRot.cost;
                        NNid = j;
                    } else if ((Double)resultRot.cost < clusterCost){
                        clusterCost = resultRot.cost;
                        NNid = j;
                    }
                }
                o[i].NNList.add(o[i].NoNNList.get(NNid));
                o[i].kNNCost.add(clusterCost);
                o[i].NoNNList.remove(NNid);
            }
        }
        
        try {
            String filename = "KNN_clustercost_2Clu_niko.txt";
            BufferedWriter out = new BufferedWriter(new FileWriter(filename));    
            for (int i = 0; i < o[i].kNNCost.size();i++){               
                for (int j = 0; j < numberObjects; j++){
                    int x_coord = i+(minKNN+1);
                    double y_coord = o[j].kNNCost.get(i);
                    if (j == 0){
                        out.write(x_coord+"\t"+y_coord);
                    }else if (j == (numberObjects-1)){
                        out.write("\t"+y_coord+"\n");
                    }else{
                        out.write("\t"+y_coord);
                    }
                }
            }
            out.close();
        } catch (IOException e) {}

        // End the runtime check of the program
        long endTime = System.currentTimeMillis();

        GregorianCalendar g1 = new GregorianCalendar();
        GregorianCalendar g2 = new GregorianCalendar();
        g1.setTimeInMillis(startTime);
        g2.setTimeInMillis(endTime);

        int hours = g2.get(GregorianCalendar.HOUR_OF_DAY) - g1.get(GregorianCalendar.HOUR_OF_DAY);
        int minutes = g2.get(GregorianCalendar.MINUTE) - g1.get(GregorianCalendar.MINUTE);
        int seconds = g2.get(GregorianCalendar.SECOND) - g1.get(GregorianCalendar.SECOND);

        System.out.println("Runtime: "+ hours+" hours "+minutes+" min "+seconds+" sec.");

        return ;
    }

    public void setInitialNNs(DataObject[] o){
        
        DataObject[] q = o.clone();
        for (int i = 0; i < numberObjects; i++){

            double[] minDistances = new double[minKNN+1];
            ArrayList<Integer> minObjects = new ArrayList<Integer>(minKNN+1);
            for (int j = 0; j < numberObjects; j++){
                double dist = o[i].euclideanDistance(q[j]);
                // if dist is smaller than the last element in minDistances (the object having 
                // the largest distance to the object i) it is replaced with dist 
                if (j >= (minKNN+1)){
                    sortArrayList(minDistances, minObjects);
                    int lastObject = minDistances.length-1;
                    if (minDistances[lastObject] > dist){
                        minDistances[lastObject] = dist;
                        minObjects.set(lastObject, q[j].number);
                    }
                    
                // the first minKNN objects are added as initial minDistances Array
                }else if (j < (minKNN+1)){
                    minDistances[j] = dist;
                    minObjects.add(q[j].number);
                }
            }
            o[i].NNList = (ArrayList<Integer>)minObjects.clone();
        }
        
        return ;
    }

    public void sortArrayList (double[] data, ArrayList<Integer> minObjects) {
        
        for (int i = 0; i < data.length; i++) {
            for (int j = 0; j < i; j++) {
                if ( data[j] > data[j+1]) {
                    int tmpO = minObjects.get(j);
                    double tmp = data[j];
                    minObjects.set(j, minObjects.get(j+1));
                    data[j] = data[j+1];
                    minObjects.set(j+1, tmpO);
                    data[j+1] = tmp;
                    
                    j = 0;
                }
            }
        }
        
        return ;
    }
    
    //COST: writes coding cost
    public IcaCluster mergedClusterNoRot(ArrayList<Integer> assignedObjects, double[][] d) {
        
        IcaCluster result = new IcaCluster();
        
        int currentNumberObjects = assignedObjects.size();
        result.codingCostObjects = new double[currentNumberObjects];
        result.assignedObjects = new int[currentNumberObjects];
        for (int i = 0; i < currentNumberObjects; i++) {
            result.assignedObjects[i] = assignedObjects.get(i);
        }
        
        if (currentNumberObjects == 0) {
            //empty cluster due to split
            System.err.println("empty");
        }
  
        double[] realCenter = new double[dim]; //center of cluster pdf
        double[][] realRotation = new double[dim][dim]; //rotation+scaling
        double[] beta = new double[dim];
        GenGauss gg = new GenGauss();
        //Ica ica = new Ica();
/*        double[][] d = new double[currentNumberObjects][];
        for (int j = 0; j < currentNumberObjects; j++) {
            d[j] = o[j].coord.clone();
        }*/
        //double[][] transform = new double[dim][dim];
        //double[][] retransform = new double[dim][dim];
        //double[] shift = new double[dim];
        //double centered[][] = ica.centerData(d, shift);
        //double whitened[][] = ica.whitenData(centered, transform, retransform);
        double bestcost = Double.MAX_VALUE;
        //Matrix localbestmw = new Matrix(1,1) ;
        double[][] localbestgenGaussParams = new double[0][];

        //double[][] w = ica.optimizeEntropy(whitened);
        //Matrix mwhitened = new Matrix(whitened);
        //Matrix mw = new Matrix(w);
        //double[][] projectedWhitened =( mwhitened.times(mw.transpose())).transpose().getArrayCopy();
        double[][] genGaussParams = new double[dim][];
        //double[] stdGauss = {0.0,1.0,0.0} ;
        // now determine the shape parameter for every dimension
        for (int j = 0; j < dim; j++) {
            double[] s = new double[currentNumberObjects];
            for (int k = 0; k < currentNumberObjects; k++) {
                s[k] = d[k][j];
            }
            genGaussParams[j] = EpdMineo.optimizeMSP(s);
        }

        double cost = 0.0;
        // determine cost for every object in d
        for (int j = 0; j < currentNumberObjects; j++) {
            for (int j2 = 0; j2 < dim; j2++) {
                double aktCost = gg.costV(d[j][j2], genGaussParams[j2]);
                cost += aktCost;
                result.codingCostObjects[j] += aktCost;
            }
        }
        if (bestcost > cost) {
            bestcost = cost;
            //localbestmw = mw ;
            localbestgenGaussParams = new double[dim][];
            for (int j = 0; j < dim; j++) {
                localbestgenGaussParams[j] = genGaussParams[j].clone();
            }
            double[] mu = new double[dim];
            for (int j = 0; j < mu.length; j++) {
                mu[j] = localbestgenGaussParams[j][0];
            }
            realCenter = mu;
            double[][] sigmaDiag = new double[dim][dim];
            for (int j = 0; j < dim; j++) {
                sigmaDiag[j][j] = 1.0 / localbestgenGaussParams[j][1];
            }
            realRotation = sigmaDiag;
            for (int j = 0; j < dim; j++) {
                beta[j] = localbestgenGaussParams[j][2];
            }
        }

        double pCostNotRot = dim * 32 * 2;
        
        result.numObj = currentNumberObjects;
        result.dim = dim;
        result.realCenter = realCenter;
        result.realRotation = realRotation;
        result.beta = beta;
        result.cost = bestcost+pCostNotRot;
        
        result.codingCost = bestcost;
        return result;
    }
    
    //COST: writes coding cost using ICA and EPD
    public IcaCluster mergedCluster(ArrayList<Integer> assignedObjects, double[][] d) {
        
        IcaCluster result = new IcaCluster();
        
        int currentNumberObjects = assignedObjects.size();
        result.codingCostObjects = new double[currentNumberObjects];
        result.assignedObjects = new int[currentNumberObjects];
        for (int i = 0; i < currentNumberObjects; i++) {
            result.assignedObjects[i] = assignedObjects.get(i);
        }
 
        if (currentNumberObjects == 0) {
            //empty cluster due to split
            System.err.println("empty");
        }

        double[] realCenter = new double[dim]; //center of cluster pdf
        double[][] realRotation = new double[dim][dim]; //rotation+scaling
        double[] beta = new double[dim];
        GenGauss gg = new GenGauss();
        // gg = new GenGaussOld() ;
        Ica ica = new Ica();
/*        double[][] d = new double[currentNumberObjects][];
        for (int j = 0; j < currentNumberObjects; j++) {
            d[j] = o[j].coord.clone();
        }*/
        double[][] transform = new double[dim][dim];
        double[][] retransform = new double[dim][dim];
        double[] shift = new double[dim];
        double centered[][] = ica.centerData(d, shift);
        double whitened[][] = ica.whitenData(centered, transform, retransform);
        double bestcost = Double.MAX_VALUE;
        Matrix localbestmw = new Matrix(1, 1);
        double[][] localbestgenGaussParams = new double[0][];
        
        double[][] w = ica.optimizeEntropy(whitened);
        Matrix mwhitened = new Matrix(whitened);
        Matrix mw = new Matrix(w);
        double[][] projectedWhitened = (mwhitened.times(mw.transpose())).transpose().getArrayCopy();
        double[][] genGaussParams = new double[dim][];
        double[] stdGauss = {0.0, 1.0, 0.0};
        // now determine the shape parameter for every dimension
        for (int j = 0; j < dim; j++) {
            try {
                //genGaussParams [j] = gg.optimizeMSB(projectedWhitened [j]) ; //GenGauss
                genGaussParams[j] = EpdMineo.optimizeMSP(projectedWhitened[j]); //GenGauss
                for (int k = 0; k < genGaussParams[j].length; k++) {
                    if (Double.isNaN(genGaussParams[j][k])) {
                        System.out.println("------genGaussParams[" + j + "][" + k + "] NaN-----------");
                        genGaussParams[j] = stdGauss;
                    }
                }
            } catch (java.lang.StackOverflowError e) {
                System.out.println("-----stack overflow error-------");
                genGaussParams[j] = stdGauss;
            }
        }
        double cost = 0.0;
        // determine cost for every object in d
        for (int j = 0; j < currentNumberObjects; j++) {
            for (int j2 = 0; j2 < dim; j2++) {
                double aktCost = gg.costV(projectedWhitened[j2][j], genGaussParams[j2]);
                cost += aktCost;
                result.codingCostObjects[j] += aktCost;
            }
            result.codingCostObjects[j] += ica.wscCost;
        }
        cost += ica.wscCost * currentNumberObjects;
        if (bestcost > cost) {
            bestcost = cost;
            localbestmw = mw;
            localbestgenGaussParams = new double[dim][];
            for (int j = 0; j < dim; j++) {
                localbestgenGaussParams[j] = genGaussParams[j].clone();
            }
            double[] mu = new double[dim];
            for (int j = 0; j < mu.length; j++) {
                mu[j] = localbestgenGaussParams[j][0];
            }
            realCenter = (rowVector(shift).plus(rowVector(mu).times(localbestmw.times(new Matrix(retransform)).transpose()))).getRowPackedCopy();
            double[][] sigmaDiag = new double[dim][dim];
            for (int j = 0; j < dim; j++) {
                sigmaDiag[j][j] = 1.0 / localbestgenGaussParams[j][1];
            }
            realRotation = new Matrix(transform).times(localbestmw.transpose()).times(new Matrix(sigmaDiag)).getArrayCopy();
            for (int j = 0; j < dim; j++) {
                beta[j] = localbestgenGaussParams[j][2];
            }
        }
        double pCostRot = dim * 32 * 2 + dim * dim * 32;
        
        result.numObj = currentNumberObjects;
        result.dim = dim;
        result.realCenter = realCenter;
        result.realRotation = realRotation;
        result.beta = beta;
        result.cost = bestcost+pCostRot;
        result.codingCost = bestcost;
      
        return result;
    }
    
    public Matrix rowVector(double[] v) {
        return new Matrix(v, 1);
    }
}
