package outlier;
import compr.*;
import java.util.*;
/*
 * LOF.java
 *
 * Created on 23. November 2005, 15:36
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

/**
 *
 * @author boehm
 */
public class LOF {
    int k;
    DataObject[] data;
    
    /** Creates a new instance of LOF */
    public LOF(DataObject[] data, int k) {
        this.data = data;
        this.k = k;
    }
    
    //reachability distance of p w.r.t. o
    private double reachDist(DataObject p, DataObject o){
        return Math.min(knnDistance(o), o.distance(p));
    }
    
    public double[] lofData(){
        double[] lofData = new double[data.length];
        for(int i = 0; i < data.length; i++)
            lofData[i] = computeLof(data[i]);
        return lofData;
    }
    
    public double computeLof(DataObject p){
        double lrd_p = lrd(p);
        double knnDist = knnDistance(p);
        DataObject[] nn = neighborhood(p, knnDist);
        double sumRel = 0.0;
        for(int i = 0; i < nn.length; i++)
            sumRel+= lrd(nn[i])/lrd_p;
        return sumRel/(double)nn.length;   
    }
    
    private double lrd(DataObject p){
        double knnDist = knnDistance(p);
        DataObject[] nn = neighborhood(p, knnDist);
        double sumReachDist = 0.0;
        for(int i = 0; i < nn.length; i++)
            sumReachDist+= reachDist(p, nn[i]);
        return (double)1/(sumReachDist/(double)nn.length);
    }
    
    private DataObject[] neighborhood(DataObject p, double knnDist){
        Vector n = new Vector();
        for(int i = 0; i < data.length; i++)
            if(data[i].euclideanDistance(p)<= knnDist)
                n.addElement(data[i]);
        DataObject[] neighbors = new DataObject[n.size()];
        for(int i = 0; i < neighbors.length; i++)
            neighbors[i] = (DataObject)n.elementAt(i);
        return neighbors;
    }
     
    
    private double knnDistance(DataObject o){
        double[] erg = new double[k];        //Array der Grösse MinPts anlegen
        DataObject[] knn = new DataObject[k];  //für knn
        for(int i=0; i<k; i++) {
            erg[i] = Double.MAX_VALUE;  //alle auf Max_Value setzen.
        }
        for(int i=0; i < data.length; i++) {   //Durchlauf alle Datenobjekte
            
            int c = k-1;
            double d = o.euclideanDistance(data[i]); //Distanz zum i-ten Datenobjekt
            while(c>=0 && d<erg[c]) { //MinPts >= 0 && Distanz zum i-ten Datenobjekt < erg[MinPts-1]
                if(c<k-1){
                    erg[c+1] = erg[c];
                    knn[c+1] = knn[c];
                }
                erg[c] = d;
                knn[c] = data[i];
                c--;                  //MinPts um 1 reduzieren
            }
        }
        return erg[k-1];
    }
    
    
    private DataObject[] knn(DataObject o){
        double[] erg = new double[k];        //Array der Grösse MinPts anlegen
        DataObject[] knn = new DataObject[k];  //für knn
        for(int i=0; i<k; i++) {
            erg[i] = Double.MAX_VALUE;  //alle auf Max_Value setzen.
        }
        for(int i=0; i < data.length; i++) {   //Durchlauf alle Datenobjekte
            
            int c = k-1;
            double d = o.distance(data[i]); //Distanz zum i-ten Datenobjekt
            while(c>=0 && d<erg[c]) { //MinPts >= 0 && Distanz zum i-ten Datenobjekt < erg[MinPts-1]
                if(c<k-1){
                    erg[c+1] = erg[c];
                    knn[c+1] = knn[c];
                }
                erg[c] = d;
                knn[c] = data[i];
                c--;                  //MinPts um 1 reduzieren
            }
        }
        return knn;
        
    }
}
