/*
 * Statistics.java
 *
 * Created on 11. September 2006, 11:32
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package utils;
import java.util.*;
import compr.DataObject;
import eval.SortPack;
import eval.Cluster;
import Jama.*;

/**
 *
 * @author boehm
 */
public class Statistics {
    int numObj;
    int dim;
    
    /** Creates a new instance of Statistics */
    public Statistics(int numObj, int dim) {
        this.numObj = numObj;
        this.dim = dim;
    }
    
    
    public double[] mean(double[][]o){
        double[] mean = new double[dim];
        for(int i = 0; i < numObj; i++)
            for(int j = 0; j < dim; j++){
            mean[j] += o[i][j];
            }
        for(int j = 0; j < dim; j++)
            mean[j] /= numObj;
        return mean;
    }
    
    
    public double sigma(double[][]o, double[] mean, int coordIndex){
        double sigma = 0.0;
        for(int i = 0; i < numObj; i++)
            sigma += (o[i][coordIndex]-mean[coordIndex]) * (o[i][coordIndex]-mean[coordIndex]);
        sigma /= (double)numObj;
        //return Math.sqrt(sigma);
        return sigma;
    }
    
    
    public double[] mean(DataObject[] d){
        double[][] o = new double[d.length][d[0].coord.length];
        for(int i = 0; i < d.length; i++)
            for(int j = 0; j < o[i].length; j++)
                o[i][j] = d[i].coord[j];
        return mean(o);
    }
    
    
    public double sigma(DataObject[] d, double[] mean, int coordIndex){
        double[][] o = new double[d.length][d[0].coord.length];
        for(int i = 0; i < d.length; i++)
            for(int j = 0; j < o[i].length; j++)
                o[i][j] = d[i].coord[j];
        return sigma(o, mean, coordIndex);
    }
    
    
    public double [] median(DataObject[] d) {
        double[] result = new double[dim];
        double [] sortarray = new double[numObj] ;
        for (int i=0 ; i<dim ; i++) {
            double median = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sortarray[j]=d[j].coord[i] ;
            Arrays.sort(sortarray) ;
            if (numObj==1)
                result[i]=sortarray[0] ;
            else if (numObj%2 == 0)
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
            else
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
        }
        return result ;
    }
    
   
    private void robustEstimate(Cluster c, DataObject[] obj){
        boolean verbose = false;
        double[] median = median(obj);
        double[][] covarm = covarByMedian(obj, median);
        Matrix covar = new Matrix(covarm);
        EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
        Matrix vectors = eig.getV();
        Matrix values = eig.getD();
        if(verbose){
            System.out.println("EW");
            values.print(5, 3);
            System.out.println("EV");
            vectors.print(5, 3);
        }
        c.dir = vectors.getArrayCopy();
        double maxEigen = -Double.MAX_VALUE ;
        for (int i=0 ; i<dim ; i++)
            maxEigen = Math.max(maxEigen, values.get(i,i));
        for (int i=0 ; i<dim ; i++)
            values.set(i,i,1/Math.max(values.get(i,i), maxEigen*0.001)) ;
        c.invDir =  (vectors.times(values).times(vectors.transpose())).getArrayCopy() ;
}  
    
     public double[][] covarByMedian(DataObject[] data, double[] mean) {
        int numObj= data.length ;
        int dim = data[0].coord.length ;
        double [] sortarray = new double[numObj] ;
        double[][] cov = new double[dim][dim];
        if (numObj <15){
            for (int i=0 ; i<dim ; i++)
                cov[i][i]=1.0 ;
            return cov ;
        }
        for(int l1 = 0; l1 < dim; l1++) {
            for(int l2 = 0; l2 < dim; l2++) {
                for(int j = 0; j < numObj; j++)
                    sortarray[j] = (data[j].coord[l1]-mean[l1])*(data[j].coord[l2]-mean[l2]) ;
                
                Arrays.sort(sortarray) ;
                if (numObj%2 == 0)
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3
                            ;
            }
            
        }
        double maxdiff = 0.0 ;
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                sum += Math.abs(cov[i][j]) ;
            double diff = 1.1*sum - 2.1*cov[i][i] ;
            if (diff>maxdiff)
                maxdiff = diff ;
        }
        if (maxdiff > 0.0)
            for (int i=0 ; i<dim ; i++)
                cov[i][i] += maxdiff ;
        
        return cov ;
    }
    
}